#include <windows.h>
#include <math.h>
#include <gl/gl.h>
#include "fhand.h"
#include "model.h"
#include "funcs.h"
#include "vert.h"
#include "face.h"
#include "bone.h"
#include "mystr.h"
#include "funcs.h"
#include "calc.h"
#include "lsmtypes.h"
#include "mouse.h"
#include "view.h"
#include "appdata.h"

BoneData::BoneData()
    : Parent((Bone*)0)
    , twistAngle(0.0f)
    , bendAngle(0.0f)
    , turnAngle(0.0f)
    , scale(1.0f)
    , translate(0.0f) { }

BoneData::~BoneData() { }

Bone::Bone() {
    BoneEntry* be=new BoneEntry;
    history.addTail(be);
}

Bone::~Bone() {
    destroyHistory();
}

bool BoneData::operator!=(BoneData& obd) {
    if(Name!=obd.Name) return true;
    if(Parent!=obd.Parent) return true;
    if(scale!=obd.scale) return true;
    if(translate!=obd.translate) return true;
    if(twistAngle!=obd.twistAngle) return true;
    if(bendAngle!=obd.bendAngle) return true;
    if(turnAngle!=obd.turnAngle) return true;
//      if(flags!=obd.flags) return true;
    return false;
}

void BoneData::operator=(BoneData& obd) {
        Name=obd.Name;
        Parent=obd.Parent;
        scale=obd.scale;
        translate=obd.translate;
        twistAngle=obd.twistAngle;
        bendAngle=obd.bendAngle;
        turnAngle=obd.turnAngle;
}

// this creates/refreshes the matrices used by the bone.
void Bone::buildTransforms() {
        Vec3 r;

        Transform temp;

// convert translate vector to new final location
        Spot=data.translate;
        if(data.Parent) {
            Spot*=data.Parent->Scaled;
        }

// build local children
        Scaled.identity();
        notScaled.identity();

// scale the scaled transform
        Scaled.scaleBy(data.scale);

// rotate both around the z axis
        r.x=r.y=0.0;
        r.z=data.twistAngle;

        Scaled.rotateBy(r);
        notScaled.rotateBy(r);

// rotate both around the x and y axes, in that order
        r.x=data.bendAngle;
        r.y=data.turnAngle;
        r.z=0.0;

        Scaled.rotateBy(r);
        notScaled.rotateBy(r);

// now multiply by inherited non-scaled transform, if there is one
        if(data.Parent) {
            Scaled.transformBy(data.Parent->notScaled);
            notScaled.transformBy(data.Parent->notScaled);
        }

// and then place the new translation vector in
        Scaled.clearTranslate();
        notScaled.clearTranslate();
        Scaled.translateBy(Spot);
        notScaled.translateBy(Spot);

        flags&=~TF_NEW; //clear the flag for needing to be processed
}

// replace later with smarter code
//float Bone::clipSize(float s) {
//    if(s>1.0f) return 1.0f;
//    return s;
//}

void Bone::draw(float size) {

    glEnable(GL_CULL_FACE);
//    size=clipSize(size);

    notScaled.loadIntoOpenGL();
    glScalef(size,size,size);
    drawDart();

    glScalef(.1,.1,.1);
    glCallList(AppData::GLObjBase+AD_OBJ_BALL);

    glTranslatef(0.0f,0.0f,10.0f);
    glScalef(.75,.75,.75);
    glCallList(AppData::GLObjBase+AD_OBJ_BALL);

    glDisable(GL_CULL_FACE);

//    glLoadIdentity();
//    glPointSize(15.0);
//    glBegin(GL_POINTS);
//    test.glPlot();
//    glEnd();
}

void Bone::drawRotation2(float size,short mode,Vec3& /*vf*/) {
    GLfloat mod[16];
    Vec3    temp;

    glMatrixMode(GL_MODELVIEW);
    if(data.Parent) {
        data.Parent->notScaled.fore.loadIntoOpenGL();
        glTranslatef(
            data.translate.x*data.Parent->data.scale.x,
            data.translate.y*data.Parent->data.scale.y,
            data.translate.z*data.Parent->data.scale.z);
    } else {
        glLoadIdentity();
        glTranslatef(data.translate.x,data.translate.y,data.translate.z);
    }
    glScalef(size,size,size);

//    glPointSize(5.0);

// draw the Y-axis ring
    glGetFloatv(GL_MODELVIEW_MATRIX,mod);
    temp.set(mod[4],mod[5],mod[6]);
    temp.normalize();

//    if(fabs(vf.dot(temp))>.01) {
        SelectGLColor(AppData::PenColor[YAXISPEN]);
        if(mode==MM_GADOVERY) glLineWidth(3.0f); else glLineWidth(1.0f);
        glLoadName(MM_GADOVERY);
        glCallList(AppData::GLObjBase+AD_OBJ_FLIPYZ);
        glCallList(AppData::GLObjBase+AD_OBJ_CIRCLE);
        glCallList(AppData::GLObjBase+AD_OBJ_FLIPYZ);
//    }

    glRotatef(data.turnAngle,0.0f,1.0f,0.0f);
    glGetFloatv(GL_MODELVIEW_MATRIX,mod);
    temp.set(mod[0],mod[1],mod[2]);
    temp.normalize();

//    if(fabs(vf.dot(temp))>.01) {
        if(mode==MM_GADOVERX) glLineWidth(3.0f); else glLineWidth(1.0f);
        SelectGLColor(AppData::PenColor[XAXISPEN]);
        glLoadName(MM_GADOVERX);
        glCallList(AppData::GLObjBase+AD_OBJ_FLIPXZ);
        glCallList(AppData::GLObjBase+AD_OBJ_CIRCLE);
        glCallList(AppData::GLObjBase+AD_OBJ_FLIPXZ);
//    }

    glRotatef(data.bendAngle,1.0f,0.0f,0.0f);
    glGetFloatv(GL_MODELVIEW_MATRIX,mod);
    temp.set(mod[8],mod[9],mod[10]);
    temp.normalize();

//    if(fabs(vf.dot(temp))>.01) {
        if(mode==MM_GADOVERZ) glLineWidth(3.0f); else glLineWidth(1.0f);
        SelectGLColor(AppData::PenColor[ZAXISPEN]);
        glLoadName(MM_GADOVERZ);
        glCallList(AppData::GLObjBase+AD_OBJ_CIRCLE);
//    }

}

void Bone::drawTranslation(float size,short /* mode */,Vec3& /*vf*/) {

    glMatrixMode(GL_MODELVIEW);

    if(data.Parent) {
        data.Parent->notScaled.loadIntoOpenGL();
        glTranslatef(
            data.Parent->data.scale.x * data.translate.x,
            data.Parent->data.scale.y * data.translate.y,
            data.Parent->data.scale.z * data.translate.z );
    } else {
        glLoadIdentity();
        glTranslatef(
            data.translate.x,
            data.translate.y,
            data.translate.z );
    }

    glScalef(size,size,size);

    for(int i=0;i<6;i++) {
        glPushMatrix();

        switch(i) {
            case 0:
            case 3:
                glLoadName(MM_GADOVERX);
                SelectGLColor(AppData::PenColor[XAXISPEN]);
                break;
            case 1:
            case 4:
                glLoadName(MM_GADOVERY);
                SelectGLColor(AppData::PenColor[YAXISPEN]);
                glCallList(AppData::GLObjBase+AD_OBJ_FLIPXY);
                break;
            case 2:
            case 5:
                glLoadName(MM_GADOVERZ);
                SelectGLColor(AppData::PenColor[ZAXISPEN]);
                glCallList(AppData::GLObjBase+AD_OBJ_FLIPXZ);
                break;
        }

        if(i>2) {
            glScalef(-.2,.075,.075);
        } else {
          glScalef(.2,.075,.075);
        }
        glTranslatef(5.0f,0.0f,0.0f);
        glCallList(AppData::GLObjBase+AD_OBJ_CONE);
        glPopMatrix();
    }
}

// returns true if b is an ancestor of this
bool Bone::isParent(Bone* b) {
        Bone*   db;

        for(db=this;db!=0;db=db->data.Parent)
        if(db==b)
        return true;

        return false;
}

void Bone::selectBase(float size) {
//    size=clipSize(size);

    notScaled.loadIntoOpenGL();
    glScalef(size*0.1f,size*0.1f,size*0.1f);
    glCallList(AppData::GLObjBase+AD_OBJ_BALL);
}

void Bone::selectEnd(float size) {
//    size=clipSize(size);

    notScaled.loadIntoOpenGL();
    glTranslatef(0.0f,0.0f,size);
    glScalef(size*.075f,size*.075f,size*.075f);
    glCallList(AppData::GLObjBase+AD_OBJ_BALL);
}

void Bone::setXAngle(float ax) {
    data.bendAngle=ax-360.0f*floorf((ax-data.bendAngle)/360.0+.5);
}

#if(0)
void Bone::setXYAngles(float ax,float ay) {
    float ux=ax-360.0f*floorf((ax-data.bendAngle)/360.0+.5);
    float uy=ay-360.0f*floorf((ay-data.turnAngle)/360.0+.5);

    float vx=180.0f-ax-360.0f*floorf((180.0f-ax-data.bendAngle)/360.0+.5);
    float vy=ay+180.0f-360.0f*floorf((ay+180.0f-data.turnAngle)/360.0+.5);

    if(fabs(data.bendAngle-ux)+fabs(data.turnAngle-uy)>fabs(data.bendAngle-vx)+fabs(data.turnAngle-vy)) {
        data.bendAngle=vx;
        data.turnAngle=vy;
    } else {
        data.bendAngle=ux;
        data.turnAngle=uy;
    }
}
#endif

void Bone::setAngles(float ax,float ay,float az) {
    float ux=ax-360.0f*floorf((ax-data.bendAngle+180.0f)/360.0);
    float uy=ay-360.0f*floorf((ay-data.turnAngle+180.0f)/360.0);
    float uz=az-360.0f*floorf((az-data.twistAngle+180.0f)/360.0);

    float vx=180.0f-ax-360.0f*floorf((180.0f-ax-data.bendAngle+180.0f)/360.0);
    float vy=ay+180.0f-360.0f*floorf((ay+180.0f-data.turnAngle+180.0f)/360.0);
    float vz=az+180.0f-360.0f*floorf((az+180.0f-data.twistAngle+180.0f)/360.0);

    if ( fabs(data.bendAngle-ux)+fabs(data.turnAngle-uy)+fabs(data.twistAngle-uz)
        > fabs(data.bendAngle-vx)+fabs(data.turnAngle-vy)+fabs(data.twistAngle-vz) ) {
        data.bendAngle=vx;
        data.turnAngle=vy;
        data.twistAngle=vz;
    } else {
        data.bendAngle=ux;
        data.turnAngle=uy;
        data.twistAngle=uz;
    }
}

// this function calculates the scale, rotate, and translate values that
// would result in the current notScaled.fore matrix, based on the Bone's
// parentage
void Bone::reparent() {
        float cosx,sinx,cosy,siny,cosz,sinz;
        Matrix temp;

        temp=notScaled.fore;

        if(data.Parent) {
            temp/=data.Parent->notScaled.fore;
        }

// calculate the rotation angles that go with the new matrix
// right now this assumes that the bend angle is from -90 to +90,
// turn is -180 to +180, and twist is -180 to +180

// a,b=cos,sin of bend angle
// c,d=cos,sin of turn angle
// e,f=cos,sin of twist angle (performed first)
// total matrix is:
//  e  f  0    1  0  0    c  0 -d   bdf+ce, af, bcf-de
// -f  e  0 x  0  a  b x  0  1  0 = bde-cf, ae, bce+df
//  0  0  1    0 -b  a    d  0  c   ad,     -b, ac

        sinx=-temp.z.y; // the value of b in above matrix
        cosx = sqrtf(1-sinx * sinx);    // the value of a in above matrix

        if(cosx !=0.0) {
                cosy=temp.z.z/cosx; // the value of c in above matrix
                siny =temp.z.x/cosx;    // the value of d in above matrix
                cosz =temp.y.y/cosx;    // the value of e in above matrix
                sinz =temp.x.y/cosx;    // the value of f in above matrix
        } else {
// if the cosine of x==0, then we cannot determine from the matrix
// whether the x and z components of the x and y rows are from rotation
// around z or around y; we will assume the former.
                cosy=1;
                siny=0;
                cosz=temp.x.x;
                sinz=-temp.y.x;
        }

        // put inverses into angles

        data.twistAngle=atan2f(sinz,cosz)*180/PI;
        data.bendAngle=asinf(sinx)*180/PI;
        data.turnAngle=atan2f(siny,cosy)*180/PI;

        // calculate the translate value

    data.translate=Spot;
    if(data.Parent) {
        data.translate/=data.Parent->Scaled;
    }
}

// this may require extensive rewriting
void Bone::writeINCBuild(fWriter& file) {
        char *n,*pn;

        n=data.Name.c_string();

        pn=NULL;
        if(data.Parent) pn=data.Parent->data.Name.c_string();

// write the non-scaling transform
// first we write the scale
        file << "#declare " << n << "_tfu=transform { " ;
// and the rotations
        file << "rotate z*" << n << "_az rotate <" << n << "_ax," << n << "_ay,0> " ;
// and the translation
        file << "translate " << n << "_t" ;

        if(pn)
        file << "*" << pn << "_s transform " << pn << "_tfu" ;

        file << " }\15\12" ;

// and now the scaling transform
        file << "#declare " << n << "_tfs=transform { scale " << n << "_s transform "
                 << n << "_tfu }\15\12" ;
}

// undo/redo functions

bool Bone::takeSnap(INDEX newVersion) {
        BoneEntry*      be;
        BoneEntry*      lbe;
        bool make;
        TYPE    newType;

        // clip all history equal to or greater than new version
        clipHistory(newVersion);

        // most recent entry should be last
        lbe=(BoneEntry*)(history.last());

/* behavior logic:
for takeSnap():

last entry      current object
NONE            invalid                 flag as TF_KILLME
et_nul          invalid                 flag as TF_KILLME
et_chg          invalid                 archive as et_del
et_del          invalid                 do nothing

NONE            valid                   archive as et_chg
et_nul          valid                   archive as et_chg
et_chg          valid                   compare and archive as et_chg
et_del          valid                   shouldn't happen
*/

        make=false; // default is don't make new entry
        newType=ET_CHG; // most common type is change

        if(lbe==NULL) {
                if(flags&TF_INVALID) {
                        flags|=TF_KILLME;
                } else {
                        make=true;
                }
        } else {
                switch(lbe->type) {
                        case ET_NUL:
                                if(flags&TF_INVALID) {
                                        flags|=TF_KILLME;
                                } else {
                                        make=true;
                                        newType=ET_CHG;
                                }
                                break;
                        case ET_CHG:
                                if(flags&TF_INVALID) {
                                        make=true;
                                        newType=ET_DEL;
                                } else {
                                        // make it if latest is different from current
                                        if (lbe->data!=data) make=true;
                                }
                                break;
// last case is either do nothing or shouldn't happen
//                      case ET_DEL:
//                              break;
                }
        }


        if(make) {
                be=new BoneEntry;
                history.addTail(be);
                be->version=newVersion;
                be->type=newType;
                be->data=data;
        }

        return make;
}

void    Bone::clipHistory(INDEX version) {
        BoneEntry *be;
        BoneEntry *nbe;

        be=(BoneEntry*)(history.first());
        while(be!=NULL) {
                nbe=(BoneEntry*)(be->next());
                if(be->version>=version) {
                        be->remove();
                        delete be;
                }
                be=nbe;
        }
}

void Bone::destroyHistory() {
        BoneEntry*  be;

        while(history.last()) {
                be=(BoneEntry*)(history.last());
                be->remove();
                delete be;
        }
}

void Bone::restoreVersion(INDEX getVersion) {
        BoneEntry*  be;

        be=(BoneEntry*)(history.last());

// there is always an entry with version 0, so this should never back up
// such that ve==0
        while(be->version > getVersion)
        be=(BoneEntry*)(be->prev());

        if(be->type==ET_CHG) {
                data=be->data;
                flags&=~TF_INVALID;
        } else {
                flags|=TF_INVALID;
        }
}
