#ifndef BONE_H
#define BONE_H

#include "link.h"
#include "list.h"
#include "mystr.h"
#include "transform.h"

// parameters for Bones:
// Scales
// Twist (z axis)
// Side-side (y-axis)
// Bend (x axis)

struct BoneData {
        class Bone  *Parent;
        Mystr   Name;
        Vec3    scale;  // scaling applied to verts and child base points
        float   twistAngle; // z axis
        float   bendAngle;  // x axis
        float   turnAngle;  // y axis
        Vec3    translate;  // translation, applied to verts and children
//      FLAGS   flags;  // no need for flags yet, but there may be in the future

        BoneData();
        ~BoneData();
        bool operator!=(BoneData&);
        void operator=(BoneData&);
};

class BoneEntry : public Entry {
    public:
        BoneData    data;
};

class Bone : public Thing {
        public:
        eList   history;
        BoneData    data;
        Transform       Scaled;
        Transform       notScaled;
        Matrix  mSave;
        float twistSave,turnSave,bendSave;
        Vec3 fore;
        Vec3     Save;
        Vec3 Spot;
//        Vec3 test;
        int gen;    // the generation of this bone

        void    buildTransforms();
//        float   clipSize(float); // made into a member function of Model
        void    draw(float);
//        void    drawScaleGadget(float);
        void    drawRotation2(float,short,struct Vec3&);
        void    drawTranslation(float,short,struct Vec3&);
        bool    isParent(Bone*);
        void    selectBase(float);
        void    selectEnd(float);
        void    setAngles(float,float,float);
        void    setXAngle(float);
//        void    setXYAngles(float,float);
        void    reparent();
        void    unleash();
        void    writeINCBuild(class fWriter&);
        void    writeToFile(class fWriter&);
        Bone();
      ~Bone();

// undo/redo
        void    clipHistory(INDEX);
        bool    takeSnap(INDEX);
        void    restoreVersion(INDEX);
        void    destroyHistory();
};

//#define BF_LOCKX    16
//#define BF_LOCKY    32
//#define BF_LOCKZ    64
//#define BF_LOCKL    128
//#define BF_BAD_ANG    256 // angles are invalid
//#define BF_LOCKMASK   (BF_LOCKX|BF_LOCKY|BF_LOCKZ|BF_LOCKL)

#endif

