#ifndef CALC_H
#define CALC_H

#define sqrtf(x) ((float)sqrt((double)(x)))
#define atan2f(a,b) ((float)atan2((double)(a),(double)(b)))
#define asinf(x) ((float)asin((double)(x)))
#define floorf(x) ((float)floor((double)(x)))
#define sinf(x) ((float)sin((double)(x)))
#define cosf(x) ((float)cos((double)(x)))
#define expf(x) ((float)exp((double)(x)))
#define absf(x) ((float)abs((double)(x)))

// defined mathematical constants
const float HALFPI=1.570796326795f;
const float PI=3.14159265359f;
const float LN10=2.302585092994f;
const float RAD2DEG=57.29577951308f;
const float DEG2RAD=0.01745329251994f;
const float SQRT_3_4=0.86602540378444f;
const float SQRT_1_2=0.7071067811865475f;
const float SQRT_1_3=0.57735026918962576f;
const float SQRT_2=1.4142135623730951f;
const float BEZ_CIRC=0.55228474983079340f;
const float ERROR_CORRECTION=2.0f;

struct Vec3 {
    float x,y,z;
    Vec3();
    Vec3(float);
    Vec3(float,float,float);
    Vec3(float*);
    Vec3& cross(Vec3&,Vec3&);
    Vec3& operator=(float*);
    Vec3& operator*=(float);
    Vec3& operator*=(struct Matrix&);
    Vec3& operator*=(struct Transform&);
    Vec3& operator/=(float);
    Vec3& operator/=(struct Matrix&);
    Vec3& operator/=(struct Transform&);
    Vec3& operator+=(Vec3&);
    Vec3& operator-=(Vec3&);
    bool operator!=(Vec3&);
    void set(float);
    void set(float a,float b,float c) { x=a; y=b; z=c; }
    float dot(Vec3&);
    float dot(float*);
    void glPlot();
    void glNormal();
    void glRasterPos();
    float normalize();
    float normalize(Vec3&);
    float length();
    float distance(Vec3&);
    float distance2(Vec3&);
    void    put(float*);
};

struct Vec2 {
    float u,v;

    void glPlot();
    void set(float a) { u=v=a; }
    Vec2& operator+=(Vec2&);
    Vec2& operator-=(Vec2&);
    Vec2& operator/=(float a);
    Vec2& operator*=(float);
    void    put(float*);
};

// globals defined in calc.cpp
float errorCorrection(float);
float frac(float);
bool isPowerof2(unsigned int);
unsigned int nearestPowerof2(unsigned int);
bool nearest2Line(Vec3&,Vec3&,Vec3&,Vec3&,Vec3&);
void nearest2Plane(Vec3&,Vec3&,Vec3&,Vec3&);
bool traceCylinder(Vec3&,Vec3&,float,Vec3&,Vec3&,Vec3&);
bool tracePlane(Vec3&,Vec3&,Vec3&,Vec3&,Vec3&);
bool traceSphere(Vec3&,float,Vec3&,Vec3&,Vec3&);

#endif

