#include "userint.h"
#include <commctrl.h>
#include <math.h>
#include <gl/glu.h>
#include "appdata.h"
#include "project.h"
#include "funcs.h"
#include "vert.h"
#include "patch.h"
#include "edge.h"
#include "face.h"
#include "uvpt.h"
#include "bone.h"
#include "mouse.h"
#include "pen.h"

void DrawMappedQuad(Vec2& ta,Vec3& pa,Vec2& tb,Vec3& pb,Vec2& tc,Vec3& pc,
        Vec2& td,Vec3& pd,COLORREF c,bool calc_norm) {
        GLfloat minx,miny;

        minx=ta.u;
        if(minx>tb.u) minx=tb.u;
        if(minx>tc.u) minx=tc.u;
        if(minx>td.u) minx=td.u;

        miny=ta.v;
        if(miny>tb.v) miny=tb.v;
        if(miny>tc.v) miny=tc.v;
        if(miny>td.v) miny=td.v;

        minx=floor(minx);
        miny=floor(miny);

        if(calc_norm) {
            Vec3 va,vb,vn;
            (va=pa)-=pc;
            (vb=pb)-=pd;
            vn.cross(va,vb);
            vn.normalize();
            vn.glNormal();
        }
        
        glBegin(GL_QUADS);

        SelectGLColor(c);
        glTexCoord2f(ta.u-minx,ta.v-miny);
        pa.glPlot();
        glTexCoord2f(tb.u-minx,tb.v-miny);
        pb.glPlot();
        glTexCoord2f(tc.u-minx,tc.v-miny);
        pc.glPlot();
        glTexCoord2f(td.u-minx,td.v-miny);
        pd.glPlot();
        glEnd();
}

void DrawQuad(Vec3& pa,Vec3& pb,Vec3& pc,Vec3& pd,COLORREF c,bool calc_norm) {

        if(calc_norm) {
            Vec3 va,vb,vn;
            (va=pa)-=pc;
            (vb=pb)-=pd;
            vn.cross(va,vb);
            vn.normalize();
            vn.glNormal();
        }
        
        glBegin(GL_QUADS);

        SelectGLColor(c);
        pa.glPlot();
        pb.glPlot();
        pc.glPlot();
        pd.glPlot();
        glEnd();
}

void Project::redoDisplay(Project* p) {
        RECT sbrect;
        int cX,cY,offS,offT;
//        PAINTSTRUCT ps;
        GLfloat lightloc[4]= {0.0f,0.0f,0.0f,0.0f };
        GLfloat ambient[4]=  {.3f,.3f,.3f,1.0f };
        GLfloat diffuse[4]=  {.5f,.5f,.5f,1.0f };
        HDC hdc;

//      COLORREF cr;
//      GLfloat m[16];

        hdc=GetDC(UserInt::hwndMain);

        if(UserInt::hrc==(HANDLE)0) {
            InvalidateView();
            UserInt::hrc=PrepareRenderingContext(hdc);
            wglMakeCurrent(hdc,UserInt::hrc);
            AppData::GLFontBase=glGenLists(0x60);
            SelectObject(hdc,UserInt::AxisFont);
            wglUseFontBitmaps(hdc,0x20,0x60,AppData::GLFontBase );
            glListBase(AppData::GLFontBase-0x20);
            AppData::GLObjBase=glGenLists(AD_OBJ_COUNT);
            buildOpenGLLists();
            glHint(GL_LINE_SMOOTH_HINT,GL_NICEST);
            glEnable(GL_LINE_SMOOTH);
            glEnable(GL_BLEND);
            glBlendFunc(GL_SRC_ALPHA,GL_ONE /*_MINUS_SRC_ALPHA */);
        }

        glClearColor(GetRValue(AppData::PenColor[BGPEN])/255.0,
                     GetGValue(AppData::PenColor[BGPEN])/255.0,
                     GetBValue(AppData::PenColor[BGPEN])/255.0,
                     0.0f);      // clear to background
        glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
        glDisable(GL_CULL_FACE);

        if(p) {
            GetClientRect(UserInt::hwndMain,&p->cView.Screen);
                
            GetClientRect(UserInt::hwndStatusBar,&sbrect);
            offS=sbrect.bottom-sbrect.top;
            p->cView.Screen.bottom-=offS;
            GetClientRect(UserInt::hwndToolBar,&sbrect);
            offT=sbrect.bottom-sbrect.top;
            p->cView.Screen.top+=offT;

            cX=p->cView.Screen.right-p->cView.Screen.left;
            cY=p->cView.Screen.bottom-p->cView.Screen.top;

            p->cView.sX = cX;
            p->cView.sY = cY;
            p->cView.sM = (cX < cY) ? cX : cY;
            if(p->cView.sM>2) {
// set up the view
                    glMatrixMode(GL_PROJECTION);
                    glViewport(0,offS,(GLsizei)cX,(GLsizei)cY);
                    glLoadIdentity();
                    p->cView.loadIntoOpenGL(cX,cY);

// get ready to render objects
//                        glPixelStorei(GL_UNPACK_ALIGNMENT,1);
                    glFrontFace(GL_CW);
                    glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
                    glMatrixMode(GL_MODELVIEW);
                    glLoadIdentity();
                    glEnable(GL_LIGHT0);
                    glEnable(GL_COLOR_MATERIAL) ;
                    glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
                    lightloc[X]=p->cView.Orient.x.x+p->cView.Orient.y.x-p->cView.Orient.z.x*2;
                    lightloc[Y]=p->cView.Orient.x.y+p->cView.Orient.y.y-p->cView.Orient.z.y*2;
                    lightloc[Z]=p->cView.Orient.x.z+p->cView.Orient.y.z-p->cView.Orient.z.z*2;

                    glLightfv(GL_LIGHT0,GL_POSITION,lightloc);

                    glLightfv(GL_LIGHT0,GL_AMBIENT,ambient);
                    glLightfv(GL_LIGHT0,GL_DIFFUSE,diffuse);

                    p->draw();
                    p->drawFocusRect();
            }
        } else {
            UserInt::updateStatusBar("No project loaded.",0,0);
            setWindowTitle(UserInt::hwndMain,(Project*)0);
        }
        glFlush();

        SwapBuffers(hdc);

        ReleaseDC(UserInt::hwndMain,hdc);
        hdc=NULL;

//      p->setScroll();
}

void InvalidateView(void) {
    RedrawWindow(
        UserInt::hwndMain,
        NULL,
        NULL,
        RDW_INTERNALPAINT | RDW_UPDATENOW | RDW_ALLCHILDREN
    );
}

void RightRect(RECT *r) {
    LONG swap;

    if(r->left > r->right) {
        swap=r->left;
        r->left=r->right;
        r->right=swap;
    }
    if(r->top > r->bottom) {
        swap=r->top;
        r->top=r->bottom;
        r->bottom=swap;
    }
}

void SelectGLColor(COLORREF cr) {
        glColor4f(GetRValue(cr)/255.0f,GetGValue(cr)/255.0f,GetBValue(cr)/255.0f,1.0f);
}

void SelectGLColorAlpha(COLORREF cr,float a) {
        glColor4f(GetRValue(cr)/255.0f,GetGValue(cr)/255.0f,GetBValue(cr)/255.0f,a);
}

void SelectCombinedColor(COLORREF c1,COLORREF c2) {
        glColor3f(GetRValue(c1)*GetRValue(c2)/65025.0f,GetGValue(c1)*GetGValue(c2)/65025.0f,GetBValue(c1)*GetBValue(c2)/65025.0f);
}

void Project::drawFocusRect() {
    Vec3 veca;

    if(focusRectValid==false) return;

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glDisable(GL_DEPTH_TEST);
    glEnable(GL_LINE_STIPPLE);

    for(int i=0;i<2;i++) {
        if(i) {
            glLineStipple(1,0x3333);
            SelectGLColor(AppData::PenColor[HIGHPEN]);
        } else {
            glLineStipple(1,0xcccc);
            SelectGLColor(~AppData::PenColor[HIGHPEN]);
        }

        glBegin(GL_LINES);

        cView.mapScr2XYZ((short)focusRect.left,(short)focusRect.top,veca);
        veca.glPlot();
        cView.mapScr2XYZ((short)focusRect.right,(short)focusRect.top,veca);
        veca.glPlot();

        veca.glPlot();
        cView.mapScr2XYZ((short)focusRect.right,(short)focusRect.bottom,veca);
        veca.glPlot();

        cView.mapScr2XYZ((short)focusRect.left,(short)focusRect.top,veca);
        veca.glPlot();
        cView.mapScr2XYZ((short)focusRect.left,(short)focusRect.bottom,veca);
        veca.glPlot();

        veca.glPlot();
        cView.mapScr2XYZ((short)focusRect.right,(short)focusRect.bottom,veca);
        veca.glPlot();

        glEnd();
    }

    glDisable(GL_LINE_STIPPLE);
    glEnable(GL_DEPTH_TEST);
}

#if false
void Project::swapHandedness() {
    float s;
        s=cView.Orient[XY];
        cView.Orient[XY]=cView.Orient[XZ];
        cView.Orient[XZ]=s;

        s=cView.Orient[XY];
        cView.Orient[XY]=cView.Orient[XZ];
        cView.Orient[XZ]=s;

        s=cView.Orient[XY];
        cView.Orient[XY]=cView.Orient[XZ];
        cView.Orient[XZ]=s;

        cView.Flags^=VF_HANDMASK;
        InvalidateView();
    }
}
#endif
