#include <windows.h>
#include <gl/glu.h>
#include "vert.h"
#include "patch.h"
#include "edge.h"
#include "face.h"
#include "funcs.h"
//#include "subdata.h"

const extern COLORREF   PenColor[];

EdgeData::EdgeData()
    : sharpLevel(0.0f)
{ }

Edge::Edge() {
// ensure that there is an initial history entry
    EdgeEntry* ee=new EdgeEntry;
    history.addTail(ee);
}

Edge::~Edge() {
    data.VertS->remEdge(this);
    data.VertE->remEdge(this);
    destroyHistory();
}

void Edge::calcSub1() {
    (subLevel1=data.VertS->data.Spot)+=data.VertE->data.Spot;
    if(sharpType()) {
        subLevel1*=.5;
    } else {
        ((subLevel1+=FaceL->getSub1())+=FaceR->getSub1())*=.25;
    }
    subLevelf.set(0);
}

void Edge::calcSubf() {
    if(sharpType()) {
        ((((subLevelf=subLevel1)*=4.0f)+=data.VertS->getSub1())+=data.VertE->getSub1())*=.1666667f;
    } else {
        subLevelf*=.25;
    }
}
// This function checks to ensure that the two faces on which it borders
// are wound in the same direction.  It only does this if one face has
// been validated and the other has not.
bool Edge::checkWindings() {
        Face* fp;   // face to be processed
        Face* fo;   // other face
        Face* swap;
        Vert* v;
        bool flip;

// sit #        FaceL   FaceR   FaceL->TF_NEW   FaceR->TF_NEW
//  0       NULL    NULL    DC              DC
//  1       !NULL   NULL    0               DC
//  2       !NULL   NULL    1               DC
//  3       NULL    !NULL   DC              0
//  4       NULL    !NULL   DC              1
//  5       !NULL   !NULL   0               0
//  6       !NULL   !NULL   0               1
//  7       !NULL   !NULL   1               0
//  8       !NULL   !NULL   1               1

// 0 1 2 3 4 5 6 7 8
// this shouldn't happen, but in case it does...
        if(FaceL==NULL || FaceR==NULL) {
                // sit 0-4: Edge has <2 bordering faces.  Mark it processed so that we won't be
                // coming back here again,
                flags|=TF_NEW;
                // but don't force another pass through the edge checks
                return false;
        }

// - - - - - 5 6 7 8
        fp=FaceL;
        fo=FaceR;

        if((fp->flags&TF_NEW)==0) { // sits 5 and 6
                // if both faces are unprocessed, return without further action
                if((fo->flags&TF_NEW)==0) return false; // sit 5 gone now
// - - - - - - 6 7 8
// fo is processed, fp is unprocessed (sit 6)
// swap pointers so that fp is the processed face
                swap=fo;fo=fp;fp=swap;
        }

// - - - - - - - 7 8
// sit #        FaceL   FaceR   FaceL->TF_NEW   FaceR->TF_NEW
//  7       !NULL   !NULL   1               0
//  8       !NULL   !NULL   1               1

// one face (fp) that is processed
// one face (fo) that could be either processed or unprocessed

// if both faces are processed, the windings are either correct or cannot be
// fixed (moebius situation); mark edge processed and do not force a repeat

        if(fo->flags&TF_NEW) {
                flags|=TF_NEW;
                return false;
        }

// ensure the edge matches the processed face
// fp's winding is right; therefore, the edge should match it
        flip=false;
        for(int i=0;i<fp->data.sides;i++)
        if(fp->edges[i]==this && fp->data.verts[i]!=data.VertS) flip=true;

        if(flip) { // but if it doesn't, then swap them
                v=data.VertS;
                data.VertS=data.VertE;
                data.VertE=v;
        }
        flags|=TF_NEW;  // edge is now processed

// check winding of other face. This winding should be in opposition to the edge
// winding, so therefore if it matches, reverse the face.
        flip=false;

        for(int i=0;i<fo->data.sides;i++)
        if(fo->edges[i]==this && fo->data.verts[i]==data.VertS) flip=true;

        if(flip) fo->reverseWinding();
        fo->flags|=TF_NEW;  // other face is now processed
        return true; // force repeat of edge processing
}

Vert*   Edge::otherEnd(Vert* ov) {
    if(ov==data.VertS) return data.VertE;
    return data.VertS;
}

//void Edge::prepSub() {
//    subLevel2.setup(8);
//    subChildS.setup(4);
//    subChildE.setup(4);
//}

void Edge::replaceFace(Face* of, Face* nf) {
    if(FaceL==of) { FaceL=nf; }
    if(FaceR==of) { FaceR=nf; }
    flags&=~(EF_NO_LEFT|EF_NO_RIGHT);
    if(FaceL==NULL) flags|=EF_NO_LEFT;
    if(FaceR==NULL) flags|=EF_NO_RIGHT;
}

short Edge::sharpType() {
    switch(flags&EF_BORDERMASK) {
        case 0:
            return (data.sharpLevel==0.0f)?0:1;
        case EF_NO_LEFT:
        case EF_NO_RIGHT:
            return -2;
        default:    // orphan edge
            return -3;
    }
}

//short Edge::sharpType2() {
//    switch(flags&EF_BORDERMASK) {
//        case 0:
//            return (data.sharpLevel<=1.0f)?0:1;
//        case EF_NO_LEFT:
//        case EF_NO_RIGHT:
//            return -2;
//        default:    // orphan edge
//            return -3;
//    }
//}

void Edge::draw(int subd) {
    switch(sharpType()) {
        case 0: // smooth interior
            glLineStipple(1,0x5555);
            glEnable(GL_LINE_STIPPLE);
            break;
        case -2: // border
            glDisable(GL_LINE_STIPPLE);
            break;
        case -3: // orphan
            glLineStipple(1,0x1111);
            glEnable(GL_LINE_STIPPLE);
            break;
        default: // sharp interior
            glLineStipple(1,0x7777);
            glEnable(GL_LINE_STIPPLE);
        break;
    }

    trace(subd);
}

void Edge::trace(int subd) {
    Vec3 coeff[4];
    if(subd==0) {
        glBegin(GL_LINES);
        data.VertS->data.Spot.glPlot();
        data.VertE->data.Spot.glPlot();
        glEnd();
    } else {
        glEnable(GL_MAP1_VERTEX_3);
//        glMapGrid1f(3,0.0f,1.0f);

        coeff[0]=data.VertS->getSubf();
        coeff[3]=getSubf();
        if(sharpType()) {
            (((coeff[1]=data.VertS->getSub1())*=2.0f)+=getSub1())*=.3333333f;
            (((coeff[2]=getSub1())*=2.0f)+=data.VertS->getSub1())*=.3333333f;
        } else {
            ((coeff[1]=patchSL->v_v)+=patchSR->v_v)*=0.5f;
            ((coeff[2]=patchSL->v_n)+=patchSR->v_p)*=0.5f;
        }
        glMap1f(GL_MAP1_VERTEX_3,0.0f,1.0f,3,4,(GLfloat*)coeff);
        glMapGrid1f(subd,0.0f,1.0f);
        glEvalMesh1(GL_LINE,0,subd);


        coeff[0]=data.VertE->getSubf();
//        coeff[3]=getSubf();
        if(sharpType()) {
          (((coeff[1]=data.VertE->getSub1())*=2.0f)+=getSub1())*=.3333333f;
          (((coeff[2]=getSub1())*=2.0f)+=data.VertE->getSub1())*=.3333333f;
        } else {
            ((coeff[1]=patchEL->v_v)+=patchER->v_v)*=0.5f;
            
            ((coeff[2]=patchEL->v_p)+=patchER->v_n)*=0.5f;
        }

        glMap1f(GL_MAP1_VERTEX_3,0.0f,1.0f,3,4,(GLfloat*)coeff);
        glMapGrid1f(subd,1.0f,0.0f);
        glEvalMesh1(GL_LINE,0,subd);

        glDisable(GL_MAP1_VERTEX_3);
    }
}

// undo/redo functions

bool Edge::takeSnap(INDEX newVersion) {
        EdgeEntry*      vh;
        EdgeEntry*      lvh;
//      INDEX   hh;
        bool make;
        TYPE    newType;

        // clip all history equal to or greater than new version
        clipHistory(newVersion);

        // most recent entry should be last
        lvh=(EdgeEntry*)(history.last());

/* behavior logic:
for takeSnap():

last entry      current object
NONE            invalid                 flag as TF_KILLME
et_nul          invalid                 flag as TF_KILLME
et_chg          invalid                 archive as et_del
et_del          invalid                 do nothing

NONE            valid                   archive as et_chg
et_nul          valid                   archive as et_chg
et_chg          valid                   compare and archive as et_chg
et_del          valid                   shouldn't happen
*/

        make=false; // default is don't make new entry
        newType=ET_CHG; // most common type is change

        if(lvh==NULL) {
                if(flags&TF_INVALID) {
                        flags|=TF_KILLME;
                } else {
                        make=true;
                }
        } else {
                switch(lvh->type) {
                        case ET_NUL:
                                if(flags&TF_INVALID) {
                                        flags|=TF_KILLME;
                                } else {
                                        make=true;
                                        newType=ET_CHG;
                                }
                                break;
                        case ET_CHG:
                                if(flags&TF_INVALID) {
                                        make=true;
                                        newType=ET_DEL;
                                } else {
                                        // make it if latest is different from current
                                        if (memcmp(&lvh->data,&data,sizeof(EdgeData))) make=true;
                                }
                                break;
// last case is either do nothing or shouldn't happen
//                      case ET_DEL:
//                              break;
                }
        }

        if(make) {
                vh=new EdgeEntry;
                history.addTail(vh);
                vh->version=newVersion;
                vh->type=newType;
                vh->data=data;
        }
        return make;
}

void Edge::clipHistory(INDEX version) {
        EdgeEntry *ee;
        EdgeEntry *nee;

        ee=(EdgeEntry*)(history.first());
        while(ee!=NULL) {
                nee=(EdgeEntry*)(ee->next());
                if(ee->version>=version) {
                        ee->remove();
                        delete ee;
                }
                ee=nee;
        }
}

void Edge::destroyHistory() {
        EdgeEntry*  ee;

        while(history.last()) {
                ee=(EdgeEntry*)(history.last());
                ee->remove();
                delete ee;
        }
}

void Edge::restoreVersion(INDEX getVersion) {
        EdgeEntry*  ee;

        ee=(EdgeEntry*)(history.last());

// there is always an entry with version 0, so this should never back up
// such that ve==0
        while(ee->version > getVersion)
        ee=(EdgeEntry*)(ee->prev());

        if(ee->type==ET_CHG) {
                data=ee->data;
                flags&=~TF_INVALID;
        } else {
                flags|=TF_INVALID;
        }
}

//Edge* Model::restoreEdge(Delta*d) {
//      Edge *e;

//      e=(Edge*)createTailObject(&Edges,sizeof(Edge));
//      if(e) finishEdgeByDelta(e,d);
//      return e;
//}

//Delta* Snap::createEdgeDelta(Edge* e,TYPE type) {
//      Delta *d;
//      LSMWORD extra[2];

//      d=createDelta(type,EDGEDELT,0);
//      if(d==NULL) {
//              ReportSystemError("Creation of Delta object failed.");
//              return d;
//      }
//      d->index=e->index;

//      if(type==DT_DEL) return d;

//      extra[0].l=e->VertS->index;
//      extra[1].l=e->VertE->index;

//      d->extend(2,extra);

//      if(e->flags&EF_SHARP) {
//              extra[0].l=SHARPFLAG;
//              d->extend(1,extra);
//      }

//      if(e->layer->index!=(INDEX)-1) {
//              extra[0].l=LAYERBLOCK;
//              extra[1].l=e->layer->index;
//              d->extend(2,extra);
//      }

//      return d;
//}
