#include "userint.h"
#include "fhand.h"
#include "vert.h"
#include "patch.h"
#include "edge.h"
#include "face.h"
#include "join.h"
#include "norm.h"
#include "uvpt.h"
#include "funcs.h"
#include "calc.h"
#include "appdata.h"

FaceData::FaceData()
        : sides(0)
        , Texture((Tint*)0)
{ }

FaceData::FaceData(int n)
        : sides(n)
        , Texture((Tint*)0) {
        verts=new Vert*[n+1];
}

FaceData::~FaceData() { }

bool FaceData::operator!=(FaceData&od) {
        if(sides!=od.sides) return true;
        if(Texture!=od.Texture) return true;
//      if(Owner!=od.Owner) return true;

        for(int i=0;i<=sides;i++) {
                if(verts[i]!=od.verts[i]) return true;
        }

        if(uvpts) {
                if(od.uvpts) {
                        for(int i=0;i<sides;i++) {
                                if(uvpts[i]!=od.uvpts[i]) return true;
                        }
                } else {
                        return true;
                }
        } else { // uvpts==NULL
                if(od.uvpts!=NULL) return true;
        }

        return false;
}

void FaceData::operator=(FaceData&od) {
        sides=od.sides;
        verts=(Vert**)0;
        if(od.verts) {
                verts = new Vert*[sides+1];
                for(int i=0;i<=sides;i++)
                verts[i]=od.verts[i];
        }
        uvpts=(UVPt**)0;
        if(od.uvpts) {
                uvpts = new UVPt*[sides];
                for(int i=0;i<sides;i++)
                uvpts[i]=od.uvpts[i];
        }
        Texture=od.Texture;
//      Owner=od.Owner;
//      flags=od.flags;
}

Face::Face()
        : Owner((Bone*)0) {
// ensure that there is an initial history entry
        FaceEntry* fe=new FaceEntry;
        history.addTail(fe);
}

Face::Face(int s_count)
        : data(s_count)
        , Owner((Bone*)0) {
// ensure that there is an initial history entry
        FaceEntry* fe=new FaceEntry;
        history.addTail(fe);
}

Face::~Face() {
    destroyHistory();
}

void Face::addCorner(Vert* v,UVPt* u,Norm* n) {
    if(data.sides==0) { // first corner to be added
        data.verts=new Vert*[2];
        data.verts[0]=
        data.verts[1]=v;
        if(u) {
            data.uvpts=new UVPt*[1];
            data.uvpts[0]=u;
        }
        if(n) {
            norms=new Norm*[1];
            norms[0]=n;
        }
        data.sides=1;
    } else { // there should be one of everything here
        Vert** nva=new Vert*[data.sides+2];
        for(int i=0;i<data.sides;i++)
        nva[i]=data.verts[i];
        nva[data.sides]=v;
        nva[data.sides+1]=nva[0];
//        delete []data.verts;
        data.verts=nva;
        if(u) {
            if(data.uvpts) { // added only if all previous corners had one
                UVPt** nua = new UVPt*[data.sides+1];
                for(int i=0;i<data.sides;i++)
                nua[i]=data.uvpts[i];
                nua[data.sides]=u;
//                delete []data.uvpts;
                data.uvpts=nua;
            }
        } else {
            if(data.uvpts) { // delete previous corner UVPts if this one is lacking
//                delete []data.uvpts;
                data.uvpts=(UVPt**)0;
            }
        }
        if(n) {
            if(norms) { // added only if all previous corners had normals
                Norm** nna = new Norm*[data.sides+1];
                for(int i=0;i<data.sides;i++)
                nna[i]=norms[i];
                nna[data.sides]=n;
                norms=nna;
            }
        } else {
            norms=(Norm**)0;
        }
        data.sides++;
        edges=0;
    }
}

float Face::calcArea() {
    Vec3 norm;

    calcNormal(norm);
    return norm.length()*.5;
}

void Face::calcCenter(Vec3& cent) {
//    if(cent==NULL) return;

    cent.set(0);
    for(int i=0;i<data.sides;i++) {
        cent+=data.verts[i]->data.Spot;
    }

    cent/=data.sides;
}

void Face::calcMapCenter(Vec2& cent) {
//        if(cent==NULL) return;
    if(data.uvpts==NULL) return;

    cent.u=cent.v=0;
    for(int i=0;i<data.sides;i++) {
        cent.u+=data.uvpts[i]->data.Spot.u;
        cent.v+=data.uvpts[i]->data.Spot.v;
    }

    cent.u/=data.sides;
    cent.v/=data.sides;
}

void Face::calcNormal(Vec3& norm) {
    Vec3 temp;

    norm.set(0);
    for(int i=0;i<data.sides;i++) {
        temp.cross(data.verts[i]->data.Spot,data.verts[i+1]->data.Spot);
        norm+=temp;
    }
}

void Face::calcSub1() {
    subLevel1=data.verts[0]->data.Spot;
    for(int i=1;i<data.sides;i++) subLevel1+=data.verts[i]->data.Spot;

    subLevel1*=1.0f/((float)data.sides);

    if(data.uvpts) {
        uSubLevel1=data.uvpts[0]->data.Spot;
        for(int i=1;i<data.sides;i++) uSubLevel1+=data.uvpts[i]->data.Spot;

        uSubLevel1*=1.0f/((float)data.sides);
    }
    subLevelf.set(0);
}

//void Face::calcSubEdges() {
//    for(int i=0;i<data.sides;i++)
//    subEdges[i].recalc();

//    if(data.uvpts) {
//        for(int i=0;i<data.sides;i++)
//        uSubEdges[i].recalc();
//    }
//}

void Face::destroyNorms() {
    norms=0;
}

// the whole purpose of this is to modify the face without making a call to
// validate() necessary.
// e is the edge to be split
// v is the new Vert to be inserted
// u is the new UVPt to be inserted, if any
void Face::insertCorner(Edge* e,Vert* v, UVPt* u) {
    int i,j,jt /*,jn*/;
    Vert**  nva;
    Edge**  nea;
    UVPt**  nua;
    Join**  nja;

//    if(data.uvpts!=NULL && u==NULL) SetWindowText(AppData::hwndMain,"NULL UVPt passed.");
    
    Edge*   es=v->findEdge(e->data.VertS);
    Edge*   ee=v->findEdge(e->data.VertE);

    jt=-1;
    for(i=0;i<data.sides;i++)
    if(e==edges[i]) jt=i;

    if(jt==-1) {
        UserInt::informUser("Could not find split edge in face.");
        return;
    }

//    jn=jt+1;
//    if(jn==data.sides) jn=0;

// build new array of Vert pointers
    nva=new Vert*[data.sides+2];
    for(i=0,j=0;i<=data.sides;i++) {
        nva[j++]=data.verts[i];
        if(i==jt) nva[j++]=v;
    }

    nea=new Edge*[data.sides+1];
    for(i=0,j=0;i<data.sides;i++) {
        if(i==jt) {
            if(e->data.VertS==data.verts[i]) {
                nea[j++]=es;
                nea[j]=ee;
            } else {
                nea[j++]=ee;
                nea[j]=es;
            }
        } else {
            nea[j]=edges[i];
        }
        j++;
    }

    if(data.uvpts) {
        nua=new UVPt*[data.sides+1];
        for(i=0,j=0;i<data.sides;i++) {
            nua[j++]=data.uvpts[i];
            if(i==jt) {
                nua[j++]=u;
            }
        }
//        delete []data.uvpts;
        data.uvpts=nua;

        Join* jc=joins[jt];
        Join* js=jc->UVPtS->findJoin(u);
        Join* je=jc->UVPtE->findJoin(u);

        nja=new Join*[data.sides+1];

        for(i=0,j=0;i<data.sides;i++) {
            if(i==jt) {
                if(jc->UVPtS==data.uvpts[i]) {
                    nja[j++]=js;
                    nja[j]=je;
                } else {
                    nja[j++]=je;
                    nja[j]=js;
                }
            } else {
                nja[j]=joins[i];
            }
            j++;
        }
        joins=nja;
    }

    edges=nea;

//    delete []data.verts;
    data.verts=nva;

    data.sides++;

    if(es->FaceL) {
        es->FaceR=this;
    } else {
        es->FaceL=this;
    }

    if(ee->FaceL) {
        ee->FaceR=this;
    } else {
        ee->FaceL=this;
    }

    norms=(Norm**)0;
}

void Face::mergeNorms() {
    Face* of;
    for(int i=0;i<data.sides;i++)
    if(edges[i]->sharpType()==0) {   // smooth edge
        of=otherSide(edges[i]);
        for(int j=0;j < of->data.sides;j++) {
            if(of->data.verts[j] == data.verts[i]) {
                norms[i]->prefer = of->norms[j];
            }
        }
    }
}

Face* Face::otherSide(Edge* e) {
    if(e==NULL) return NULL;

    if(this==e->FaceL) {
        return e->FaceR;
    } else {
        return e->FaceL;
    }
}

//void Face::prepSub() {
//    subLevel1.setup(data.sides);
//    subLevel2.setup(data.sides*2+1);

//    subEdges=new WSum3[data.sides];
//    for(int i=0;i<data.sides;i++) subEdges[i].setup(4);

//    if(data.uvpts) {
//        uSubLevel1.setup(data.sides);
//        uSubLevel2.setup(data.sides*2+1);

//        uSubEdges=new WSum2[data.sides];
//        for(int i=0;i<data.sides;i++) uSubEdges[i].setup(4);
//    }
//}

void Face::removeEdgeXRef() {
    edges=0;
}

void Face::removeMapping() {
    data.uvpts=(UVPt**)0;
}

void Face::replaceEdge(Edge* e,Edge* ne) {
    if(!edges) return;

    for(int i=0;i<data.sides;i++)
    if(edges[i]==e)
    edges[i]=ne;
}

void Face::reverseWinding() {
        Vert*   v;
        Edge*   e;
        UVPt*   u;

        for(int i=1;i<(data.sides-i);i++) {
                int j=data.sides-i;

                v=data.verts[i];
                data.verts[i]=data.verts[j];
                data.verts[j]=v;

                if(data.uvpts) {
                        u=data.uvpts[i];
                        data.uvpts[i]=data.uvpts[j];
                        data.uvpts[j]=u;
                }

                j=j-1;

                if(i!=j) {
                        e=edges[i];
                        edges[i]=edges[j];
                        edges[j]=e;
                }
        }

//      for(int i=0;i<data.sides;i++)
//      if(data.verts[i]==NULL) ReportSystemError("Blow time!");

        e=edges[0];
        edges[0]=edges[data.sides-1];
        edges[data.sides-1]=e;
}

void Face::writeToOBJFile(fWriter& file) {
        file << "f";
        for(int i=0;i<data.sides;i++) {
                file << ' ';
                file << (1+(int)(data.verts[i]->index));
                if(data.uvpts) {
                        file << '/';
                        file << (1+(int)(data.uvpts[i]->index));
                }
//              WriteString(fhand,"/");
//              WriteIntString(fhand,norms[i]->index+1);
        }
        file << '\n';
}

// undo/redo functions

bool Face::takeSnap(INDEX newVersion) {
        FaceEntry*      vh;
        FaceEntry*      lvh;
//        INDEX   hh;
        bool make;
        TYPE    newType;

        // clip all history equal to or greater than new version
        clipHistory(newVersion);

        // most recent entry should be last
        lvh=(FaceEntry*)(history.last());

/* behavior logic:
for takeSnap():

last entry      current object
NONE            invalid                 flag as TF_KILLME
et_nul          invalid                 flag as TF_KILLME
et_chg          invalid                 archive as et_del
et_del          invalid                 do nothing

NONE            valid                   archive as et_chg
et_nul          valid                   archive as et_chg
et_chg          valid                   compare and archive as et_chg
et_del          valid                   shouldn't happen
*/

        make=false; // default is don't make new entry
        newType=ET_CHG; // most common type is change

        if(lvh==NULL) {
                if(flags&TF_INVALID) {
                        flags|=TF_KILLME;
                } else {
                        make=true;
                }
        } else {
                switch(lvh->type) {
                        case ET_NUL:
                                if(flags&TF_INVALID) {
                                        flags|=TF_KILLME;
                                } else {
                                        make=true;
                                        newType=ET_CHG;
                                }
                                break;
                        case ET_CHG:
                                if(flags&TF_INVALID) {
                                        make=true;
                                        newType=ET_DEL;
                                } else {
                                        // make it if latest is different from current
                                        if (lvh->data!=data) make=true;
                                }
                                break;
// last case is either do nothing or shouldn't happen
//                      case ET_DEL:
//                              break;
                }
        }

        if(make) {
                vh=new FaceEntry;
                history.addTail(vh);
                vh->version=newVersion;
                vh->type=newType;
                vh->data=data;
        }

        return make;
}

void Face::clipHistory(INDEX version) {
        FaceEntry *fe;
        FaceEntry *nfe;

        fe=(FaceEntry*)(history.first());
        while(fe!=NULL) {
                nfe=(FaceEntry*)(fe->next());
                if(fe->version>=version) {
                        fe->remove();
                        delete fe;
                }
                fe=nfe;
        }
}

void Face::destroyHistory() {
        FaceEntry*  fe;

        while(history.last()) {
                fe=(FaceEntry*)(history.last());
                fe->remove();
                delete fe;
        }
}

void Face::restoreVersion(INDEX getVersion) {
        FaceEntry*  fe;

        fe=(FaceEntry*)(history.last());

// there is always an entry with version 0, so this should never back up
// such that ve==0
        while(fe->version > getVersion)
        fe=(FaceEntry*)(fe->prev());

        if(fe->type==ET_CHG) {
                data=fe->data;
                flags&=~TF_INVALID;
        } else {
                flags|=TF_INVALID;
        }
}


