#ifndef FACE_H
#define FACE_H

#include "link.h"
#include "list.h"
#include "norm.h"
#include "pointer.h"
#include "obvec.h"

struct FaceData {
    short   sides;
    Array<class Vert*>  verts;
    Array<class UVPt*>  uvpts;
    class Tint*     Texture;
    FaceData();
    FaceData(int);
    ~FaceData();
    bool    operator!=(FaceData&);
    void    operator=(FaceData&);
};

class FaceEntry : public Entry {
    public:
        FaceData    data;
};

class Face : public Thing {
    protected:
        Vec3   subLevel1;
        Vec3   subLevelf;
//        Array<WSum3>    subEdges;
        Vec2   uSubLevel1;
//        WSum2   uSubLevel2;
//        Array<WSum2>    uSubEdges;
    public:
        FaceData    data;
        eList  history;
        Array<class Edge*>  edges;
        Array<class Join*>  joins;
        Array<class Norm*>  norms;
        Ref<Bone>   Owner;

        Face();
        Face(int);
        ~Face();

// in face.cpp
        void    addCorner(Vert*,UVPt*,Norm*);

//      void    extendMapping(Edge*,float*);
        float   calcArea();
        void    calcCenter(struct Vec3&);
        void    calcMapCenter(struct Vec2&);
        void    calcNormal(struct Vec3&);
        void    destroyNorms();
        void    insertCorner(Edge*,Vert*,UVPt*);
        void    mergeNorms();
        Face*   otherSide(class Edge*);
        void    pullUVData(class Edge*);
        void    removeEdgeXRef();
        void    removeMapping();    // removes uv-mapping from face
        void    replaceEdge(Edge*,Edge*);
        void    reverseWinding();
        void    validate();
        void    writeToFile(class fWriter&);

// subdivision
//    void    prepSub();

//    void    weightSub1(Vec3& v,float w=1.0f) { subLevel1.add(v,w); }
//    void    weightSub2(Vec3& v,float w=1.0f) { subLevel2.add(v,w); }
//    void    weightSubEdge(int i,Vec3& v,float w=1.0f) { subEdges[i].add(v,w); }
//    void    weightUSub1(Vec2& v,float w=1.0f) { uSubLevel1.add(v,w); }
//    void    weightUSub2(Vec2& v,float w=1.0f) { uSubLevel2.add(v,w); }
//    void    weightUSubEdge(int i,Vec2& v,float w=1.0f) { uSubEdges[i].add(v,w); }

    void    calcSub1();
//    void    calcSub2() { subLevel2.recalc(); uSubLevel2.recalc(); }
//    void    calcSubEdges();

    Vec3&   getSub1() { return subLevel1; }
    Vec3&   getSubf() { return subLevelf; }
    Vec2&   getUSub1() { return uSubLevel1; }
//    Vec2&   getUSub2() { return uSubLevel2.result(); }

// undo/redo
        void    clipHistory(INDEX);
        bool    takeSnap(INDEX);
        void    restoreVersion(INDEX);
        void    destroyHistory();

// in obj_file.cpp
        void    writeToOBJFile(class fWriter&);

    friend class Patch;
};

// flags for Face.flags
// #define      TF_MARKED       1       // marked by the user for some reason
// #define      TF_DRAGGED      2       // needs to be redrawn with mouse move
// #define      TF_NEW          4       // created in current operation
// #define      TF_INVALID      8       // not valid due to delete/redo/undo operation
// #define      TF_KILLME   16  // to be destroyed during next sweep
// #define      TF_HIDDEN   32  // don't display this

#endif
