// fhand.cpp -- implements the file handling classes

#include <windows.h>
#include <stdio.h>
#include "fhand.h"
#include "mystr.h"
#include "calc.h"

fHandler::fHandler() : fhand(INVALID_HANDLE_VALUE) { }

fHandler::~fHandler() {
        close();
}

void fHandler::close() {
        if(fhand!=INVALID_HANDLE_VALUE) {
                CloseHandle(fhand);
                fhand=INVALID_HANDLE_VALUE;
        }
}

void fWriter::open(char* filePath) {
        close();

        if(filePath)
        fhand=CreateFile(filePath,GENERIC_WRITE,0,NULL,CREATE_ALWAYS,
                FILE_ATTRIBUTE_NORMAL,NULL);
}

fWriter& fWriter::operator<<(const char& val) {
        DWORD count;
        if(fhand!=INVALID_HANDLE_VALUE) {
                WriteFile(fhand,&val,sizeof(char),&count,NULL);
                if(count<sizeof(char)) close();
        }
        return *this;
}

fWriter& fWriter::operator<<(const char* str) {
        DWORD count;
        DWORD len;

        for(len=0;str[len];len++);

        if(len)
        if(fhand!=INVALID_HANDLE_VALUE) {
                WriteFile(fhand,str,len,&count,NULL);
                if(count<len) close();
        }
        return *this;
}

fWriter& fWriter::operator<<(const int val) {
        char buff[20];

        sprintf(buff,"%d",val);
        return (*this)<<buff;
}

fWriter& fWriter::operator<<(const long val) {
        char buff[20];

        sprintf(buff,"%d",val);
        return (*this)<<buff;
}

fWriter& fWriter::operator<<(const float& val) {
        char buff[20];

        sprintf(buff,"%g",(double)val);
        return (*this)<<buff;
}

fWriter& fWriter::writeVecString(const struct Vec3& vec) {
    *this << '<' << vec.x << ',' << vec.y << ',' << vec.z << '>';
    return *this;
}

//void fWriter::writeUVecString(const float* vec) {
//    *this << '<' << vec[0] << ',' << vec[1] << '>';
//}

fWriter& fWriter::operator<<(const struct Vec2& v) {
    *this << '<' << v.u << ',' << v.v << '>';
    return *this;
}

void fWriter::writeHexString(const int val) {
        char buff[20];

        sprintf(buff,"%x",val);
        (*this) << buff;
}

fReader::fReader() : eof_yet(true) { }

fReader::operator bool() const {
        if(fhand==INVALID_HANDLE_VALUE) return false;
        return !eof_yet;
}

void fReader::open(char* filePath) {
        close();
        fhand=CreateFile(filePath,GENERIC_READ,0,NULL,OPEN_EXISTING,
                        FILE_FLAG_SEQUENTIAL_SCAN,NULL);
        if(fhand!=INVALID_HANDLE_VALUE) eof_yet=false;
}

fReader& fReader::operator>>(char& val) {
        DWORD count;
        char c;

        if(fhand!=INVALID_HANDLE_VALUE) {
            ReadFile(fhand,&c,sizeof(char),&count,NULL);
                if(count<sizeof(char)) eof_yet=true;
        }
        val=c;
        return *this;
}

fReader& fReader::operator>>(short& val) {
        DWORD count;
        short s;

        if(fhand!=INVALID_HANDLE_VALUE) {
            ReadFile(fhand,&s,sizeof(short),&count,NULL);
            if(count<sizeof(short)) eof_yet=true;
        }
        val=s;
        return *this;
}

fReader& fReader::operator>>(long& val) {
        DWORD count;
        long l;

        if(fhand!=INVALID_HANDLE_VALUE) {
            ReadFile(fhand,&l,sizeof(long),&count,NULL);
                if(count<sizeof(long)) eof_yet=true;
        }
        val=l;
        return *this;
}

fReader& fReader::operator>>(float& val) {
        DWORD count;
        float f;

        if(fhand!=INVALID_HANDLE_VALUE) {
            ReadFile(fhand,&f,sizeof(float),&count,NULL);
                if(count<sizeof(float)) eof_yet=true;
        }
        val=f;
        
        return *this;
}

void fReader::getLine(Mystr& buff) {
        buff.own((char*)0);
        char c;

        do {
                c='\012';
                if(this) *this >> c;
                if(*this)
                switch(c) {
                        case '\012':
                                return;
                        case '\015':
                                if(peek()=='\012') *this+=1;
                                return;
                        default:
                                buff+=c;
                }
        } while(*this);
}

char fReader::peek() {
        char c;

        *this >> c;
        *this+=-1;
        return c;
}

fReader& fReader::operator+=(const int n) {
        if(*this)   // if this is valid
        if(SetFilePointer(fhand,n,NULL,FILE_CURRENT)==INVALID_SET_FILE_POINTER)
        if(GetLastError()!=NO_ERROR) // if error from this operation
        close();    // we close the file

        return *this;
}

bool fReader::getOBJLine(Mystr& buff) {

        getLine(buff);
        int l=buff.len();

        if(buff.last()=='\\') {
                Mystr add;
                getOBJLine(add);
        buff.c_string()[l-1]='\0';
                buff+=add;
        }
        return ((l==0)?false:true);
}

fReader& fReader::read(void* buffer, int len) {
        DWORD count;

        if(buffer)
        if(*this) {
            ReadFile(fhand,buffer,len,&count,NULL);
                if(count<len) eof_yet=true;
        }

        return *this;
}

void fReader::readPaddedString(char *s,int l) {
        int ptr;
        bool drop;

        ptr=0;
        drop=false;
        do {
                read(s+ptr,4);
                if(s[ptr++]==0) drop=true;
                if(s[ptr++]==0) drop=true;
                if(s[ptr++]==0) drop=true;
                if(s[ptr++]==0) drop=true;
    } while(drop==false && ptr<l);
}
