// fhand.h -- my file handler, less bloated than standard streaming

#ifndef FHAND_H
#define FHAND_H

#include <windows.h>

// the basic file-handling class
class fHandler {
    protected:
        HANDLE  fhand;
    public:
        fHandler();
        ~fHandler();
        void close();
};

// text file reader
class fReader : public fHandler {
        protected:
                bool eof_yet;
        public:
                fReader();
                void open(char*);
                operator bool() const;  // reports whether stream is valid
                fReader& operator>>(char&); // reads one character
                fReader& operator>>(short&);    // reads one short integer (binary)
                fReader& operator>>(int&);  // reads one int (binary)
                fReader& operator>>(long&); // reads one long (binary)
                fReader& read(void*,int);   // reads into a buffer (binary)
                fReader& operator>>(float&); // reads one single-precision variable (binary)
                fReader& operator+=(const int);  // skips forward by this amount
                bool    getOBJLine(class Mystr&);    // reads an OBJ file line into Mystr
                void    getLine(class Mystr&);  // reads until the next EOL (or file end)
                char    peek(); // get the next char without advancing the file pointer
                void    readPaddedString(char*,int); // reads a padded string
};

// text file writer
class fWriter : public fHandler {
        public:
                void open(char*);
                operator bool() const;  // reports whether file handle is valid
                fWriter& operator << (const char&); // writes one character in text format
                fWriter& operator << (const int);    // writes one integer
                fWriter& operator << (const long);    // writes one integer
                fWriter& operator << (const float&); // writes one single-precision variable
                fWriter& operator << (const double&); // writes one double-precision float
                fWriter& operator << (const char*); // writes one string
                fWriter& operator << (const struct Vec2&); // writes mapping vector
                void writeHexString(const int);
                fWriter& writeVecString(const struct Vec3&);
                fWriter& writeUVecString(const struct Vec2&);
};

inline fWriter::operator bool() const {
        return fhand!=INVALID_HANDLE_VALUE;
}

#endif
