#include "userint.h"
//#include "appdata.h"
#include "model.h"
#include "scene.h"
#include "fhand.h"
#include "lsmres.h"
#include "funcs.h"
#include "userint.h"

/* saving logic:
  The flag PF_NOTDEFAULTFILENAME is cleared when the project is created.
  The flag is set when the project is saved, or when it is loaded.

  If the user specifies "save as" or if the project has never been saved,
  a file is requested through the dialog.
*/

// extern HWND UserInt::hwndMain;

char FullPathBuffer[256];
char TitleBuffer[40];
char GraphicFilter[]={ "Windows bitmap files (*.bmp)\0*.bmp\0" };
char MeshFilter[]={ "Alias/Wavefront file\0*.obj\0" };
char TextureFilter[]={ "Alias/Wavefront MTL file\0*.mtl\0" };
//char SkeletonFilter[]={ "LionSnake skeleton file\0*.bon\0" };
char ImportFilter[]={ "Alias/Wavefront OBJ file (*.obj)\0*.obj\0" };
char ProjectFilter[]={ "Lionsnake model (*.lsm)\0*.lsm\0" };
char ModelFilter[]={ "LionSnake model (*.lsm)\0*.lsm\0" };
char ModelExportFilter[]= {
    "POV-Ray posable include file (*.inc)\0*.inc\0Alias/Wavefront OBJ file (*.obj)\0*.obj\0"
};

char *ModelExt="obj";
char InputBuffer[64];

#define PI_MODEL 1
#define MEI_INC 1
#define MEI_OBJ 2
#define MII_OBJ 1

OPENFILENAME FileData= {
 sizeof(OPENFILENAME),  // lStructSize, never changes
 (HWND) NULL,           // hwndOwner, set to main window
 (HINSTANCE) NULL,      // hInstance, not used
 (LPCTSTR) NULL,                // lpstrFilter, indicates file type filter to use
 (LPTSTR) NULL,         // lpstrCustomFilter, not used
 (DWORD) 0,                     // nMaxCustomFilter, not used
 (DWORD) 0,                     // nFilterIndex, returns index of selected filter type
 (LPTSTR) FullPathBuffer,       // lpstrFile, full path of selected file
 (DWORD) 256,           // nMaxFile, size of buffer for full path
 (LPTSTR) TitleBuffer,  // lpstrFileTitle, buffer to receive file title
 (DWORD) 40,            // nMaxFileTitle, size of file title buffer
 (LPCTSTR) NULL,                // lpstrInitialDir, not used
 (LPCTSTR) NULL,                // lpstrTitle, for specifying dialog box window title
 (DWORD) NULL,          // Flags, set by code
 (WORD) 0,                      // nFileOffset, offset of file name within returned string;
 (WORD) 0,                      // nFileExtension, offset of extension within returned string;
 (LPCTSTR) NULL,                // lpstrDefExt, pointer to default extension;
 (LPARAM) 0,            // lCustData, probably not used;
 (LPOFNHOOKPROC) NULL,  // lpfnHook, probably not used;
 (LPCTSTR) NULL         // lpTemplateName, probably not used;
};

// following fields are used in getting file names:
// FileData.hwndOwner, set to main window
// FileData.lpstrFilter, indicates file type filter to use
// FileData.nFilterIndex, returns index of selected filter type
// FileData.lpstrTitle, for specifying dialog box window title
// FileData.Flags, set by code
// FileData.nFileOffset, offset of file name within returned string;
// FileData.nFileExtension, offset of extension within returned string;
// FileData.lpstrDefExt, pointer to default extension;

// Flags OFN_CREATEPROMPT ask to create if file does not exist
// OFN_DONTADDTORECENT prevents addition to recent documents list
// OFN_FILEMUSTEXIST prevents specification of non-existent file
// OFN_HIDEREADONLY hides the read-only check box.
// OFN_OVERWRITEPROMPT prompts user to approve overwriting if file exists
// OFN_EXTENSIONDIFFERENT user specified a different extension other than default

// <header> { <data> }

void Model::saveProject(bool getname) {
        fWriter saveHandle;

// getname is TRUE when user selects "Save As"
// PF_NOTDEFAULTNAME is set when the user has saved the file at least once
// but all that is commented out for the time being
        if(getname ||(projectFlags&PF_NOTDEFAULTNAME)==0) {
                FileData.hwndOwner=UserInt::hwndMain;
// should set the filter according to the project type
                FileData.lpstrFilter=ModelFilter;
                FileData.nFilterIndex=1;
                FileData.lpstrTitle=NULL;
                FileData.Flags=OFN_OVERWRITEPROMPT|OFN_HIDEREADONLY;
// this should also be set according to the project type
                FileData.lpstrDefExt=ModelExt;
// the save title is derived from the project name
                if(SaveTitle) delete []SaveTitle;
                SaveTitle=ConcatStrings(name,".lsm");
// and it is the default file path for saving
                strcpy(FullPathBuffer,SaveTitle);
                if(GetSaveFileNameA(&FileData)==0) return;
// copy the results to the save path
                if(SavePath) delete []SavePath;
                SavePath=CloneString(FullPathBuffer);
        } else {
// use the save path from the last save operation
                strcpy(FullPathBuffer,SavePath);
        }

// prepare model for saving
        removeUnusedUVPts();
        destroyAllHistory();
        currentVersion=(INDEX)-1;
        takeSnap();
        reindexObjects();
// write LSM file
        saveHandle.open(FullPathBuffer);
        if(saveHandle) {
        writeLSMFile(saveHandle);
        } else {
                UserInt::informUser("Couldn't create model file.");
        }

        projectFlags|=PF_NOTDEFAULTNAME;

// name project to what the user specified
        if(SaveTitle) delete []SaveTitle;
        SaveTitle=CloneString(FullPathBuffer+FileData.nFileOffset);
        name.copy(SaveTitle);
        name.cut_from('.'); // chops off the file extension
        activateProject(activeProject);
}

void Model::exportFile() {
        fWriter exportFile;

        FileData.hwndOwner=UserInt::hwndMain;
// filter should be set according to the project type
        FileData.lpstrFilter=ModelExportFilter;
        FileData.nFilterIndex=1;
        FileData.lpstrTitle="Export";
        FileData.Flags=OFN_OVERWRITEPROMPT|OFN_HIDEREADONLY;
        FileData.lpstrDefExt=NULL; // no default for exporting
        FullPathBuffer[0]=0;

        if(GetSaveFileNameA(&FileData)==0) return;

        if(FileData.nFileExtension==0) {
//              UserInt::informUser("Ending file name with period.");
                FileData.nFileExtension=(WORD)strlen(FullPathBuffer);
                if(FullPathBuffer[FileData.nFileExtension-1]=='.') {
                        FileData.nFileExtension--;
                        FullPathBuffer[FileData.nFileExtension]='\0';
                }
        }

        if(FullPathBuffer[FileData.nFileExtension]=='\0') {
                switch(FileData.nFilterIndex) {
                        case MEI_INC:
                                FullPathBuffer[FileData.nFileExtension]='.';
                                FullPathBuffer[FileData.nFileExtension+1]='i';
                                FullPathBuffer[FileData.nFileExtension+2]='n';
                                FullPathBuffer[FileData.nFileExtension+3]='c';
                                FullPathBuffer[FileData.nFileExtension+4]='\0';
                        break;
                        case MEI_OBJ:
                                FullPathBuffer[FileData.nFileExtension]='.';
                                FullPathBuffer[FileData.nFileExtension+1]='o';
                                FullPathBuffer[FileData.nFileExtension+2]='b';
                                FullPathBuffer[FileData.nFileExtension+3]='j';
                                FullPathBuffer[FileData.nFileExtension+4]='\0';
                        break;
                }
        }

        exportFile.open(FullPathBuffer);
        if(!exportFile) {
                UserInt::informUser("Couldn't create export file");
                return;
        }

        switch(FileData.nFilterIndex) {
                case MEI_INC:
                        writeINCFile(exportFile);
                break;
                case MEI_OBJ:
                        replaceExtension(FullPathBuffer,"mtl");
                        writeOBJFile(exportFile,FullPathBuffer);
                        exportFile.close();
                        exportFile.open(FullPathBuffer);
                        writeMTLFile(exportFile);
                break;
        }
}

void OpenProject(void) {
        fReader loadFile;
        Model* m;

// following fields are used in getting file names:
        FileData.hwndOwner=UserInt::hwndMain;
        FileData.lpstrFilter=ProjectFilter;
        FileData.nFilterIndex=1;
        FileData.lpstrTitle=NULL;
        FileData.Flags=OFN_FILEMUSTEXIST|OFN_HIDEREADONLY;

        if(GetOpenFileNameA(&FileData)==0) return; // user bailed

        loadFile.open(FullPathBuffer);
        if(!loadFile) {
            UserInt::informUser("Could not open file"); return;
        }

        // determine file type
        switch(FileData.nFilterIndex) {
            case PI_MODEL:
                m=Model::readLSMFile(loadFile);
                if(m!=NULL) {
// name project to match the title
                    if(m->SaveTitle) delete [](m->SaveTitle);
                    m->SaveTitle=CloneString(FullPathBuffer+FileData.nFileOffset);
                    m->name.copy(m->SaveTitle);
                    m->name.cut_from('.'); // chops off the file extension
                    m->activate();
                    SendMessage(UserInt::hwndMain,WM_PAINT,0,0);
                    m->zoomToFit();
                }
            break;
        }
}

void OpenFromCommandLine(LPSTR fname) {
    fReader file;
    Model* m;
    int length;
    char *nname,*qname;

    length=strlen(fname);
    if(length==0) {
        m=new Model;
        Project::activateFirst();
        return;
    }

    nname=new char[length+1];
    ScrubString((char *)fname,'"','\0',nname);
//      strrepl(fname,"\"",NULL,nname);
    GetFullPathNameA(nname,256,FullPathBuffer,&qname);
    strcpy(TitleBuffer,qname);

    file.open(FullPathBuffer);
    delete []nname;
    nname=NULL;
    if(!file) {
        UserInt::informUser("Could not open file");
        return;
    }

    if(strstr(fname,".lsm")!=0) {
        m=Model::readLSMFile(file);
        if(m) {
// name project to match the title
            if(m->SaveTitle) delete [](m->SaveTitle);
            m->SaveTitle=CloneString(FullPathBuffer);
            m->name.copy(m->SaveTitle);
            m->name.cut_from('.'); // chops off the file extension
            m->activate();
            SendMessage(UserInt::hwndMain,WM_PAINT,0,0);
            m->zoomToFit();
        } else {
            UserInt::informUser("Unable to create Model.");
        }
    }
}

void ImportProject(void) {
    fReader file;
    Model*  m;

// following fields are used in getting file names:
    FileData.hwndOwner=UserInt::hwndMain;
    FileData.lpstrFilter=ImportFilter;
    FileData.nFilterIndex=1;
    FileData.lpstrTitle="Import";
    FileData.Flags=OFN_FILEMUSTEXIST|OFN_HIDEREADONLY;
    FileData.lpstrDefExt=NULL; // no default for importing
    FullPathBuffer[0]=0;

    if(GetOpenFileNameA(&FileData)==0) return;

    file.open(FileData.lpstrFile);
    if(!file) {
        UserInt::informUser("Could not open import file");
        return;
    }

    // determine file type
    switch(FileData.nFilterIndex) {
        case MII_OBJ:
            m=new Model;
            if(m->SaveTitle) delete [](m->SaveTitle);
            m->SaveTitle=CloneString(FullPathBuffer);
            m->name.copy(m->SaveTitle);
            m->name.cut_from('.'); // chops off the file extension

            m->readOBJFile(file);
            m->projectFlags|=PF_NOTDEFAULTNAME;
            m->activate();
            SendMessage(UserInt::hwndMain,WM_PAINT,0,0);
            m->zoomToFit();
            break;
    }
}

void GetGraphicFileName(HWND hwnd,int item) {
        FileData.hwndOwner=UserInt::hwndMain;
        FileData.lpstrFilter=GraphicFilter;
        FileData.nFilterIndex=1;
        FileData.lpstrTitle=NULL;
        FileData.Flags=OFN_FILEMUSTEXIST|OFN_HIDEREADONLY;

        if(GetOpenFileNameA(&FileData))
        SetDlgItemText(hwnd,item,FileData.lpstrFile);
}

void GetMeshFileName(HWND hwnd,int item) {
        FileData.hwndOwner=UserInt::hwndMain;
        FileData.lpstrFilter=MeshFilter;
        FileData.nFilterIndex=1;
        FileData.lpstrTitle=NULL;
        FileData.Flags=OFN_FILEMUSTEXIST|OFN_HIDEREADONLY;

        if(GetOpenFileNameA(&FileData))
        SetDlgItemText(hwnd,item,FileData.lpstrFile);
}

void GetTextureFileName(HWND hwnd,int item) {
        FileData.hwndOwner=UserInt::hwndMain;
        FileData.lpstrFilter=TextureFilter;
        FileData.nFilterIndex=1;
        FileData.lpstrTitle=NULL;
        FileData.Flags=OFN_FILEMUSTEXIST|OFN_HIDEREADONLY;

        if(GetOpenFileNameA(&FileData))
        SetDlgItemText(hwnd,item,FileData.lpstrFile);
}

/*
void GetSkeletonFileName(HWND hwnd,int item) {
        FileData.hwndOwner=UserInt::hwndMain;
        FileData.lpstrFilter=SkeletonFilter;
        FileData.nFilterIndex=1;
        FileData.lpstrTitle=NULL;
        FileData.Flags=OFN_FILEMUSTEXIST|OFN_HIDEREADONLY;

        if(GetOpenFileNameA(&FileData))
        SetDlgItemText(hwnd,item,FileData.lpstrFile);
}
*/

// the following code was rendered obsolete by the fReader and fWriter classes
/*
void WritePaddedString(HANDLE fhand,char *s) {
        DWORD count;

        WriteFile(fhand,s,(strlen(s)+4)&0xffc,&count,NULL);
}

void ReadLine(HANDLE fhand,char* buffer,int max_length,char end_char) {
        int i;
        DWORD   count;

        for(i=0;i<max_length;i++) {
                ReadFile(fhand,buffer+i,1,&count,NULL);
                if(count==0) {
                        buffer[i]='\0';
                        EOF_yet=true;
                        return;
                }
                if(buffer[i]==end_char) {
                        buffer[i]='\0';
                        return;
                }
        }
}

float ReadFloat(HANDLE fhand) {
        DWORD count;
        float f;

        ReadFile(fhand,&f,sizeof(float),&count,NULL);
        if(count<sizeof(float)) { EOF_yet=true; return 0.0; }

        return f;
}

void WriteFloat(HANDLE fhand,float f) {
        DWORD count;

        WriteFile(fhand,&f,sizeof(float),&count,NULL);
}

long ReadLong(HANDLE fhand) {
        DWORD count;
        long l;

        ReadFile(fhand,&l,sizeof(long),&count,NULL);
        if(count<sizeof(long)) { EOF_yet=true; return 0; }

        return l;
}

void WriteLong(HANDLE fhand,long l) {
        DWORD count;

        WriteFile(fhand,&l,sizeof(long),&count,NULL);
}

short ReadShort(HANDLE fhand) {
        DWORD count;
        short s;

        ReadFile(fhand,&s,sizeof(short),&count,NULL);
        if(count<sizeof(short)) { EOF_yet=true; return 0; }

        return s;
}

void WriteShort(HANDLE fhand,short s) {
        DWORD count;

        WriteFile(fhand,&s,sizeof(short),&count,NULL);
}
*/
