// gadget.cpp -- implements the Gadget class and its subclasses

#include <windows.h>
#include <math.h>
#include <gl/gl.h>
#include "model.h"
#include "gadget.h"
#include "view.h"
#include "pen.h"
#include "funcs.h"
#include "calc.h"
#include "mouse.h"
#include "appdata.h"

//void BoneGadget::draw(lsmView&) { }

//void BoneGadget::select(lsmView&) { }

bool Gadget::release() {
    if(mode) {
        mode=0;
        return true;
    }
    return false;
}

void BoneGadget::rescale() { }

void RotBoneGadget::select() {
    draw();

    glLoadName(MM_GADOVERK);
    glEnable(GL_CULL_FACE);
    float size=model->getBoneLength(*bone);

    bone->notScaled.loadIntoOpenGL();
    glScalef(size,size,size);
    glTranslatef(0.0f,0.0f,1.0f);
    glScalef(.075,.075,.075);
    glCallList(AppData::GLObjBase+AD_OBJ_BALL);
    glDisable(GL_CULL_FACE);
}

void RotBoneGadget::draw() {
    bone->drawRotation2(model->cView.Scale,mode,model->cView.Orient.z);

    if(mode==MM_GADOVERX || mode==MM_GADOVERY || mode==MM_GADOVERZ) {
        Vec3 right;
        GLfloat m[16];
        right.cross(dir,model->cView.Orient.z);
        m[ 0]=right.x;
        m[ 1]=right.y;
        m[ 2]=right.z;
        m[ 4]=dir.x;
        m[ 5]=dir.y;
        m[ 6]=dir.z;
        m[ 8]=model->cView.Orient.z.x;
        m[ 9]=model->cView.Orient.z.y;
        m[10]=model->cView.Orient.z.z;
        for(int i=0;i<11;i++) m[i]*=model->cView.Scale;
        m[3]=m[7]=m[11]=0.0f;
        m[12]=startloc.x;
        m[13]=startloc.y;
        m[14]=startloc.z;
        m[15]=1.0f;
        glLineWidth(1.0f);
        glDisable(GL_DEPTH_TEST);
        glMatrixMode(GL_MODELVIEW);
//        glLoadIdentity();
        glLoadMatrixf(m);
        SelectGLColor(AppData::PenColor[HIGHPEN]);
        glBegin(GL_LINE_STRIP);
        glVertex2f(.075,.9);
        glVertex2f(0,1.0f);
        glVertex2f(0.,-1.0);
        glVertex2f(.075,-.9);
        glEnd();
        glBegin(GL_LINES);
        glVertex2f(-.075,.9);
        glVertex2f(0,1.0f);
        glVertex2f(0.,-1.0);
        glVertex2f(-.075,-.9);
        glEnd();
    }
}

void RotBoneGadget::start(short dragMode,short sX,short sY) {
    Vec3 temp;
    GLfloat mv[16];
    Vec3 fore(model->cView.Orient.z);
//    float size=model->cView.Scale*.5;

    glMatrixMode(GL_MODELVIEW);
    if(bone->data.Parent) {
        bone->data.Parent->notScaled.fore.loadIntoOpenGL();
        glTranslatef(
            bone->data.translate.x*bone->data.Parent->data.scale.x,
            bone->data.translate.y*bone->data.Parent->data.scale.y,
            bone->data.translate.z*bone->data.Parent->data.scale.z);
    } else {
        glLoadIdentity();
        glTranslatef(bone->data.translate.x,bone->data.translate.y,bone->data.translate.z);
    }
//    glScalef(size,size,size);
    mode=dragMode;
    model->cView.mapScr2XYZ(sX,sY,temp);

    if(mode==MM_GADOVERY) {
        glGetFloatv(GL_MODELVIEW_MATRIX,mv);
        normal.x=mv[4];
        normal.y=mv[5];
        normal.z=mv[6];
    }

    glRotatef(bone->data.turnAngle,0.0f,1.0f,0.0f);

    if(mode==MM_GADOVERX) {
        glGetFloatv(GL_MODELVIEW_MATRIX,mv);
        normal.x=mv[0];
        normal.y=mv[1];
        normal.z=mv[2];
    }

    glRotatef(bone->data.bendAngle,1.0f,0.0f,0.0f);

    if(mode==MM_GADOVERZ) {
        glGetFloatv(GL_MODELVIEW_MATRIX,mv);
        normal.x=mv[8];
        normal.y=mv[9];
        normal.z=mv[10];
    }

    if(mode==MM_GADOVERX || mode==MM_GADOVERY || mode==MM_GADOVERZ) {
        tracePlane(bone->Spot,normal,temp,fore,startloc);
        (temp=startloc)-=bone->Spot;
        if(fabs(fore.dot(normal))<.1) { // too oblique
            dir.cross(fore,normal);            
            startloc=bone->Spot;
        } else {
            dir.cross(normal,temp);
            float q;
            q=fore.dot(dir);
            (temp=fore)*=q;
            dir-=temp;
        }
        dir.normalize();
    }

    if(mode==MM_GADOVERK) {
        bone->mSave=bone->notScaled.fore;

        normal=fore;

        if(normal.dot(bone->notScaled.fore.z)>0) {
            normal*=-1.0f;
        }

        ((startloc=bone->notScaled.fore.z)*=model->getBoneLength(*bone))+=bone->Spot;
    }

    bone->bendSave=bone->data.bendAngle;
    bone->turnSave=bone->data.turnAngle;
    bone->twistSave=bone->data.twistAngle;
    model->prepBoneDrag();
}

bool RotBoneGadget::drag(short cX,short cY) {
    Vec3 fore(model->cView.Orient.z);
    GLfloat mv[16];
    if(mode==MM_GADNULL) return false;

    Vec3 vs,ve,cp;
    float s,c;

    glMatrixMode(GL_MODELVIEW);
    if(bone->data.Parent) {
        bone->data.Parent->notScaled.fore.loadIntoOpenGL();
        glTranslatef(
            bone->data.translate.x*bone->data.Parent->data.scale.x,
            bone->data.translate.y*bone->data.Parent->data.scale.y,
            bone->data.translate.z*bone->data.Parent->data.scale.z);
    } else {
        glLoadIdentity();
        glTranslatef(bone->data.translate.x,bone->data.translate.y,bone->data.translate.z);
    }

    model->cView.mapScr2XYZ(cX,cY,vs);
    glGetFloatv(GL_MODELVIEW_MATRIX,mv);

    float rad;
    switch(mode) {
        case MM_GADOVERX:
            vs-=startloc;
            bone->data.bendAngle=bone->bendSave + dir.dot(vs)/model->cView.Scale * RAD2DEG*2.0f;
            break;
        case MM_GADOVERY:
            vs-=startloc;
            bone->data.turnAngle=bone->turnSave + dir.dot(vs)/model->cView.Scale * RAD2DEG*2.0f;
            break;
        case MM_GADOVERZ:
            vs-=startloc;
            bone->data.twistAngle=bone->twistSave + dir.dot(vs)/model->cView.Scale * RAD2DEG*2.0f;
            break;
//            tracePlane(bone->Spot,normal,vs,fore,current);
//            break;
        case MM_GADOVERK:
            rad=model->getBoneLength(*bone);
            traceSphere(bone->Spot,rad,vs,normal,current);

//            break;
//    }
        
    (ve=current)-=bone->Spot;
    ve.normalize();

    (vs=startloc)-=bone->Spot;
    vs.normalize();

    cp.cross(vs,ve);
    c=vs.dot(ve);
    s=normal.dot(cp);
    float ang=atan2(s,c)*180.0/PI;

//    switch(mode) {
//        case MM_GADOVERX:
//            bone->data.bendAngle=bone->bendSave+ang;
//            break;
//        case MM_GADOVERY:
//            bone->data.turnAngle=bone->turnSave+ang;
//            break;
//        case MM_GADOVERZ:
//            bone->data.twistAngle=bone->twistSave+ang;
//            break;
//        case MM_GADOVERK:
            if(c<=-1.0f) {
                InvalidateView();
                return true;
            }
            if(c>=1.0f) {
                // restore original angles
                bone->data.twistAngle=bone->twistSave;
                bone->data.turnAngle=bone->turnSave;
                bone->data.bendAngle=bone->bendSave;
            } else {
                Matrix sMat,eMat;
                sMat.x=vs;
                eMat.x=ve;
                eMat.y=cp;

                eMat.y.normalize();
                sMat.y=eMat.y;
                sMat.z.cross(sMat.x,sMat.y);
                eMat.z.cross(eMat.x,eMat.y);

                Matrix fMat;

                ((fMat=bone->mSave)/=sMat)*=eMat;

                if(bone->data.Parent) {
                    fMat*=bone->data.Parent->notScaled.back;
                }

                if(fMat.z.y<=-1.0f) {
                    bone->setXAngle( 90.0f);
                }
                else if(fMat.z.y>=1.0f) {
                    bone->setXAngle(-90.0f);
                }
                else {
                    bone->setAngles(
                     asinf(-fMat.z.y)*180/PI,
                     atan2(fMat.z.x,fMat.z.z)*180/PI,
                     atan2(fMat.x.y,fMat.y.y)*180/PI );
                }
            }
            break;
    }
    model->updateSkeleton(true);
    InvalidateView();
    return true;
}

//void RotBoneGadget::rescale() { }

ScaleBoneGadget::ScaleBoneGadget(class Model* m,class Bone* b)
    : BoneGadget(m,b) {
    float smax=bone->data.scale.x;
    if(bone->data.scale.y > smax) smax=bone->data.scale.y;
    if(bone->data.scale.z > smax) smax=bone->data.scale.z;
    resize=model->cView.Scale/smax;
}

void ScaleBoneGadget::select() {
    // draw the little knobs that the user can click
    glMatrixMode(GL_MODELVIEW);
    bone->notScaled.loadIntoOpenGL();
    float scale=model->cView.Scale*.05;

// draw x-axis stuff
    if(fabs(bone->notScaled.fore.x.dot(model->cView.Orient.z))<.985) {
        glLoadName(MM_GADOVERX);
        SelectGLColor(AppData::PenColor[XAXISPEN]);
        glPushMatrix();
        glTranslatef(bone->data.scale.x*resize,0,0);
        glScalef(scale,scale,scale);
        glCallList(AppData::GLObjBase+AD_OBJ_CUBE);
        glPopMatrix();
    }
// draw y-axis stuff
    if(fabs(bone->notScaled.fore.y.dot(model->cView.Orient.z))<.985) {
        glLoadName(MM_GADOVERY);
        SelectGLColor(AppData::PenColor[YAXISPEN]);
        glPushMatrix();
        glTranslatef(0,bone->data.scale.y*resize,0);
        glScalef(scale,scale,scale);
        glCallList(AppData::GLObjBase+AD_OBJ_CUBE);
        glPopMatrix();
    }

// draw z-axis stuff
    if(fabs(bone->notScaled.fore.z.dot(model->cView.Orient.z))<.985) {
        glLoadName(MM_GADOVERZ);
        SelectGLColor(AppData::PenColor[ZAXISPEN]);
        glPushMatrix();
        glTranslatef(0,0,bone->data.scale.z*resize);
        glScalef(scale,scale,scale);
        glCallList(AppData::GLObjBase+AD_OBJ_CUBE);
        glPopMatrix();
    }
}

void ScaleBoneGadget::draw() {
    select();

    glScalef(bone->data.scale.x*resize,bone->data.scale.y*resize,bone->data.scale.z*resize);

    glBegin(GL_LINES);

    if(fabs(bone->notScaled.fore.x.dot(model->cView.Orient.z))<.985) {
        SelectGLColor(AppData::PenColor[XAXISPEN]);
        glVertex2s(0,0);
        glVertex2s(1,0);
    }

    if(fabs(bone->notScaled.fore.y.dot(model->cView.Orient.z))<.985) {
        SelectGLColor(AppData::PenColor[YAXISPEN]);
        glVertex2s(0,0);
        glVertex2s(0,1);
    }

    if(fabs(bone->notScaled.fore.z.dot(model->cView.Orient.z))<.985) {
        SelectGLColor(AppData::PenColor[ZAXISPEN]);
        glVertex2s(0,0);
        glVertex3s(0,0,1);
    }

    glEnd();
}

void ScaleBoneGadget::start(short dragMode,short sX,short sY) {
    mode=dragMode;
    if(mode==MM_GADNULL) return;

    Vec3    clickpoint;

    save=bone->data.scale;

    switch(mode) {
        case MM_GADOVERX:
            dir=bone->notScaled.fore.x;
            break;
        case MM_GADOVERY:
            dir=bone->notScaled.fore.y;
            break;
        case MM_GADOVERZ:
            dir=bone->notScaled.fore.z;
            break;
    }

    model->cView.mapScr2XYZ(sX,sY,clickpoint);

    nearest2Line(bone->Spot,dir,clickpoint,model->cView.Orient.z,startloc);
}

bool ScaleBoneGadget::drag(short sX,short sY) {
    if(mode==MM_GADNULL) return false;

    Vec3    clickpoint;
    model->cView.mapScr2XYZ(sX,sY,clickpoint);
    nearest2Line(bone->Spot,dir,clickpoint,model->cView.Orient.z,current);

    float diff=((current-=startloc).dot(dir)/resize);

    switch(mode) {
        case MM_GADOVERX:
            diff+=save.x;
            if(diff!=0.0f) bone->data.scale.x=diff;
            break;
        case MM_GADOVERY:
            diff+=save.y;
            if(diff!=0.0f) bone->data.scale.y=diff;
            break;
        case MM_GADOVERZ:
            diff+=save.z;
            if(diff!=0.0f) bone->data.scale.z=diff;
            break;
    }    

    model->updateSkeleton(true);

    return true;
}

void ScaleBoneGadget::rescale() {
    float smax=bone->data.scale.x;
    if(bone->data.scale.y > smax) smax=bone->data.scale.y;
    if(bone->data.scale.z > smax) smax=bone->data.scale.z;
    resize=model->cView.Scale*.5/smax;
}

void TransBoneGadget::draw() {
    bone->drawTranslation(model->cView.Scale,mode,model->cView.Orient.z);
}

void TransBoneGadget::select() {
    bone->drawTranslation(model->cView.Scale,mode,model->cView.Orient.z);

    glLoadName(MM_GADOVERK);
    bone->selectBase(model->getBoneLength(*bone));
}

void TransBoneGadget::start(short dragMode,short sX,short sY) {
    mode=dragMode;
    if(mode==MM_GADNULL) return;

    Vec3    clickpoint;

    if(bone->data.Parent) {
        switch(mode) {
            case MM_GADOVERX:
                dir=bone->data.Parent->notScaled.fore.x;
                break;
            case MM_GADOVERY:
                dir=bone->data.Parent->notScaled.fore.y;
                break;
            case MM_GADOVERZ:
                dir=bone->data.Parent->notScaled.fore.z;
                break;
        }
    } else {
        dir.set(0);
        switch(mode) {
            case MM_GADOVERX:
                dir.x=1.0f;
                break;
            case MM_GADOVERY:
                dir.y=1.0f;
                break;
            case MM_GADOVERZ:
                dir.z=1.0f;
                break;
        }
    }        

    model->cView.mapScr2XYZ(sX,sY,clickpoint);

    if(mode==MM_GADOVERK) {
        startloc=clickpoint;
        if(bone->data.Parent) startloc/=bone->data.Parent->Scaled;
    } else {
        nearest2Line(bone->Spot,dir,clickpoint,model->cView.Orient.z,startloc);
    }

    bone->Save=bone->data.translate;
}

bool TransBoneGadget::drag(short cX,short cY) {
    Vec3 clickpoint;

    if(mode==MM_GADNULL) return false;

    model->cView.mapScr2XYZ(cX,cY,clickpoint);

    if(mode==MM_GADOVERK) {
        if(bone->data.Parent) clickpoint/=bone->data.Parent->Scaled;
        clickpoint-=startloc;
        (bone->data.translate=bone->Save)+=clickpoint;
        model->updateSkeleton(true);
        return true;
    }
    
    nearest2Line(bone->Spot,dir,clickpoint,model->cView.Orient.z,current);
    current-=startloc;

    float diff=current.dot(dir);
    if(bone->data.Parent) {
        switch(mode) {
            case MM_GADOVERX:
                diff/=bone->data.Parent->data.scale.x;
                break;
            case MM_GADOVERY:
                diff/=bone->data.Parent->data.scale.y;
                break;
            case MM_GADOVERZ:
                diff/=bone->data.Parent->data.scale.z;
                break;
        }
    }

    switch(mode) {
        case MM_GADOVERX:
            bone->data.translate.x=bone->Save.x+diff;
            break;
        case MM_GADOVERY:
            bone->data.translate.y=bone->Save.y+diff;
            break;
        case MM_GADOVERZ:
            bone->data.translate.z=bone->Save.z+diff;
            break;
    }
    model->updateSkeleton(true);
    return true;
/*
    switch(mode) {
        case MM_GADDRAGX:
            nearest2Line(bone->Save,bone->notScaled.fore.getXVec(),
                bone->data.translate,v.Orient.getZVec(),
                bone->data.translate);
            break;
        case MM_GADDRAGY:
            nearest2Line(bone->Save,bone->notScaled.fore.getYVec(),
                bone->data.translate,v.Orient.getZVec(),
                bone->data.translate);
            break;
        case MM_GADDRAGZ:
            nearest2Line(bone->Save,bone->notScaled.fore.getZVec(),
                bone->data.translate,v.Orient.getZVec(),
                bone->data.translate);
            break;
    }
*/
}

//void TransBoneGadget::rescale() { }

void MeshGadget::rescale() { }

// and now for NormTransMeshGadget stuff

void NormTransMeshGadget::draw() {
    glMatrixMode(GL_MODELVIEW);

    Vec3 dy;
    Vec3 dz;

    dy.x=dir.y;
    dy.y=dir.z;
    dy.z=-dir.x;

    dz.cross(dir,dy);
    dz.normalize();
    dy.cross(dir,dz);
    dz.normalize();

    GLfloat m[16];
    m[0]=dir.x;
    m[1]=dir.y;
    m[2]=dir.z;
    m[3]=m[7]=m[11]=0.0f;
    m[4]=dy.x;
    m[5]=dy.y;
    m[6]=dy.z;
    m[8]=dz.x;
    m[9]=dz.y;
    m[10]=dz.z;
    m[12]=center.x;
    m[13]=center.y;
    m[14]=center.z;
    m[15]=1.0;

    glLoadName(MM_GADOVERN);
    glDepthMask(GL_FALSE);

    for(int i=0;i<2;i++) {
        if(i) {
            glLineWidth(1.0f);
            SelectGLColor(AppData::PenColor[EDGEPEN]);
        } else {
            glLineWidth(3.0f);
            SelectGLColor(AppData::PenColor[GRIDPEN]);
        }

        glLoadMatrixf(m);

        float size=model->cView.Scale;

        glScalef(size,size,size);

        glBegin(GL_LINES);
        glVertex2f( .75,0.0);
        glVertex2f(-.75,0.0);
        glEnd();

        for(int j=0;j<2;j++) {
            glPushMatrix();
            switch(i*2+j) {
                case 0:
                    glScalef(-1,1,1);
                case 1:
                    glScalef(.2362777,.0787589,.0787589);
                    glTranslatef(3.124647,0,0);
                    break;
                case 2:
                    glScalef(-1,1,1);
                case 3:
                    glScalef(.1875,.0625,.0625);
                    glTranslatef(4,0,0);
                    break;
            }
            glCallList(AppData::GLObjBase+AD_OBJ_CONE);
            glPopMatrix();
        }
    }
    glDepthMask(GL_TRUE);
}

void NormTransMeshGadget::select() { draw(); }

// need a method to set the direction from inside Model::extrudeMesh()
// need another method to set the center as well

NormTransMeshGadget::NormTransMeshGadget(class Model* m,class Face* f) : MeshGadget(m) {
    f->calcCenter(center);
    f->calcNormal(dir);
    dir.normalize();
}

void NormTransMeshGadget::start(short dragMode,short sX,short sY) {
    mode=dragMode;
    if(mode==MM_GADNULL) return;

    Vec3    clickpoint;

    model->cView.mapScr2XYZ(sX,sY,clickpoint);

    if(mode==MM_GADOVERN) { // should be only value
        nearest2Line(center,dir,clickpoint,model->cView.Orient.z,startloc);
    }

    model->finishVertDrag();

    centsave=center;
}

bool NormTransMeshGadget::drag(short cX,short cY) {
    Vec3 clickpoint;

    if(mode==MM_GADNULL) return false;

    model->cView.mapScr2XYZ(cX,cY,clickpoint);

    if(mode==MM_GADOVERN) {
        nearest2Line(center,dir,clickpoint,model->cView.Orient.z,current);
        current-=startloc;

        Vert*   v;

        STEPVERTS(model,v)
        if(v->flags&TF_MARKED)
        (v->data.Spot=v->Save)+=current;

        (center=centsave)+=current;        

        model->modelFlags|=MF_CALC_SUBD;
        InvalidateView();
    }

    return true;
}

// void NormTransMeshGadget::rescale() { }

UniScaleMeshGadget::UniScaleMeshGadget(class Model* m,class Face* f) : MeshGadget(m) {
    f->calcCenter(center);
    radius = model->cView.Scale;
}

void UniScaleMeshGadget::draw() {
    GLfloat m[16];

    glMatrixMode(GL_MODELVIEW);
    m[ 0]=model->cView.Orient.x.x;
    m[ 1]=model->cView.Orient.x.y;
    m[ 2]=model->cView.Orient.x.z;
    m[ 4]=model->cView.Orient.y.x;
    m[ 5]=model->cView.Orient.y.y;
    m[ 6]=model->cView.Orient.y.z;
    m[ 8]=model->cView.Orient.z.x;
    m[ 9]=model->cView.Orient.z.y;
    m[10]=model->cView.Orient.z.z;
    m[12]=center.x;
    m[13]=center.y;
    m[14]=center.z;
    m[3]=m[7]=m[11]=0.0f;
    m[15]=1.0f;

    glLoadMatrixf(m);

    glLoadName(MM_GADOVERN);
    glScalef(radius,radius,radius);
//    glCallList(AppData::GLObjBase+AD_OBJ_FLIPYZ);
    glDisable(GL_DEPTH_TEST);
    glDepthMask(GL_FALSE);
    glLineWidth(3.0);
    SelectGLColor(AppData::PenColor[GRIDPEN]);
    glCallList(AppData::GLObjBase+AD_OBJ_CIRCLE);

    glLineWidth(1.25);
    SelectGLColor(AppData::PenColor[EDGEPEN]);
    glCallList(AppData::GLObjBase+AD_OBJ_CIRCLE);
    glDepthMask(GL_TRUE);
    glEnable(GL_DEPTH_TEST);
}

void UniScaleMeshGadget::select() { draw(); }

void UniScaleMeshGadget::start(short dragMode,short sX,short sY) {
    float qdist;

    mode=dragMode;
    if(mode==MM_GADNULL) return;

    Vec3    clickpoint;

    model->cView.mapScr2XYZ(sX,sY,clickpoint);

    if(mode==MM_GADOVERN) { // should be only value
        clickpoint-=center;
        qdist=clickpoint.dot(model->cView.Orient.z);
        startrad=clickpoint.dot(clickpoint);
        startrad-=qdist*qdist;
        if(startrad<0.0f) {
            startrad=0.0f;
        } else {
            startrad=sqrtf(startrad);
        }

        model->finishVertDrag();
    }
}

bool UniScaleMeshGadget::drag(short cX,short cY) {
    Vec3 clickpoint;
    float qdist;


    if(mode==MM_GADNULL) return false;

    model->cView.mapScr2XYZ(cX,cY,clickpoint);

    if(mode==MM_GADOVERN) {
        clickpoint-=center;
        radius=clickpoint.dot(clickpoint);
        qdist=clickpoint.dot(model->cView.Orient.z);
        radius-=qdist*qdist;
        if(radius<0.0f) {
            radius=0.0f;
        } else {
            radius=sqrtf(radius);
        }

        Vert*   v;

        STEPVERTS(model,v)
        if(v->flags&TF_MARKED)
        (((v->data.Spot=v->Save)-=center)*=(radius/startrad))+=center;

        model->modelFlags|=MF_CALC_SUBD;
        InvalidateView();
    }

    return true;
}

void UniScaleMeshGadget::rescale() {
    radius = model->cView.Scale;
}

