// gadget.h -- declares the Gadget class and subclasses

#ifndef GADGET_H
#define GADGET_H

#include "pointer.h"
#include "bone.h"
#include "calc.h"

class Model;    // necessary to make the Ref template use the Model class

class Gadget {
    protected:
        Ref<class Model> model;  // needs to communicate with the Model
        short   mode;
//        short   flags;
        Vec3 startloc;
        Vec3 current;
    public:
        Gadget(class Model* m) : mode(0) { model=m; }
        virtual void    select()=0;
        virtual void    draw()=0;
        virtual void    start(short,short,short)=0;
        virtual bool    drag(short,short)=0;
        virtual void    rescale()=0;
        bool    release();
};

class BoneGadget : public Gadget {
    protected:
        Ref<class Bone> bone;
        Vec3    dir;    // the direction that turns the gadget
    public:
        BoneGadget(class Model*,class Bone*);
        virtual void    select()=0;
        virtual void    draw()=0;
        virtual void    start(short,short,short)=0;
        virtual bool    drag(short,short)=0;
        virtual void    rescale();
};

inline BoneGadget::BoneGadget(class Model* m,class Bone* b) : Gadget(m) { bone=b; }

class TransBoneGadget : public BoneGadget {
    protected:
    public:
        TransBoneGadget(class Model*,class Bone*);
        void    select();
        void    draw();
        void    start(short,short,short);
        bool    drag(short,short);
//        void    rescale();
};

inline TransBoneGadget::TransBoneGadget(class Model* m,class Bone* b) : BoneGadget(m,b) { }

class ScaleBoneGadget : public BoneGadget {
    protected:
        float    resize;   // current rescaling value for display gadget
        Vec3    save;   // save the scaling values
    public:
        ScaleBoneGadget(class Model*,class Bone*);
        void    select();
        void    draw();
        void    start(short,short,short);
        bool    drag(short,short);
        void    rescale();
};

class RotBoneGadget : public BoneGadget {
    protected:
        Vec3 normal;
    public:
        RotBoneGadget(class Model*,class Bone*);
        void    select();
        void    draw();
        void    start(short,short,short);
        bool    drag(short,short);
//        void    rescale();
};

inline RotBoneGadget::RotBoneGadget(class Model* m,class Bone* b) : BoneGadget(m,b) { }

class MeshGadget : public Gadget {
    protected:
        Vec3    center,centsave;
    public:
        MeshGadget(class Model*);
        virtual void    select()=0;
        virtual void    draw()=0;
        virtual void    start(short,short,short)=0;
        virtual bool    drag(short,short)=0;
        virtual void    rescale();
};

inline MeshGadget::MeshGadget(class Model* m) : Gadget(m) { }

class NormTransMeshGadget : public MeshGadget {
    protected:
        Vec3 dir;
    public:
        NormTransMeshGadget(class Model*,class Face*);
        void    select();
        void    draw();
        void    start(short,short,short);
        bool    drag(short,short);
//        void    rescale();
};

class UniScaleMeshGadget : public MeshGadget {
    protected:
        float radius;
        float startrad;
    public:
        UniScaleMeshGadget(class Model*,class Face*);
        void    select();
        void    draw();
        void    start(short,short,short);
        bool    drag(short,short);
        void    rescale();
};

#endif

