// lsm_file.cpp -- reads and writes the .lsm format
#include "userint.h"
#include "calc.h"
#include "fhand.h"
#include "model.h"
#include "vert.h"
#include "edge.h"
#include "face.h"
#include "part.h"
#include "uvpt.h"
#include "bone.h"
#include "parser.h"
#include "funcs.h"

extern char InputBuffer[64];
extern char FullPathBuffer[256];

// function to read all native formats from a file handle
Model* Model::readLSMFile(fReader& file,bool insert) {
        char c;
        long type;
        Model* m;

        m=new Model;
        if(m) {
            file >> c;
            type=c;
            file >> c;
            type=(type<<8)|c;
            file >> c;
            type=(type<<8)|c;
            file >> c;
            type=(type<<8)|c;

            switch(type) {
                case LSM0BLOCK: m->readLSM0File(file,insert); break;
                case LSM1BLOCK:
                case LSM2BLOCK:
                case LSM3BLOCK:
                case LSM4BLOCK:
                case LSM5BLOCK: m->readLSM1File(file,insert); break;
                case LSM6BLOCK:
                case LSM7BLOCK: m->readLSM3File(file); break;
                default:
                    UserInt::informUser("This is not a recognized file type.");
                    delete m;
                    return (Model*)0;
            }
            if(insert==false) {
                m->updateSkeleton(false);
                m->destroyAllHistory();
                m->currentVersion=-1;
            }
            m->validateMapping();
            m->validate();
            m->takeSnap();
            m->modelFlags|=MF_BUILD_SUBD;
            m->projectFlags|=PF_NOTDEFAULTNAME;
        }
        return m;
}

// this reads the current LSM format (v1.7)
void Model::readLSM3File(fReader& file,bool /*insert*/) {
    Mystr buffer;
    Mystr name;
    MYPTR m;
    char    type;
    Vert* va;
    Vert* vb;
    Vert* vc;
    Vert* vd;
    int cr,cg,cb;
    float sharp;
    int start_vert=NextVertIndex;
    int start_uvpt=NextUVPtIndex;
    int start_bone=NextBoneIndex;
    int start_tint=NextTintIndex;
    bool bonedupewarn=false;

    while(file) {
//                UserInt::informUser("Reading a line from the file.");
        file.getLine(buffer);
//                UserInt::informUser(buffer);
        buffer.reduceWhiteSpace();

        Parser p(buffer);

//              if(!p.isValid()) UserInt::informUser("Parser object is invalid");

        while(p.isValid())
        switch(p.getChar()) {
            case '#':   // comment line
            case 'L':   // the "LSM*" header
                break;  // these lines are skipped
            case 'b':   // a Bone
                    // skip the space after the b
                    p.getChar();
                    m.b=newBone(0,0);
                    // get the new Bone's parent
                    m.b->data.Parent=indexBone(p.getIndex()+start_bone);
//              if(m.b->data.Name) DestroyString(m.b->data.Name); // no longer needed
                    name.own(p.getString());
                    if(!isUniqueBoneName(name,NULL)) {
                        UserInt::informUser("Duplicate Bone names found, \"_add\" will be appended to each duplicate.");
                        name+="_add";
                    }
                    m.b->data.Name=name;
                    m.b->data.scale.x=p.getFloat();
                    m.b->data.scale.y=p.getFloat();
                    m.b->data.scale.z=p.getFloat();
                    m.b->data.twistAngle=p.getFloat();
                    m.b->data.bendAngle=p.getFloat();
                    m.b->data.turnAngle=p.getFloat();
                    m.b->data.translate.x=p.getFloat();
                    m.b->data.translate.y=p.getFloat();
                    m.b->data.translate.z=p.getFloat();
                break;
            case 'u':   // a UV mapping point
                // skip the space after the u
                p.getChar();
                // and make the new item
                m.u=newUVPt();
                m.u->data.Spot.u=p.getFloat();
                m.u->data.Spot.v=p.getFloat();
                break;
            case 't':   // a Tint
                p.getChar();
                m.t=newTint(0xffffff);
//              if(m.t->data.Name) DestroyString(m.t->data.Name);
                m.t->data.Name.own(p.getString());
//              if(m.t->data.File) DestroyString(m.t->data.File);
                m.t->data.File.own(p.getString());
                cr=p.getIndex();
                cg=p.getIndex();
                cb=p.getIndex();
                m.t->data.color=RGB(cr,cg,cb);
                break;
            case 'l':   // a Layer
//                              p.getChar();
//                              m.l=newLayer();
//                              if(m.l->data.Name) DestroyString(m.l->data.Name);
//                              m.l->data.Name=p.getString();
//                              m.l->data.color=RGB(p.getIndex(),p.getIndex(),p.getIndex());
//                              if(p.getBool()) {
//                                      m.l->data.flags|=LF_HIDDEN;
//                              } else {
//                                      m.l->data.flags&=~LF_HIDDEN;
//                              }
                break;
            case 'v':   // a Vert object
                p.getChar();
                m.v=newVert(0,0);
                m.v->data.Spot.x=p.getFloat();
                m.v->data.Spot.y=p.getFloat();
                m.v->data.Spot.z=p.getFloat();
                m.v->data.Owner=indexBone(p.getIndex()+start_bone);
                if(p.getBool()) {
//                  m.v->data.flags|=VF_LOCKED;
                } else {
//                  m.v->data.flags&=~VF_LOCKED;
                }
                break;
            case 'e':   // an Edge object
                p.getChar();
                m.e=getEdge(indexVert(p.getIndex()+start_vert),indexVert(p.getIndex()+start_vert));
                sharp=p.getFloat();
                if(sharp<0.0f) {
                    m.e->setSharpLevel(255.0f);
                } else {
                    m.e->setSharpLevel(sharp);
                }
                break;
            case 'f':   // a Face object from v 1.6.x
                type=p.getChar();
                p.getChar();

                va=indexVert(p.getIndex()+start_vert);
                vb=indexVert(p.getIndex()+start_vert);
                vc=indexVert(p.getIndex()+start_vert);
                vd=indexVert(p.getIndex()+start_vert);

                m.f=newFace(va,vb,vc,vd);
                m.f->data.Texture=indexTint(p.getIndex()+start_tint);
                if(type=='t') {
                    m.f->data.uvpts=new UVPt*[m.f->data.sides];
                    m.f->data.uvpts[0]=indexUVPt(p.getIndex()+start_uvpt);
                    m.f->data.uvpts[1]=indexUVPt(p.getIndex()+start_uvpt);
                    m.f->data.uvpts[2]=indexUVPt(p.getIndex()+start_uvpt);
                    if(m.f->data.sides>3)
                    m.f->data.uvpts[3]=indexUVPt(p.getIndex()+start_uvpt);
                }
//              m.f->data.layer=indexLayer(p.getIndex());
                break;
            case 'p':   // a Face object from 1.7 and onwards
                type=p.getChar();
                p.getChar();

                int s;

                s=p.getIndex();

                m.f=new Face(s);
                Faces.addTail(m.f);

                if(type=='t')
                m.f->data.uvpts= new UVPt*[s];

                for(int i=0;i<s;i++) {
                    m.f->data.verts[i]=indexVert(p.getIndex()+start_vert);
                    if(type=='t') m.f->data.uvpts[i]=indexUVPt(p.getIndex()+start_uvpt);
                }

                m.f->data.verts[s]=m.f->data.verts[0];
                m.f->data.Texture=indexTint(p.getIndex()+start_tint);

//              for(int i=0;i<s;i++)
//              m.f->edges[i]=getEdge(m.f->data.verts[i],m.f->data.verts[i+1]);
                break;
        } // end of while(p.isValid())-switch(p.getChar()) block
    } // end of while(file)
//        UserInt::informUser("All lines read.");
}

// this reads an older file format
void Model::readLSM1File(fReader& fhand,bool /* insert */) {
    short   type;
    int start_vert=NextVertIndex;
    int start_uvpt=NextUVPtIndex;
    int start_bone=NextBoneIndex;
    int start_tint=NextTintIndex;
    bool bonedupewarn=false;

        do {
                fhand >> type;
                switch(type) {
                        case KNOBDELT:  readLSM1Knob(fhand); break;
                        case PARTDELT:  readLSM1Part(fhand); break;
                        case UVPTDELT:  readLSM1UVPt(fhand); break;
                        case TINTDELT:  readLSM1Tint(fhand); break;
                        case LAYERDELT: readLSM1Layer(fhand); break;
                        case EDGEDELT:  readLSM1Edge(fhand); break;
                        case VERTDELT:  readLSM1Vert(fhand); break;
                        case FACEDELT:  readLSM1Face(fhand); break;
                        case QUADDELT:  readLSM1Quad(fhand); break;
                }
        } while(fhand);
        convertPartModel();
}

void Model::readLSM1Knob(fReader& fhand) {
        Knob*   k;
        short len;
        LSMWORD*    buff;

        fhand >> len;
        buff=new LSMWORD[len];
        fhand.read(buff,len*sizeof(LSMWORD));

        k=newKnob();
        k->Spot.x=buff[0].f;
        k->Spot.y=buff[1].f;
        k->Spot.z=buff[2].f;

        delete []buff;
}

void Model::readLSM1UVPt(fReader& fhand) {
        UVPt*   u;
        short len;
        LSMWORD*    buff;
//      UINT    i;

        fhand >> len ;
        buff=new LSMWORD[len];
        fhand.read(buff,len*sizeof(LSMWORD));

        u=newUVPt();
        u->data.Spot.u=buff[0].f;
        u->data.Spot.v=buff[1].f;

        delete []buff;
}

void Model::readLSM1Tint(fReader& fhand) {
        Tint*   t;
        short len;
        LSMWORD*    buff;
        UINT    i;

        fhand >> len ;
        buff=new LSMWORD[len];
        fhand.read(buff,len*sizeof(LSMWORD));

        t=newTint(0x7f7f7f);

        t->data.color=(COLORREF)(buff[0].l);

        for(i=1;i<len;)
        switch(buff[i].l) {
                case NAMEBLOCK:
                        i++;
//                      if(t->data.Name) DestroyString(t->data.Name);
                        t->data.Name.copy((char*)(buff+i));
                        i+=((t->data.Name.len()+4)>>2);
                break;
                case FILEBLOCK:
                        i++;
//                      if(t->data.File) DestroyString(t->data.File);
                        t->data.File.copy((char*)(buff+i));
                        i+=((t->data.File.len()+4)>>2);
                break;
                default: i++;
        }

        delete []buff;
}

void Model::readLSM1Layer(fReader& fhand) {
//      Layer*  l;
        short len;
        LSMWORD*    buff;
//      UINT    i;

        fhand >> len;
        buff=new LSMWORD[len];
        fhand.read(buff,len*sizeof(LSMWORD));
/*
        l=newLayer();

        l->data.color=(COLORREF)(buff[0].l);

        for(i=1;i<len;)
        switch(buff[i].l) {
                case NAMEBLOCK:
                        i++;
                        if(l->data.Name) DestroyString(l->data.Name);
                        l->data.Name=CloneString((char*)(buff+i));
                        i+=((strlen(l->data.Name)+4)>>2);
                break;
                default: i++;
        }
*/
        delete []buff;
}

void Model::readLSM1Part(fReader& fhand) {
        Part*  p;
        short len;
        LSMWORD*    buff;
        UINT    i;

        fhand >> len ;
        buff=new LSMWORD[len];
        fhand.read(buff,len*sizeof(LSMWORD));

        p=newPart(indexKnob(buff[0].l),indexKnob(buff[1].l),indexKnob(buff[2].l));

        if(p->Base==NULL)
            UserInt::informUser("Base==NULL");
        if(p->Rise==NULL)
            UserInt::informUser("Rise==NULL");
        if(p->End==NULL)
            UserInt::informUser("End==NULL");

        i=3;
        while(i<len)
        switch(buff[i].l) {
                case BALLFLAG:
                        p->flags&=~PT_MASK;
                        p->flags|=PT_BALL;
                        i++;
                break;
                case HINGEFLAG:
                        p->flags&=~PT_MASK;
                        p->flags|=PT_HINGE;
                        i++;
                break;
                case LIMXBLOCK:                 // no longer supported
                case LIMYBLOCK:                 // no longer supported
                case LIMZBLOCK:                 // no longer supported
                case LIMLBLOCK:                 // no longer supported
                        i+=3;
                break;
                case NAMEBLOCK:
                        i++;
//                      if(p->Name) DestroyString(p->Name);
                        p->Name.copy((char*)(buff+i));
                        i+=((p->Name.len()+4)>>2);
                break;
//              case LOCKXBLOCK:
//              case LOCKYBLOCK:
//              case LOCKZBLOCK:
//              case LOCKLBLOCK:
                default:
                        i++;
        }
        delete []buff;
}

void Model::readLSM1Vert(fReader& fhand) {
        Part*   p;
//        Bone*   b;
        Vert*   v;
        bool    use_end;
        short len;
        LSMWORD*        buff;
        int     i;

        fhand >> len ;
        buff=new LSMWORD[len];
        fhand.read(buff,len*sizeof(LSMWORD));

        v=newVert(0,0);

        v->data.Spot.x=buff[0].f;
        v->data.Spot.y=buff[1].f;
        v->data.Spot.z=buff[2].f;

        use_end=true;

        i=4;
        while(i<len)
        switch(buff[i].l) {
                case LAYERBLOCK:
                        // vertex layering no longer supported as of v1.6
                        i+=2;
                break;
                case ATBASEFLAG:
                        use_end=false;
                        i++;
                break;
                case LOCKVFLAG:
//                      v->data.flags|=VF_LOCKED;
                        i++;
                break;
                default: i++;
        }
        delete []buff;

        p=indexPart(buff[3].l);

        if(p) {
//            UserInt::informUser("A Vert with a Part.");
            if(use_end) {
                    v->data.Owner=p->EndBone;
            } else {
                    v->data.Owner=p->BaseBone;
            }
//            if(v->data.Owner==NULL) UserInt::informUser("Vert->Owner wound up NULL anyway.");
        }
}

void Model::readLSM1Edge(fReader& fhand) {
    Edge*   e;
    short len;
    LSMWORD*        buff;
    UINT    i;

    fhand >> len ;
    buff=new LSMWORD[len];
    fhand.read(buff,len*sizeof(LSMWORD));

    e=newEdge(indexVert(buff[0].l),indexVert(buff[1].l));

    i=2;
    while(i<len)
    switch(buff[i].l) {
        case SHARPFLAG:
            e->setSharpLevel(255.0f);
            i++;
        break;
        case LAYERBLOCK:
        // edge layering no longer supported as of v1.6
            i+=2;
        break;
        default: i++;
    }
    delete []buff;
}

// I forget when this object was taken out of the file format--now I'll have to
// keep it around for a while longer...
void Model::readLSM1Face(fReader& fhand) {
        Face*   f;
        short len;
        LSMWORD*        buff;
        UINT    i;

        fhand >> len ;
        buff=new LSMWORD[len];
        fhand.read(buff,len*sizeof(LSMWORD));

        f=newFace(indexVert(buff[0].l),indexVert(buff[1].l),indexVert(buff[2].l),NULL);

        i=3;
        while(i<len)
        switch(buff[i].l) {
                case LAYERBLOCK:
//                      f->data.layer=indexLayer(buff[i+1].l);
                        i+=2;
                break;
                case TINTATTR:
                        f->data.Texture=indexTint(buff[i+1].l);
                        i+=2;
                break;
                case UVMAPDATA:
                        f->data.uvpts=new UVPt*[3];
                        f->data.uvpts[0]=indexUVPt(buff[i+1].l);
                        f->data.uvpts[1]=indexUVPt(buff[i+2].l);
                        f->data.uvpts[2]=indexUVPt(buff[i+3].l);
//                      f->data.MapD=(UVPt*)0;
                        i+=4;
                break;
                default: i++;
        }

        delete []buff;
}

void Model::readLSM1Quad(fReader& fhand) {
        Face*   f;
        short len;
        LSMWORD*        buff;
        UINT    i;

        fhand >> len ;
        buff=new LSMWORD[len];
        fhand.read(buff,len*sizeof(LSMWORD));

        f=newFace(indexVert(buff[0].l),indexVert(buff[1].l),indexVert(buff[2].l),
                indexVert(buff[3].l));

        i=4;
        while(i<len)
        switch(buff[i].l) {
                case LAYERBLOCK:
//                      f->data.layer=indexLayer(buff[i+1].l);
                        i+=2;
                break;
                case TINTATTR:
                        f->data.Texture=indexTint(buff[i+1].l);
                        i+=2;
                break;
                case UVMAPDATA:
                        f->data.uvpts=new UVPt*[4];
                        f->data.uvpts[0]=indexUVPt(buff[i+1].l);
                        f->data.uvpts[1]=indexUVPt(buff[i+2].l);
                        f->data.uvpts[2]=indexUVPt(buff[i+3].l);
                        f->data.uvpts[3]=indexUVPt(buff[i+4].l);
                        i+=5;
                break;
                default: i++;
        }
        delete []buff;
}

// this reads the original file format.
void Model::readLSM0File(fReader& file,bool insert) {
    long type;
    int start_vert=insert?NextVertIndex:0;
    int start_uvpt=insert?NextUVPtIndex:0;

        type=0;
        do {
                if(type==0) file >> type;
                switch(type) {
                        case OLDKNOBBLOCK:      type=readLSM0Knob(file);        break;
                        case OLDPARTBLOCK:      type=readLSM0Part(file);        break;
                        case OLDVERTBLOCK:      type=readLSM0Vert(file);        break;
                        case OLDEDGEBLOCK:      type=readLSM0Edge(file);        break;
                        case OLDFACEBLOCK:      type=readLSM0Face(file);        break;
                        case OLDQUADBLOCK:      type=readLSM0Quad(file);        break;
                        case OLDTINTBLOCK:      type=readLSM0Tint(file);        break;
                        case OLDUVPTBLOCK:      type=readLSM0UVPt(file);        break;
// only listed block types are legal at this level
                        default:        type=0; break;
                }
        } while(file);
        convertPartModel();
}

// this reads a knob block from the original format
long Model::readLSM0Knob(fReader& fhand) {
        Knob *k;

        k=newKnob();
        fhand >> k->Spot.x >> k->Spot.y >> k->Spot.z ;

        return 0;
}

// this reads a part block from the original format
long Model::readLSM0Part(fReader& fhand) {
    Knob *kb,*ke,*kr;
    Part* p;
    long type;
    float discard;
    Vec3 va,vb,vc;

    fhand >> type;
    kb=indexKnob(type);
    fhand >> type;
    ke=indexKnob(type);
    fhand >> type;
    kr=indexKnob(type);
    if(kr==0) {
        kr=newKnob();
        (va=kb->Spot)-=ke->Spot;
        vb.x=va.z;
        vb.y=0;
        vb.z=-va.x;
        vc.cross(va,vb);
        vc.normalize();
        vc*=kb->Spot.distance(ke->Spot);
        ((((kr->Spot=ke->Spot)*=3.0f)+=kb->Spot)+=vc)*=.25;
    }        
    p=newPart(kb,ke,kr);
    ke->Owner=p;

        type=0;
        do {
                if(type==0) fhand >> type ;
                switch(type) {
                case LIMXBLOCK: // was in former formats, no longer used
                        fhand >> discard ;
                        fhand >> discard ;
                        type=0;
                break;
                case LIMYBLOCK: // was in former formats, no longer used
                        fhand >> discard ;
                        fhand >> discard ;
                        type=0;
                break;
                case LIMZBLOCK: // was in former formats, no longer used
                        fhand >> discard ;
                        fhand >> discard ;
                        type=0;
                break;
                case LIMLBLOCK: // was in former formats, no longer used
                        fhand >> discard ;
                        fhand >> discard ;
                        type=0;
                break;
                case LOCKXBLOCK:
//                      p->flags|=PF_LOCKX;
                        type=0;
                break;
                case LOCKYBLOCK:
//                      p->flags|=PF_LOCKY;
                        type=0;
                break;
                case LOCKZBLOCK:
//                      p->flags|=PF_LOCKZ;
                        type=0;
                break;
                case LOCKLBLOCK:
//                      p->flags|=PF_LOCKL;
                        type=0;
                break;
                case NAMEBLOCK:
                        fhand.readPaddedString(InputBuffer,(int)60);
                        InputBuffer[60]=0;
//                      if(p->Name!=NULL) DestroyString(p->Name);
                        p->Name.copy(InputBuffer);
                        type=0;
                break;
                default:        return type;
                }
        } while(fhand);
        return 0;
}

// this reads a UVPt block from the original format
long Model::readLSM0UVPt(fReader& file) {
        UVPt* u;

        u=newUVPt();
        file >> u->data.Spot.u ;
        file >> u->data.Spot.v ;

        return 0;
}

// this reads a vertex block from the original format
long Model::readLSM0Vert(fReader& file) {
        Vert *v;
        long type;
        Part*   p;

        v=newVert(0,0);
        file >> v->data.Spot.x ;
        file >> v->data.Spot.y ;
        file >> v->data.Spot.z ;
        long i;
        file >> i ;
        p=indexPart(i);
        if(p)
        v->data.Owner=p->EndBone;

        type=0;
        do {
                if(type==0) file >> type ;
                switch(type) {
                case OLDFLAG0:
                case ATBASEFLAG:
                        v->flags|=VF_WITHB;
                        if(p)
                        v->data.Owner=p->BaseBone;
                        type=0;
                break;
                case SHARPFLAG:
//                      v->data.flags|=VF_SHARP;
                        type=0;
                break;
                default:
                return type;
                }
        } while(file);
        return 0;
}

// this reads an edge block from the original format
long Model::readLSM0Edge(fReader& file) {
        Edge* e /*,*te */ ;
        Vert *vs,*ve;
        long type;

        file >> type ;
        vs=indexVert(type);
        file >> type ;
        ve=indexVert(type);
        e=getEdge(vs,ve);

        type=0;
        do {
                if(type==0) file >> type ;
                switch(type) {
                        case OLDFLAG1:
                        case SHARPFLAG:
                                e->setSharpLevel(255.0f);
                                type=0;
                        break;
                        default:
                        return type;
                }
        } while(file);
        return 0;
}

// this reads a face block from the original format
long Model::readLSM0Face(fReader& file) {
        long i;
        Face* f;
        Vert *va,*vb,*vc;

        file >> i;
        va=indexVert(i);
        file >> i;
        vb=indexVert(i);
        file >> i;
        vc=indexVert(i);
//      if(va==NULL || vb==NULL || vc==NULL)
//              UserInt::informUser("NULL vert alert!");
        f=newFace(va,vb,vc,NULL);

        return readPaintInfo(file,f);
}

// this reads a quad face from the original format
long Model::readLSM0Quad(fReader& file) {
        long i;
        Face* f;
        Vert *va,*vb,*vc,*vd;

        file >> i ;
        va=indexVert(i);
        file >> i ;
        vb=indexVert(i);
        file >> i ;
        vc=indexVert(i);
        file >> i ;
        vd=indexVert(i);
        f=newFace(va,vb,vc,vd);

        return readPaintInfo(file,f);
}

// this reads the texturing info for a quad or face
long Model::readPaintInfo(fReader& file,Face* f) {
        long type;
        long i;

        type=0;
        do {
                if(type==0) file >> type ;
                switch(type) {
                case TINTATTR:
                        file >> i ;
                        if(f)
                        f->data.Texture=indexTint(i);
                        type=0;
                break;
                case UVMAPDATA:
                        if(f)
                        f->data.uvpts=new UVPt*[f->data.sides];
                        file >> i ;
                        if(f) f->data.uvpts[0]=indexUVPt(i);
                        file >> i ;
                        if(f) f->data.uvpts[1]=indexUVPt(i);
                        file >> i ;
                        if(f) f->data.uvpts[2]=indexUVPt(i);

                        if(f)
                        if(f->data.sides>3)
                                file >> i ;
                                f->data.uvpts[3]=indexUVPt(i);
                        type=0;
                break;
                default: return type;
                }

        } while (file);
        return type;
}

// this reads the Tint object from the original format
long Model::readLSM0Tint(fReader& file) {
        Tint *t;
        long type;

        file >> type ;
        t=newTint((COLORREF)type);

        type=0;
        do {
                if(type==0) file >> type ;
                switch(type) {
                case NAMEBLOCK:
                        file.readPaddedString(InputBuffer,(int)60);
//                      DestroyString(t->data.Name);
                        t->data.Name.copy(InputBuffer);
                        type=0;
                break;
                case FILEBLOCK:
                        file.readPaddedString(FullPathBuffer,(int)252);
//                      DestroyString(t->data.File);
                        t->data.File.copy(FullPathBuffer);
                        type=0;
                break;
                default: return type;
                }

        } while (file);
        return type;
}

void Bone::writeToFile(fWriter& fhand) {
// write the type
        fhand << "b " ;

// write the index of the parent, if any
        if(data.Parent) {
                fhand << data.Parent->index ;
        } else {
                fhand << -1 ;
        }

// write the name in a string
        fhand << " \"" << data.Name << "\" " ;

// write the scale vectors
        fhand << data.scale.x << ' '
                  << data.scale.y << ' '
                  << data.scale.z << ' ' ;

// write the twist, bend, and turn angles
        fhand << data.twistAngle << ' '
                  << data.bendAngle  << ' '
                  << data.turnAngle  << ' ' ;

// write the translation value
        fhand << data.translate.x << ' '
                  << data.translate.y << ' '
                  << data.translate.z << "\15\12" ;
}

void UVPt::writeToFile(fWriter& fhand) {
        fhand << "u "
                  << data.Spot.u << ' '
                  << data.Spot.v << "\15\12" ;
}

void Tint::writeToFile(fWriter& fhand) {
        fhand << "t \"" << data.Name << "\" \"" ;

        if(data.File) fhand << data.File ;
        fhand << "\" " ;

        fhand << GetRValue(data.color) << ' '
                  << GetGValue(data.color) << ' '
                  << GetBValue(data.color) << "\15\12" ;
}

//void Layer::writeToFile(HANDLE fhand) {
//      WriteStrings(fhand,"l \"",data.Name,"\" ",NULL);

//      WriteIntString(fhand,GetRValue(data.color));
//      WriteSpace(fhand);
//      WriteIntString(fhand,GetGValue(data.color));
//      WriteSpace(fhand);
//      WriteIntString(fhand,GetBValue(data.color));

//      WriteSpace(fhand);

//      if(data.flags&LF_HIDDEN) {
//              WriteString(fhand,"1");
//      } else {
//              WriteString(fhand,"0");
//      }

//      WriteString(fhand,"\15\12");
//}

void Vert::writeToFile(fWriter& fhand) {
        fhand << "v " ;

        fhand << data.Spot.x << ' '
                  << data.Spot.y << ' '
                  << data.Spot.z << ' ' ;

        if(data.Owner) {
                fhand << data.Owner->index << ' ' ;
        } else {
                fhand << "-1 " ;
        }

        fhand << "0\15\12" ;
}

// only interior sharp edges need to be saved
void Edge::writeToFile(fWriter& fhand) {
        if(sharpType()!=0 && sharpType() != -2) {
                fhand << "e " ;

                fhand << data.VertS->index << ' '
                          << data.VertE->index << ' ' ;

                fhand << getSharpLevel() ;

                fhand << "\15\12" ;
        }
}

void Face::writeToFile(fWriter& fhand) {
        if(data.uvpts) {
                fhand << "pt " ;
        } else {
                fhand << "pp " ;
        }

        fhand << data.sides << ' ' ;

        for(int i=0;i<data.sides;i++) {
        fhand << data.verts[i]->index << ' ' ;
                if(data.uvpts)
                fhand << data.uvpts[i]->index << ' ' ;
        }

        fhand << data.Texture->index << "\15\12" ;
}

void Model::writeLSMFile(fWriter& fhand) {
        MYPTR   m;

        fhand << "LSM7\15\12" ;
        for(INDEX i=0;i<NextBoneIndex;i++) {
                m.b=indexBone(i);
                if(m.b) m.b->writeToFile(fhand);
        }
        STEPUVPTS(this,m.u) m.u->writeToFile(fhand);
        STEPTINTS(this,m.t) m.t->writeToFile(fhand);
        STEPVERTS(this,m.v) m.v->writeToFile(fhand);
        STEPEDGES(this,m.e) m.e->writeToFile(fhand);
        STEPFACES(this,m.f) m.f->writeToFile(fhand);
}
