#include "userint.h"
#include <windowsx.h>
#include <commctrl.h>
#include <shellapi.h>
#include <gl/glu.h>
#include "appdata.h"
#include "model.h"
#include "face.h"
#include "lsmres.h"
#include "scene.h"
#include "funcs.h"
#include "mouse.h"

const char WINDCLASSNAME[]="lsnWndClass";

// HWND hwndProp=(HWND)0;      // the properties window

LRESULT CALLBACK MainWndProc(HWND,UINT,WPARAM,LPARAM);

static BOOL InitWndClass(HINSTANCE hInst) {
    WNDCLASS wc;

    memset(&wc,0,sizeof(WNDCLASS));
    wc.style = CS_HREDRAW|CS_VREDRAW|CS_DBLCLKS ;
    wc.lpfnWndProc = (WNDPROC)MainWndProc;
    wc.hInstance = hInst;
    wc.hbrBackground = (HBRUSH)0;
    wc.lpszClassName = WINDCLASSNAME;
    wc.lpszMenuName = NULL;
    wc.hCursor = LoadCursor(NULL,IDC_ARROW);
    wc.hIcon = LoadIcon(hInst,MAKEINTRESOURCE(IDI_LARGEICON));
    return (RegisterClass(&wc)==(ATOM)0)?FALSE:TRUE;
}

// this is for the commands that are executed whether there is a project open or not.
bool doCommonCommands(HWND hwnd, int id,HWND /* hwndCtl */, UINT /* codeNofity */) {
    switch(id) {
        case IDM_IMPORT:
            ImportProject();
            return true;
        case IDM_OPEN:
            OpenProject();
            return true;
        case IDM_NEWMODEL:
            Model::createModel(hwnd);
            return true;
//                case IDM_NEWSCENE:
//                        new Scene;
//                        Project::activateFirst();
//                        return true;
        case IDM_ABOUT:
            DialogBox(UserInt::hInst,MAKEINTRESOURCE(DLG_ABOUT), UserInt::hwndMain,AboutDlg);
            return true;
        case IDM_PANIC:
            ShellExecute(hwnd,NULL,"lsm.htm",NULL,NULL,SW_SHOW);
            return true;
        case IDM_EXIT:
            PostMessage(hwnd,WM_CLOSE,0,0);
            return true;
        case IDM_PREVPROJ:
            Project::activatePrev();
            break;
        case IDM_NEXTPROJ:
            Project::activateNext();
            break;
    }

    return false;
}

void MainWndProc_OnCommand(HWND hwnd, int id, HWND hwndCtl, UINT codeNotify)
{
        if(doCommonCommands(hwnd,id,hwndCtl,codeNotify)) return;

        if(Project::activeProject) Project::activeProject->doCommands(hwnd,id,hwndCtl,codeNotify);
}

void Model::doCommands(HWND hwnd, int id, HWND /* hwndCtl */, UINT /* codeNotify */ ) {
    RECT wRect;

    switch(id) {
        case IDM_ADDBONE:
            if(HighType==HT_BONE) {
                drag.b=HighItem.b;
            } else {
                drag.b=(Bone*)0;
            }
            setMouseMode(MM_ADDBONE);
            break;
        case IDM_ADDMESH:
            clearAll();
            blankSelection();
            setMouseMode(MM_ADDMESH0);
            break;
        case IDM_ADDVERT:
            setMouseMode(MM_ADDVERT);
            break;
            
        case IDM_ADJUST_ROT:
            if(HighType==HT_BONE || HighType==HT_KNOB)
            if(HighItem.b) {
                gadget=new RotBoneGadget(this,HighItem.b);
                InvalidateView();
            }
            break;
        case IDM_ADJUST_SCALE:
            if(HighType==HT_BONE)
            if(HighItem.b) {
                gadget=new ScaleBoneGadget(this,HighItem.b);
                InvalidateView();
            }
            break;
        case IDM_ADJUST_TRANS:
            if(HighType==HT_BONE)
            if(HighItem.b) {
                gadget=new TransBoneGadget(this,HighItem.b);
                InvalidateView();
            }
            break;

        case IDM_CLOSE:
            Project::closeActiveProject();
            break;

        case IDM_EXT_HIGH:
            extrudeHigh();
            break;

        case IDM_HIDE_HIGH:
            hideHigh();
            InvalidateView();
            break;
        case IDM_HIDE_MARK:
            hideMarked();
            InvalidateView();
            break;
        case IDM_HIDE_UNMARKED:
            hideUnmarked();
            InvalidateView();
            break;

        case    IDM_INSERT_CUBE:
            insertCube();
            break;
        case IDM_INSET_HIGH:
            insetHigh();
            break;

        case IDM_MARK_CONT:
            markContiguous();
            break;

        case IDM_NORM_DRAG_HIGH:
            if(HighType==HT_FACE)
            if(HighItem.f) {
                gadget=new NormTransMeshGadget(this,HighItem.f);
                InvalidateView();
            }
            break;

        case IDM_PANVIEW:
            clearHighlight();
            setMouseMode(MM_PANVIEW0);
            break;
        case IDM_PUSH_BACK:
            pushBackHigh();
            break;

        case IDM_ROTVIEW:
            clearHighlight();
            setMouseMode(MM_ROTVIEW0);
            break;

        case IDM_SELECT:
            setMouseMode(MM_NOTHING);
            break;
        case IDM_SETGRID:
            setGridSnap(hwnd);
            break;
        case IDM_SETFINE:
            setFineLevel(hwnd);
            break;

        case IDM_SHOWFACES:
            modelFlags^=MF_SHOW_FACES;
            InvalidateView();
            updateMenu();
            updateToolBar();
            if(HighType==HT_FACE) {
                HighType=HT_NOTHING;
                HighItem.f=NULL;
            }
            break;

        case IDM_SHOWSUBD:
            modelFlags^=MF_SHOW_SUBD;
            if(modelFlags&MF_SHOW_SUBD) {
                modelFlags|=MF_BUILD_SUBD;
            } else {
                deleteSubData();
            }
            updateMenu();
            updateToolBar();
            if(HighType==HT_FACE) {
                HighType=HT_NOTHING;
                HighItem.p=NULL;
            }
            InvalidateView();
            break;

        case IDM_SPLIT_HIGH:
            splitHigh(false);
            break;
        case IDM_SPLIT_MARK:
            splitMarked(false);
            break;
        case IDM_SUBDIVIDE:
            splitMarked(true);
            break;

        case IDM_TOGGLEHIGH:
            toggleHighlight();
            break;

        case IDM_UNHIDE_ALL:
            unhideAll();
            InvalidateView();
            break;

        case IDM_ZOOMVIEW:
            clearHighlight();
            setMouseMode(MM_ZOOMVIEW0);
            break;

//              case IDM_IMPORT:        ImportProject();        break;
        case IDM_EXPORT:
            exportFile();
            break;
        case IDM_SAVE:
            saveProject(FALSE);
            break;
        case IDM_SAVEAS:
            saveProject(TRUE);
            break;
        case IDM_MATRIX:
            matrixMarked(hwnd,EA_NADA);
            break;
        case IDM_TRANSLATE:
            translateMarked(hwnd,EA_NADA);
            break;
        case IDM_SCALE:
            scaleMarked(hwnd,EA_NADA);
            break;
        case IDM_ROTATE:
            rotateMarked(hwnd,EA_NADA);
            break;
        case IDM_REPARENT:
            editHighParent(hwnd);
            break;
        case IDM_MARKALL:
            markAll();
            break;
        case IDM_MARKCLR:
            clearAll();
            break;
        case IDM_MARKTOG:
            toggleAll();
            break;
        case IDM_MARK_PROP:
            editMarkedProperties(hwnd);
            break;
        case IDM_HIGH_PROP:
            editHighProperties(hwnd);
            break;
        case IDM_EXTRUDE:
            GetWindowRect(hwnd,&wRect);
            TrackPopupMenuEx(GetSubMenu(UserInt::ExtrudePopupMenu,0),TPM_LEFTALIGN|TPM_LEFTBUTTON,
                wRect.left+lX,wRect.top+lY,hwnd,NULL);
            break;
        case IDM_COPY:
            GetWindowRect(hwnd,&wRect);
            TrackPopupMenuEx(GetSubMenu(UserInt::CopyPopupMenu,0),TPM_LEFTALIGN|TPM_LEFTBUTTON,
                wRect.left+lX,wRect.top+lY,hwnd,NULL);
            break;
        case IDM_EXT_STAY:
            extrudeMarked();
            takeSnap();
            break;
        case IDM_EXT_SCALE:
            scaleMarked(hwnd,EA_EXTRUDE);
            break;
        case IDM_EXT_ROTATE:
            rotateMarked(hwnd,EA_EXTRUDE);
            break;
        case IDM_EXT_TRANSLATE:
            translateMarked(hwnd,EA_EXTRUDE);
            break;
        case IDM_EXT_MATRIX:
            matrixMarked(hwnd,EA_EXTRUDE);
            break;
        case IDM_COPY_STAY:
            copyMarked();
            takeSnap();
            break;
        case IDM_COPY_SCALE:
            scaleMarked(hwnd,EA_COPY);
            break;
        case IDM_COPY_ROTATE:
            rotateMarked(hwnd,EA_COPY);
            break;
        case IDM_COPY_TRANSLATE:
            translateMarked(hwnd,EA_COPY);
            break;
        case IDM_COPY_MATRIX:
            matrixMarked(hwnd,EA_COPY);
            break;
        case IDM_DELMARK:
            GetWindowRect(hwnd,&wRect);
            TrackPopupMenuEx(GetSubMenu(UserInt::DelPopupMenu,0),TPM_LEFTALIGN|TPM_LEFTBUTTON,
                wRect.left+lX,wRect.top+lY,hwnd,NULL);
            break;
        case IDM_DELVERTS:
            deleteVerts();
            break;
        case IDM_DELEDGES:
            deleteEdges();
            break;
        case IDM_DELFACES:
            deleteFaces();
            break;
        case IDM_DELHIGH:
            deleteHighlight();
            break;
        case IDM_SHOWAXES:
            projectFlags^=PF_SHOWAXES;
            InvalidateView();
            updateMenu();
            updateToolBar();
            break;
        case IDM_FROML:
            if(EditMode!=EM_MAPEDIT)
            cView.ViewFromLeft();
            break;
        case IDM_FROMR:
            if(EditMode!=EM_MAPEDIT)
            cView.ViewFromRight();
            break;
        case IDM_FROMU:
            if(EditMode!=EM_MAPEDIT)
            cView.ViewFromUp();
            break;
        case IDM_FROMD:
            if(EditMode!=EM_MAPEDIT)
            cView.ViewFromDown();
            break;
        case IDM_FROMF:
            if(EditMode!=EM_MAPEDIT)
            cView.ViewFromFront();
            break;
        case IDM_FROMB:
            if(EditMode!=EM_MAPEDIT)
            cView.ViewFromBack();
            break;
        case IDM_ZOOMIN:
            cView.zoomInTenPercent();
            break;
        case IDM_ZOOMOUT:
            cView.zoomOutTenPercent();
            break;
        case IDM_ZOOMFIT:
            zoomToFit();
            break;
        case IDM_ZOOMMARK:
            zoomToSelection();
            break;
        case IDM_JUMPORG:
            cView.jumpToOrigin();
            break;
        case IDM_MESHMODE:
            enterMode(EM_MESHEDIT);
            break;
        case IDM_MAPMODE:
            enterMode(EM_MAPEDIT);
            break;
        case IDM_EDIT_HIGH_TINT:
            editHighTint(hwnd);
            break;
        case IDM_EDITMARKTINT:
            editMarkedTint(hwnd);
            break;
        case IDM_ADDMAPPING:
            if(HighItem.f)
            /* if(HighItem.f->data.uvpts==NULL) */ {
                addMapping(HighItem.f,true);
                validateMapping();
                takeSnap();
                InvalidateView();
            }
            break;
        case IDM_REMMAPPING:
            if(HighItem.f) {
                HighItem.f->removeMapping();
                validateMapping();
                takeSnap();
            }
            InvalidateView();
            break;
        case IDM_EDITMAPPING:
            if(HighType==HT_FACE)
            if(HighItem.f->data.uvpts) {
                CurTint=HighItem.f->data.Texture;
                enterMode(EM_MAPEDIT);
            }
            break;
        case IDM_DEFTINT:
            if(HighType==HT_FACE) {
                HighItem.f->data.Texture=CurTint;
                InvalidateView();
                takeSnap();
            }
            break;
        case IDM_SHOWPARTS:
            modelFlags^=MF_SHOW_PARTS;
            updateMenu();
            updateToolBar();
            if(HighType==HT_BONE) {
                HighType=HT_NOTHING;
                HighItem.b=NULL;
            }
            InvalidateView();
            break;
        case IDM_SHOWVERTS:
            modelFlags^=MF_SHOW_VERTS;
            InvalidateView();
            updateMenu();
            updateToolBar();
            if(HighType==HT_VERT) {
                HighType=HT_NOTHING;
                HighItem.p=NULL;
            }
            break;
        case IDM_SHOWEDGES:
            modelFlags^=MF_SHOW_EDGES;
            InvalidateView();
            updateMenu();
            updateToolBar();
            if(HighType==HT_EDGE) {
                HighType=HT_NOTHING;
                HighItem.p=NULL;
            }
            break;

        case IDM_UNDO:
                        undoChanges();
                        InvalidateView();
                        break;
        case IDM_REDO:
                        redoChanges();
                        InvalidateView();
                        break;
        case IDM_LATHE:
                        if(Project::activeProject)
                        makeLathe(hwnd);
                        break;
        case IDM_LOCK_X:
                    modelFlags^=MF_LOCK_X;
                    updateMenu();
                    updateToolBar();
                    break;
        case IDM_LOCK_Y:
                    modelFlags^=MF_LOCK_Y;
                    updateMenu();
                    updateToolBar();
                    break;
        case IDM_LOCK_Z:
                    modelFlags^=MF_LOCK_Z;
                    updateMenu();
                    updateToolBar();
                    break;
        case IDM_JOIN:
                    mergeAtEdge(HighItem.e);
                    break;
    }
}

void Scene::doCommands(HWND /* hwnd */, int /* id */ , HWND /* hwndCtl */, UINT /* codeNotify */ ) { }

LRESULT CALLBACK MainWndProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam)
{
    HMENU tpu;
    TOOLTIPTEXT *tipinfo;
    PAINTSTRUCT ps;
//      NMHDR   *header;

    switch (msg) {
        case WM_ERASEBKGND:
            return 1;
        case WM_PAINT:
            BeginPaint(UserInt::hwndMain,&ps);
            Project::redoDisplay(Project::activeProject);
            EndPaint(UserInt::hwndMain,&ps);
            break;
        case WM_MOUSEMOVE:
            if(Project::activeProject) Project::activeProject->mouseMove(hwnd,wParam,lParam);
            break;
        case WM_LBUTTONDOWN:
                if(Project::activeProject) Project::activeProject->leftClick(hwnd,wParam,lParam);
                break;
        case WM_LBUTTONDBLCLK:
                if(Project::activeProject) Project::activeProject->leftDoubleClick(hwnd,wParam,lParam);
                break;
        case WM_LBUTTONUP:
                if(Project::activeProject) Project::activeProject->leftRelease(hwnd,wParam,lParam);
                break;
        case WM_RBUTTONDOWN:
                if(Project::activeProject) Project::activeProject->rightClick(hwnd,wParam,lParam);
                break;
        case WM_RBUTTONUP:
                if(Project::activeProject) {
                        Project::activeProject->rightRelease(hwnd,wParam,lParam);
                        if(AppData::Popup) return DefWindowProc(hwnd,msg,wParam,lParam);
                }
        break;
        case WM_CONTEXTMENU:
                if(Project::activeProject)
                switch(Project::activeProject->HighType) {
                        case HT_KNOB:
                        case HT_BONE:
                            tpu=LoadMenu(UserInt::hInst,MAKEINTRESOURCE(IDPOPUPBONE));
                            TrackPopupMenuEx(GetSubMenu(tpu,0),TPM_LEFTALIGN|TPM_LEFTBUTTON,
                                LOWORD(lParam),HIWORD(lParam),hwnd,NULL);
                            DestroyMenu(tpu);
                        break;
                        case HT_EDGE:
                            tpu=((Model*)Project::activeProject)->buildEdgePopup();
//                            tpu=LoadMenu(UserInt::hInst,MAKEINTRESOURCE(IDPOPUPEDGE));
                            TrackPopupMenuEx(tpu,TPM_LEFTALIGN|TPM_LEFTBUTTON,
                                LOWORD(lParam),HIWORD(lParam),hwnd,NULL);
                            DestroyMenu(tpu);
                        break;
                        case HT_VERT:
                            tpu=LoadMenu(UserInt::hInst,MAKEINTRESOURCE(IDPOPUPVERT));
                            TrackPopupMenuEx(GetSubMenu(tpu,0),TPM_LEFTALIGN|TPM_LEFTBUTTON,
                                LOWORD(lParam),HIWORD(lParam),hwnd,NULL);
                            DestroyMenu(tpu);
                        break;
                        case HT_UVPT:
                            tpu=LoadMenu(UserInt::hInst,MAKEINTRESOURCE(IDPOPUPUVPT));
                            TrackPopupMenuEx(GetSubMenu(tpu,0),TPM_LEFTALIGN|TPM_LEFTBUTTON,
                                LOWORD(lParam),HIWORD(lParam),hwnd,NULL);
                            DestroyMenu(tpu);
                        break;
                        case HT_FACE:
                            tpu=((Model*)Project::activeProject)->buildFacePopup();
                            TrackPopupMenuEx(tpu,TPM_LEFTALIGN|TPM_LEFTBUTTON,
                                LOWORD(lParam),HIWORD(lParam),hwnd,NULL);
                            DestroyMenu(tpu);
                        break;
                        default:
                            switch(Project::activeProject->EditMode) {
                                case EM_MESHEDIT:
                                    tpu=Project::activeProject->getPopupMenu();
                                    TrackPopupMenuEx(GetSubMenu(tpu,0),TPM_LEFTALIGN|TPM_LEFTBUTTON,
                                        LOWORD(lParam),HIWORD(lParam),hwnd,NULL);
                                    DestroyMenu(tpu);
                                    break;
                                case EM_MAPEDIT:
                                    tpu=LoadMenu(UserInt::hInst,MAKEINTRESOURCE(IDPOPUPMAP));
                                    TrackPopupMenuEx(GetSubMenu(tpu,0),TPM_LEFTALIGN|TPM_LEFTBUTTON,
                                        LOWORD(lParam),HIWORD(lParam),hwnd,NULL);
                                    DestroyMenu(tpu);
                                    break;
                                }
                        }
        break;
        case WM_KILLFOCUS:      // if window loses focus
        case WM_CANCELMODE:     // or an unexpected dialog pops up
                if(Project::activeProject) Project::activeProject->abortMode();
        break;
//      case WM_HSCROLL: Project::activeProject->hScroll(wParam); break;
//      case WM_VSCROLL: Project::activeProject->vScroll(wParam); break;
        case WM_CLOSE:
            if(Project::destroyAllProjects()) {
                DestroyWindow(UserInt::hwndMain);
            }
        break;
        case WM_SIZE:
                SendMessage(UserInt::hwndStatusBar,msg,wParam,lParam);
                UserInt::resizeStatusBar(UserInt::hwndStatusBar,1);
                SendMessage(UserInt::hwndToolBar, TB_AUTOSIZE, 0, 0);
        break;
//      case WM_MENUSELECT:
//              return MsgMenuSelect(hwnd,msg,wParam,lParam);
    case WM_COMMAND:
        MainWndProc_OnCommand(hwnd, LOWORD(wParam),(HWND)lParam, HIWORD(wParam));
        break;
        case WM_DESTROY:
                wglMakeCurrent(NULL,NULL);
                if(UserInt::hrc) {
                        wglDeleteContext(UserInt::hrc);
                        UserInt::hrc=NULL;
                }
                PostQuitMessage(0);
        break;
        case WM_HELP:
                        switch((int)ShellExecute(hwnd,"open","lsm.htm",NULL,NULL,SW_SHOW)) {
                                case ERROR_FILE_NOT_FOUND:
                                        MessageBoxEx(hwnd,"File \42lsm.htm\42 not found","Error",MB_OK,0);
                                break;
                        }
        break;
        case WM_CREATE:
            Project::activateProject((Project*)0);
            InvalidateView();
        break;
        case WM_NOTIFY:
                tipinfo=(LPTOOLTIPTEXT)lParam;
//              header=(LPNMHDR)lParam;

    if ((int)(tipinfo->hdr.code) == (int)TTN_NEEDTEXT) {
        tipinfo->hinst = UserInt::hInst;
        tipinfo->lpszText = MAKEINTRESOURCE(tipinfo->hdr.idFrom);
    }
        return 0;

//        break;
        default:
                return DefWindowProc(hwnd,msg,wParam,lParam);
        }
        return 0;
}

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE /* hPrevInstance */ , LPSTR lpCmdLine, INT /* nCmdShow */ )
{
    MSG msg;
    HACCEL hAccelTable;

    UserInt ui(hInstance);    // initializes all application data
    AppData ad;

    if(InitWndClass(hInstance)==FALSE) return 0;
    hAccelTable = LoadAccelerators(UserInt::hInst,MAKEINTRESOURCE(IDACCEL));
    UserInt::hwndMain=CreateWindow(WINDCLASSNAME,NULL,WS_TILEDWINDOW|WS_VISIBLE|
        /* WS_VSCROLL|WS_HSCROLL| */ WS_CLIPSIBLINGS|WS_CLIPCHILDREN,
        CW_USEDEFAULT,0,CW_USEDEFAULT,0,NULL,NULL,UserInt::hInst,NULL);
    if (UserInt::hwndMain==(HWND)0) return 0;
    UserInt::createStatusBar(UserInt::hwndMain,"Ready",1);
    UserInt::hwndToolBar=CreateAToolBar(UserInt::hwndMain,UserInt::hInst);
    SetMenu(UserInt::hwndMain,UserInt::LameMenu);
    ShowWindow(UserInt::hwndMain,SW_SHOW);
    Project::setWindowTitle(UserInt::hwndMain,NULL);
//      hwndTab=createTabControl(hwndMain,hInst);
    OpenFromCommandLine(lpCmdLine);
    PostMessage(UserInt::hwndMain,WM_PAINT,0,0);
    while (GetMessage(&msg,NULL,0,0)) {
        if(!TranslateAccelerator(msg.hwnd,hAccelTable,&msg)) {
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        }
    }
    return msg.wParam;
}

