#ifndef MODEL_H
#define MODEL_H

#include "project.h"
#include "list.h"
#include "part.h"
#include "vert.h"
#include "face.h"
#include "bone.h"
#include "join.h"
#include "edge.h"
#include "tint.h"
#include "uvpt.h"
#include "norm.h"
#include "obvec.h"

class Model : public Project {
        public:
// objects is use
        tList<class Knob>   Knobs;
        tList<class Part>   Parts;
        List  Joins;
//        Chain<class Join>   delJoins; // I doubt that I need these
        tList<class Vert>   Vertices;
        tList<class Vert>   delVerts;
        tList<class Edge>   Edges;
        tList<class Edge>   delEdges;
        tList<class Face>   Faces;
        tList<class Face>   delFaces;
        tList<class Tint>   Tints;
        tList<class Tint>   delTints;
        tList<class UVPt>   UVPts;
        tList<class UVPt>   delUVPts;
        tList<class Norm>   Norms;
        tList<class Bone>   Bones;
        tList<class Bone>   delBones;
// deleted objects (saved for later restoration)
        INDEX   NextEdgeIndex,NextFaceIndex,NextKnobIndex,NextPartIndex,
                NextTintIndex,NextUVPtIndex,NextVertIndex, /* NextLayerIndex, */
                NextBoneIndex /* ,NextNormIndex */ ;
//        Tint    defTint;        // the default tint
        Tint    *CurTint;
        lsmView mView;  // save from the mesh view
        lsmView uView;  // save from uv-map editing
        float   MeshSnap,MapSnap;       // the current grid snap precision
        int             LatheSteps;                     // the number of lathe divisions
        int modelFlags; // flags used by the model and not by the project
        int subPreviewLevel;
        MYPTR   drag;
        MYPTR   anch;
        Ref<Bone>   defBone;
        Array<Vert*>    verttable;  // for rapid retrieval of verts via index
        Array<UVPt*>    uvpttable;  // for rapid retrieval of mapping points via index
        struct Patch*   patches;    // for subdivision preview

// member functions

// implemented in files.cpp
        void    exportFile();
        void    saveProject(bool);

// implemented in lsm_file.cpp
        static Model* readLSMFile(class fReader&,bool insert=false);
        void    readLSM0File(class fReader&,bool insert=false);
        void    readLSM1File(class fReader&,bool insert=false);
//      void    readLSM2File(class fReader&,bool insert=false);
        void    readLSM3File(class fReader&,bool insert=false);
        void    writeLSMFile(class fWriter&);

        long    readLSM0Knob(class fReader&);
        void    readLSM1Knob(class fReader&);
        long    readLSM0Part(class fReader&);
        void    readLSM1Part(class fReader&);
        long    readLSM0UVPt(class fReader&);
        void    readLSM1UVPt(class fReader&);
        long    readLSM0Tint(class fReader&);
        void    readLSM1Tint(class fReader&);
        void    readLSM1Layer(class fReader&);
        long    readLSM0Vert(class fReader&);
        void    readLSM1Vert(class fReader&);
        long    readLSM0Edge(class fReader&);
        void    readLSM1Edge(class fReader&);
        long    readLSM0Face(class fReader&);
        void    readLSM1Face(class fReader&);
        long    readLSM0Quad(class fReader&);
        void    readLSM1Quad(class fReader&);

        long    readPaintInfo(class fReader&,Face*);

// implemented in main.cpp
        void    doCommands(HWND,int,HWND,UINT);

// implemented in menu.cpp
        HMENU   buildEdgePopup();
        HMENU   buildFacePopup();

// implemented in model.cpp
        Model();
        ~Model();

// implemented in model_a.cpp
        void    activate();
        void    addMapping(Face*,bool);
        void    applyMarkLogic();
        void    assignFaces();
        void    assignVerts();

// implemented in model_b.cpp
        void    baseParts();
        void    blankSelection();
        void    buildBoneComboBox(HWND,int,Bone*);
        void    buildBoneHierarchy();
        void    buildNorms();
        void    buildTintComboBox(HWND,int);

// implemented in model_c.cpp
        void    calcNorms();
        bool    canMergeAtEdge(Edge*);
        void    checkHighlight(int);
        short   checkGadget();
        void    checkSelectBox(RECT *,short);
        void    clearAll();
        void    clearAllDragging();
        void    clearHighlight();
        void    convertPartModel(); // for converting Part-based models to Bone-based
        Face*   copyFace(Face*);
        void    copyMarked();
        UVPt*   copyUVPt(UVPt*);
        char*   createBoneName();
        static Model*   createModel(HWND);
        static char*    createModelName();
        char*   createPartName();
        char*   createRootName();
        char*   createTintName();

// implemented in model_d.cpp
        void    decSelection();
        void    deleteBone(Bone*);
        void    deleteEdge(Edge*);
        void    deleteEdges();
        void    deleteFace(Face*);
        void    deleteFaces();
        void    deleteTint(Tint*);
        void    deleteUVPt(UVPt*);
        void    deleteVert(Vert*);
        void    deleteVerts();
        void    deleteHighlight();
        void    dragBone(Bone*,Vec3&);
        void    draw(); // instantiates inherited virtual void function
        void    drawBones();
        void    drawEdges();
        void    drawFaces();
        void    drawHighlight();
        void    drawMarking();
        void    drawSubdivision();
        void    drawUVEdges();
        void    drawUVPts();
        void    drawVertices();

// implemented in model_e.cpp
        void    editHighParent(HWND);
        void    editHighProperties(HWND);
        void    editHighTint(HWND);
        void    editMarkedProperties(HWND);
        void    editMarkedTint(HWND);
        void    enterMode(short);
        void    extrudeHigh();
        void    extrudeMarked();

// implemented in model_f.cpp
        Bone*   findBoneByName(Mystr&);
        Tint*   findTintByName(Mystr&);
        void    finishVertDrag();

// implemented in model_g.cpp
        void    getBoneBound(Bone&,class Bound&);
        float   getBoneLength(Bone&);
        Edge*   getEdge(Vert*,Vert*);
        Join*   getJoin(UVPt*,UVPt*);
        HMENU   getPopupMenu();
        Tint*   getTintByName(Mystr&);

// implemented in model_h.cpp
        void    hideHigh();
        void    hideMarked();
        void    hideUnmarked();

// implemented in model_i.cpp
        void    incSelection();
        Bone*   indexBone(INDEX i);
        Knob*   indexKnob(INDEX i);
        Norm*   indexNorm(INDEX i);
        Part*   indexPart(INDEX i);
        Tint*   indexTint(INDEX i);
        UVPt*   indexUVPt(INDEX i);
        Vert*   indexVert(INDEX i);
        void    insertCube();
        void    insetHigh();
        bool    isUniqueBoneName(char *,Bone*);
        bool    isUniquePartName(char *,Part*);
        bool    isUniqueTintName(char *,Tint*);

// implemented in model_m.cpp
        void    makeLathe(HWND);
        void    markAll();
        void    markContiguous();
        void    matrixMarked(HWND,short);
        void    mergeVerts(Vert*,Vert*);
        void    mergeAtEdge(Edge*);

// implemented in model_n.cpp
        Bone*   newBone(short,short);
        Edge*   newEdge(Vert*,Vert*);
        Face*   newFace(Vert*,Vert*,Vert*,Vert*);
        Join*   newJoin(UVPt*,UVPt*);
        Knob*   newKnob();
        Norm*   newNorm();
        Part*   newPart(Knob*,Knob*,Knob*);
        Tint*   newTint(COLORREF);
        UVPt*   newUVPt();
        Vert*   newVert(short,short);
        void    nukeBone(Bone*);
        void    nukeEdge(Edge*);
        void    nukeFace(Face*);
        void    nukeJoin(Join*);    // may just use delete
        void    nukeKnob(Knob*);
        void    nukeNorm(Norm*);
        void    nukePart(Part*);
        void    nukeTint(Tint*);
        void    nukeUVPt(UVPt*);
        void    nukeVertex(Vert*);

// implemented in model_o.cpp
        void    orientWindings();

// implemented in model_p.cpp
        void    paintMapBackground();
        void    prepBoneDrag();
        void    prepSingleVertDrag(Vert*);
        void    prepVertDrag();
        void    pushBackHigh();

// implemented in model_r.cpp
        void    readMTLFile(class fReader&);
        void    readOBJFile(class fReader&);
        void    reindexObjects();
        void    removeUnusedUVPts();
//        void    rootCheck();
        void    rotateMarked(HWND,short);

// implemented in model_s.cpp
        void    saveToFile(HANDLE);
        void    saveView();
        void    scaleMarked(HWND,short);
        Vert*   selectVert(ULONG i);
        void    setFineLevel(HWND);
        void    setGridSnap(HWND);
        void    setWindowTitle(HWND);
        void    splitEdge(Edge*,bool);
        void    splitFace(Face*,bool);
        void    splitHigh(bool);
        void    splitMarked(bool);

// implemented in model_t.cpp
        void    toggleAll();
        void    toggleHighlight();
        void    translateMarked(HWND,short);

// implemented in model_u.cpp
        void    unhideAll();
        void    updateMenu();
        void    updateSkeleton(bool);
        void    updateToolBar();

// implemented in model_w.cpp
//      void    writeBONFile(class fWriter&);
        void    writeINCFile(class fWriter&);
        void    writeMTLFile(class fWriter&);
        void    writeOBJFile(class fWriter&,char*);

// implemented in model_v.cpp
        void    validate();
        void    validateMapping();

// implemented in model_z.cpp
        void    zoomToFit();
        void    zoomToSelection();

// implemented in mouse.cpp
        void    leftClick(HWND,WPARAM,LPARAM);
        void    leftDoubleClick(HWND,WPARAM,LPARAM);
        void    leftRelease(HWND,WPARAM,LPARAM);
        void    mouseMove(HWND,WPARAM,LPARAM);
        void    rightClick(HWND,WPARAM,LPARAM);
        void    rightRelease(HWND,WPARAM,LPARAM);

// implemented in scroll.cpp
        void    hScroll(WPARAM);
        void    setScroll();
        void    vScroll(WPARAM);

// implemented in sub.cpp
    void    buildSubData();
    void    calcSubData();
    void    deleteSubData();

// implemented in undo.cpp
        void    destroyAllHistory();
        void    takeSnap();
        void    undoChanges();
        void    redoChanges();
        void    mergeDeleted();
        void    separateDeleted();

};

//model flags
#define MF_SHOW_PARTS   1
#define MF_SHOW_VERTS   2
#define MF_SHOW_EDGES   4
#define MF_SHOW_FACES   8
#define MF_SHOW_SUBD    16
#define MF_BUILD_SUBD  32
#define MF_CALC_SUBD  64
#define MF_REORIENT 128
#define MF_MARK_BY_VERTEX  256
#define MF_MARK_BY_EDGE    512
#define MF_MARK_BY_FACE    1024
#define MF_MARK_MASK    MF_MARK_BY_VERTEX|MF_MARK_BY_EDGE|MF_MARK_BY_FACE
#define MF_REBUILD_NORMS  2048
#define MF_RECALC_NORMS   4096
#define MF_LOCK_X   8192
#define MF_LOCK_Y   16384
#define MF_LOCK_Z   32768
#define MF_LOCK_MASK    (MF_LOCK_X|MF_LOCK_Y|MF_LOCK_Z)

// edit modes
#define EM_MESHEDIT     0
#define EM_MAPEDIT      256

#define FIRSTBONE(m) ((Bone*)(((Model*)m)->Bones.first()))
#define STEPBONES(m,b) for(b=FIRSTBONE(m);b!=NULL;b=(Bone*)b->next())

#define FIRSTEDGE(m) ((Edge*)(((Model*)m)->Edges.first()))
#define STEPEDGES(m,e) for(e=FIRSTEDGE(m);e!=NULL;e=(Edge*)e->next())

#define FIRSTFACE(m) ((Face*)(((Model*)m)->Faces.first()))
#define STEPFACES(m,f) for(f=FIRSTFACE(m);f!=NULL;f=(Face*)f->next())

#define FIRSTJOIN(m) ((Join*)(((Model*)m)->Joins.first()))
#define STEPJOINS(m,j) for(j=FIRSTJOIN(m);j!=NULL;j=(Join*)j->next())

#define FIRSTKNOB(m) ((Knob*)(((Model*)m)->Knobs.first()))
#define STEPKNOBS(m,k) for(k=FIRSTKNOB(m);k!=NULL;k=(Knob*)k->next())

#define FIRSTNORM(m) ((Norm*)(((Model*)m)->Norms.first()))
#define STEPNORMS(m,n) for(n=FIRSTNORM(m);n!=NULL;n=(Norm*)n->next())

#define FIRSTPART(m) ((Part*)(((Model*)m)->Parts.first()))
#define STEPPARTS(m,p) for(p=FIRSTPART(m);p!=NULL;p=(Part*)p->next())

#define FIRSTTINT(m) ((Tint*)(((Model*)m)->Tints.first()))
#define STEPTINTS(m,t) for(t=FIRSTTINT(m);t!=NULL;t=(Tint*)t->next())

#define FIRSTUVPT(m) ((UVPt*)(((Model*)m)->UVPts.first()))
#define STEPUVPTS(m,u) for(u=FIRSTUVPT(m);u!=NULL;u=(UVPt*)u->next())

#define FIRSTVERT(m) (m->Vertices.first())
#define STEPVERTS(m,v) for(v=FIRSTVERT(m);v!=NULL;v=(Vert*)v->next())

#endif
