// model_a.cpp -- implements functions starting with 'a'

#include "userint.h"
#include "model.h"
#include "bone.h"
#include "face.h"
#include "edge.h"
#include "uvpt.h"
#include "vert.h"
#include "mouse.h"
#include "funcs.h"

void Model::activate() {
        activeProject=this;
        updateMenu();
        SetMenu(UserInt::hwndMain,UserInt::ModelMenu);
        setWindowTitle(UserInt::hwndMain);
        updateToolBar();
        InvalidateView();
        clearHighlight();
        clearAll();
}

void Model::addMapping(Face* f,bool /* pull */ ) {
        if(f==NULL) return;

        if(f->data.uvpts==NULL)
        f->data.uvpts=new UVPt*[f->data.sides];

        for(int i=0;i<f->data.sides;i++)
        f->data.uvpts[i]=(UVPt*)0;

//      if(pull)
        for(int i=0;i<f->data.sides;i++)
        f->pullUVData(f->edges[i]);

        for(int i=0;i<f->data.sides;i++)
        if(f->data.uvpts[i]==(UVPt*)0) {
                f->data.uvpts[i]=newUVPt();
                cView.mapXYZ2UV(f->data.verts[i]->data.Spot,f->data.uvpts[i]->data.Spot);
        }

        bool nuke=false;
        for(int i=0;i<f->data.sides;i++)
        if(f->data.uvpts[i]==0) nuke=true;

        if(nuke) {
                UserInt::informUser("Was not able to add texture mapping.");
//                delete [](f->data.uvpts);
                f->data.uvpts=(UVPt**)0;
        } //else {
//        UpdateStatusBar("texture mapping added.",0,0);
//      }
        validateMapping();
}

void Model::applyMarkLogic() {
        MYPTR o;

        switch(MarkType) {
                case HT_VERT:
                        STEPEDGES(this,o.e) {
                                o.e->flags&=~TF_MARKED;
                                o.e->flags|=(o.e->data.VertS->flags & o.e->data.VertE->flags & TF_MARKED);
                        }
                        STEPFACES(this,o.f) {
                                o.f->flags|=TF_MARKED;
                                for(int i=0;i<o.f->data.sides;i++)
                                o.f->flags &= o.f->data.verts[i]->flags | ~TF_MARKED;
                        }
                break;
                case HT_EDGE:
                        // clear vertex marking
                        STEPVERTS(this,o.v) o.v->flags&=~TF_MARKED;
                        // mark both ends of marked edges
                        STEPEDGES(this,o.e)
                        if(o.e->flags&TF_MARKED) {
                                o.e->data.VertS->flags|=TF_MARKED;
                                o.e->data.VertE->flags|=TF_MARKED;
                        }
                        // mark all faces with all edges marked
                        STEPFACES(this,o.f) {
                                o.f->flags|=TF_MARKED;
                                for(int i=0;i<o.f->data.sides;i++)
                                o.f->flags &= (o.f->edges[i]->flags|(~TF_MARKED));
                        }
                break;
                case HT_FACE:
                        // clear vertex marking
                        STEPVERTS(this,o.v) o.v->flags&=~TF_MARKED;

                        // clear edge marking
                        STEPEDGES(this,o.e) o.e->flags&=~TF_MARKED;

                        STEPFACES(this,o.f)
                        if(o.f->flags&TF_MARKED)
                        for(int i=0;i<o.f->data.sides;i++) {
                                o.f->data.verts[i]->flags|=TF_MARKED;
                                o.f->edges[i]->flags|=TF_MARKED;
                        }
                break;
        }
        InvalidateView();
}

// derives ownership of Face objects from Vert objects
// Faces are attached to the same bone as the Vert that is nearest to the
// root.  Used prior to OBJ export.
void Model::assignFaces() {
        Bone*   b;
        Face*   f;

// establish ownership of Faces
        buildBoneHierarchy();

        int high=-1;

        STEPBONES(this,b)
                if(b->gen > high)
                        high=b->gen;

        STEPFACES(this,f)
                f->Owner=(Bone*)0;

        for(int i=high;i>0;i--)
        STEPFACES(this,f) {
                for(int i=0;i<f->data.sides;i++)
                if(f->data.verts[i]->data.Owner)
                if(f->data.verts[i]->data.Owner->gen==i)
                f->Owner=f->data.verts[i]->data.Owner;
        }

        STEPFACES(this,f) {
                for(int i=0;i<f->data.sides;i++)
                if(f->data.verts[i]->data.Owner==(Bone*)0)
                f->Owner=(Bone*)0;
        }

}

// derives Vert ownership from Face ownership
// needed for loading OBJ files
// Verts shared by faces with different Owners go with the Face that is
// farthest from the root

void Model::assignVerts() {
        Vert*   v;
        Face*   f;

        buildBoneHierarchy();

        STEPVERTS(this,v)
        v->data.Owner=NULL;

        STEPFACES(this,f)
        if(f->Owner!=NULL)
        for(int i=0;i<f->data.sides;i++) {
                v=f->data.verts[i];
                if(v->data.Owner==NULL) {
                        v->data.Owner=f->Owner;
                } else {
                        if(v->data.Owner->gen < f->Owner->gen)
                        v->data.Owner=f->Owner;
                }
        }
}

