// model_b.cpp -- member functions of Model that start with 'b'

#include <windows.h> // still needed
#include "model.h"
#include "bone.h"
#include "edge.h"
#include "face.h"
#include "part.h"
#include "tint.h"
#include "vert.h"
#include "appdata.h"

void Model::baseParts() {
        Part* p;

        STEPPARTS(this,p) {
                p->Rise->Owner=p;
                p->End->Owner=p;
        }

        STEPPARTS(this,p)
        p->Parent=p->Base->Owner;
}

void Model::blankSelection() {
        Vert* v;

        STEPVERTS(this,v)
        v->Order=0;
}

void Model::buildBoneComboBox(HWND hwnd,int id,Bone* root) {
        int index;
        Bone*   b;

        // put the default layer at the start of the list
        index=SendDlgItemMessage(hwnd,id,CB_ADDSTRING,0,(LPARAM)"-- not attached --");
        SendDlgItemMessage(hwnd,id,CB_SETITEMDATA,index,(LPARAM)0);
        // and then add the rest

        STEPBONES(this,b)
        if(b->isParent(root)==false) {
                index=SendDlgItemMessage(hwnd,id,CB_ADDSTRING,0,(LPARAM)(b->data.Name.c_string()));
                SendDlgItemMessage(hwnd,id,CB_SETITEMDATA,index,(LPARAM)b);
        }
}

void Model::buildBoneHierarchy() {
        Bone*   b;
        bool    redo;

        // assign generation values to all bones
        // first, clear them all
        STEPBONES(this,b)   b->gen=-1;

        // each root is gen=0, every other is equal to parent's gen+1
        do {
                redo=false;
                STEPBONES(this,b) {
                        if(b->data.Parent==NULL) {
                                b->gen=0;
                        } else {
                                if(b->data.Parent->gen==-1) {
                                        redo=true;
                                } else {
                                        b->gen=b->data.Parent->gen+1;
                                }
                        }
                }
        } while(redo);
}

// this is the function for building the normals, meaning that this one decides
// which normals go to which vertices on each face.
// This would be easier if I wasn't supporting creases...
void Model::buildNorms() {
        Face*   f;
//        Face*   of;
        Norm*   n;
        Norm*   pn;
        Norm*   qn;
        bool redo;

        orientWindings();

// destroy old norms
        STEPFACES(this,f)
        f->destroyNorms();

        n=FIRSTNORM(this);
        while(n) {
                qn=(Norm*)n->next();
                nukeNorm(n);
                n=qn;
        }

// create norms for each corner of each face
        STEPFACES(this,f)
        if(f->norms==NULL) {
                f->norms=new Norm*[f->data.sides];
                for(int i=0;i<f->data.sides;i++)
                f->norms[i]=newNorm();
        }

// and now merge the ones that mees on smooth edges
        STEPFACES(this,f)
        f->mergeNorms();

// now we need to find those norms which are part of a circle that goes around
// forever; making one of them point to NULL will make them terminate properly

// to do this, we get two pointers.  each is updated by making it point to its
// preference.  One is updated twice as fast as the other, and if there is a
// circle, the fast one will equal the slow one in a non-NULL state; if there
// is no circle, the slow will catch the fast one in a NULL state.
        do {
                redo=false;
                STEPNORMS(this,n) {
                        qn=pn=n->prefer;
                        do {
                                if(pn) pn=pn->prefer;
                                if(qn) qn=qn->prefer;
                                if(qn) qn=qn->prefer;
                        } while (pn!=qn);
                        if(pn) {
                                pn->prefer=NULL;
                        }
                }
        } while (redo);


// And now every norm is replaced by its ultimate preference (if there is one)
// in the face-to-normal assignments
        STEPFACES(this,f) {
                for(int i=0;i<f->data.sides;i++) {
                        pn=f->norms[i];
                        if(pn) while(pn->prefer) pn=pn->prefer;
                        f->norms[i]=pn;
                }
        }

        STEPNORMS(this,n)
        n->index=0;

        STEPFACES(this,f)
        for(int i=0;i<f->data.sides;i++)
        f->norms[i]->index=1;   // show that norm is used

        n=FIRSTNORM(this);
        while(n) {
                pn=(Norm*)(n->next());
                if (n->index==0) nukeNorm(n);
                n=pn;
        }

        int i=0;
        STEPNORMS(this,n) n->index=i++;

        modelFlags&=~MF_REBUILD_NORMS;
        modelFlags|=MF_RECALC_NORMS;
}

void Model::buildTintComboBox(HWND hwnd,int id) {
    int index;
    Tint*   t;

    // put the default layer at the start of the list
    index=SendDlgItemMessage(hwnd,id,CB_ADDSTRING,0,(LPARAM)AppData::defTint.data.Name.c_string());
    SendDlgItemMessage(hwnd,id,CB_SETITEMDATA,(WPARAM)index,(LPARAM)&AppData::defTint);
    STEPTINTS(this,t) {
        index=SendDlgItemMessage(hwnd,id,CB_ADDSTRING,0,(LPARAM)t->data.Name.c_string());
        SendDlgItemMessage(hwnd,id,CB_SETITEMDATA,index,(LPARAM)t);
    }
}
