// model_c.cpp -- implements Model member functions starting with 'c'

#include "userint.h"
#include <gl/gl.h>
#include <gl/glu.h>
#include "appdata.h"
#include "fhand.h"
#include "model.h"
#include "bone.h"
#include "part.h"
#include "edge.h"
#include "face.h"
#include "norm.h"
#include "uvpt.h"
#include "vert.h"
#include "dialex.h"
#include "funcs.h"
#include "mouse.h"
#include "lsmtypes.h"
#include "lsmres.h"
#include "obvec.h"
#include "patch.h"

#define SELBUFFSIZE 512
#define MINIBUFFSIZE    4
#define GADBUFFSIZE 64

//extern HINSTANCE hInst;
//extern HWND hwndMain;

void Model::calcNorms() {
        Norm*   n;
        Face*   f;
        Vec3 x;
        Vec3 y;
        Vec3 res;

        STEPNORMS(this,n) n->dir.set(0.0f);

        STEPFACES(this,f)
        if(f->norms) {
            (x = f->data.verts[0]->data.Spot) -= f->data.verts[f->data.sides-1]->data.Spot;
            for (int i=0;i<f->data.sides;i++) {
                (y = f->data.verts[i+1]->data.Spot) -= f->data.verts[i]->data.Spot;
                res.cross(y,x);
                f->norms[i]->dir+=res;
                x=y;
            }
        }

        modelFlags&=~MF_RECALC_NORMS;
}

bool Model::canMergeAtEdge(Edge* e) {
    Edge* te;

    if(e==(Edge*)0) return false;

    if(e->FaceL==NULL || e->FaceR==NULL) return false;

    if(e->FaceL==e->FaceR) return false;

    STEPEDGES(this,te)
    if(te!=e) {
        if((te->FaceL==e->FaceL && te->FaceR==e->FaceR) ||
           (te->FaceL==e->FaceR && te->FaceR==e->FaceL)) return false;
    }
    return true;
}

void Model::checkHighlight(int hi_flags) {
    GLuint *selbuff;
    GLint vp[4];
    short NewType,HitType;
    MYPTR NewItem;
    INDEX i,j;
    MYPTR m;
    UINT hits;
    unsigned int cmin,hmin,cmax,hmax;
    unsigned int skip;
    char buff[128];

// make space for the select buffer
    selbuff=new GLuint[SELBUFFSIZE];

    if(selbuff==NULL) {
        UserInt::updateStatusBar("Selection buffer allocation failed.",0,0);
        return;
    }

// clear out or preserve current highlight, depending on what's needed
    if(hi_flags&HT_CLEAR) { // discard if not the current item
        NewType=HT_NOTHING;
        NewItem.e=NULL;
    } else { // discard only if replaced by another
        NewType=HighType;
        NewItem.e=HighItem.e;
    }

// set up select mode
// prepare the selection buffer
    glSelectBuffer(SELBUFFSIZE,selbuff);

// ensure that all objects are drawn correctly
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

// get the viewport
    glGetIntegerv(GL_VIEWPORT,vp);

// now go into projection mode
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
// set up clipping region
    glLoadIdentity();
    gluPickMatrix((GLdouble)lX,(GLdouble)(vp[3]+vp[1]-lY+cView.Screen.top),(GLdouble)7,(GLdouble)7,vp);
    cView.loadIntoOpenGL(vp[2],vp[3]);

// go into selection mode
    glRenderMode(GL_SELECT);

// prepare name stack
    glInitNames();
    glPushName((GLuint)0);

// search faces
    if(((hi_flags&HT_FACE)!=0) && ((modelFlags&MF_SHOW_FACES)!=0)) {
        glDisable(GL_CULL_FACE);
        glLoadName((GLuint)HT_FACE);
        glPushName((GLuint)FIRSTFACE(this));
        if(modelFlags&MF_SHOW_SUBD) {
            for(Patch* p=patches;p;p=p->next) {
                m.f=p->f;
                glLoadName((GLuint)m.f);
                p->draw(subPreviewLevel,false);
            }
        } else {
            STEPFACES(this,m.f)
            if((m.f->flags&TF_HIDDEN)==0) {
                glLoadName((GLuint)m.f);
// render the face
                glBegin(GL_POLYGON);
                for(int i=0;i<m.f->data.sides;i++)
                m.f->data.verts[i]->data.Spot.glPlot();
                glEnd();
            }
        }

        glPopName();
    }

// search edges
    if(((hi_flags&HT_EDGE)!=0) && ((modelFlags&MF_SHOW_EDGES)!=0)) {
        int subd=((modelFlags&MF_SHOW_SUBD)==0)?0:subPreviewLevel;
        glLoadName((GLuint)HT_EDGE);
        glPushName((GLuint)0);
        STEPEDGES(this,m.e)
//              if((m.e->flags&TF_INVALID)==0)
        if((m.e->flags&TF_HIDDEN)==0) {
            glLoadName((GLuint)m.e);
// render the edge
            m.e->trace(subd);
        }
        glPopName();
    }

    bool showsubd=modelFlags&MF_SHOW_SUBD;
// search vertices
    if(((hi_flags&HT_VERT)!=0) && ((modelFlags&MF_SHOW_VERTS)!=0)) {
        glLoadName((GLuint)HT_VERT);
        glPushName((GLuint)0);
        STEPVERTS(this,m.v)
//      if((m.v->flags&TF_INVALID)==0)
        if((m.v->flags&TF_HIDDEN)==0)
        if(m.v!=drag.v) {
            glLoadName((GLuint)m.v);
// render the vertex
            glBegin(GL_POINTS);
            m.v->draw(showsubd);
            glEnd();
        }
        glPopName();
    }

    if((hi_flags&HT_UVPT)!=0) {
        glLoadName((GLuint)HT_UVPT);
        glPushName((GLuint)0);
        STEPUVPTS(this,m.u)
        if(m.u->flags&TF_NEW)
        if(m.u!=drag.u) {
            glLoadName((GLuint)m.u);
// render the uv point
            glBegin(GL_POINTS);
            m.u->data.Spot.glPlot();
            glEnd();
        }
        glPopName();
    }

// select the bones
        if(((hi_flags&HT_BONE)!=0) && ((modelFlags&MF_SHOW_PARTS)!=0)) {
            glLoadName((GLuint)HT_BONE);
            glPushName((GLuint)0);
            STEPBONES(this,m.b)
            if(m.b!=drag.b) {
                glLoadName((GLuint)m.b);
                m.b->selectBase(getBoneLength(*m.b));
            }
            glPopName();

            glLoadName((GLuint)HT_KNOB);
            glPushName((GLuint)0);
            STEPBONES(this,m.b)
            if(m.b!=drag.b) {
                glLoadName((GLuint)m.b);
                m.b->selectEnd(getBoneLength(*m.b));
            }
            glPopName();
        }

        glPopName();

        hits=glRenderMode(GL_RENDER);

        hmin=0xffffffff;
        hmax=0xffffffff;

        j=0;
        HitType=HT_NOTHING;
        while(hits>0 && j<SELBUFFSIZE) {
            skip=selbuff[j++];      // get count of names
            cmin=selbuff[j++];      // get shallow end of hit
            cmax=selbuff[j++];  // get deep end of hit
            while(skip) {
                if((selbuff[j]>(UINT)HitType && cmin<=hmax && cmax>=hmin) // overlapped by higher priority item
                    || cmax<hmin ) { // closer item found
                    HitType=(short)selbuff[j];
                    i=selbuff[j+1];
                    hmin=cmin;
                    hmax=cmax;
                }
                j+=2;
                skip-=2;
            }
            hits--;
        }

        if(HitType!=HT_NOTHING) {
            NewItem.u=(UVPt*)i;
            NewType=HitType;
            if(NewItem.u!=HighItem.u || NewType!=HighType) InvalidateView();
        }

        HighType=NewType;
        HighItem.e=NewItem.e;
        if(HighItem.e==NULL) HighType=HT_NOTHING;

// restore old viewing matrix
        glMatrixMode(GL_PROJECTION);
        glPopMatrix();

        switch(HighType) {
        case HT_BONE:
                if(HighItem.b->data.Name.len()) {
                        sprintf(buff,"Bone: %s",HighItem.b->data.Name.c_string());
                        UserInt::updateStatusBar(buff,0,0);
                }
        break;
        case HT_VERT:
                // do nothing
        break;
        case HT_EDGE:
                if(HighItem.e->sharpType()) {
                        UserInt::updateStatusBar("This edge is sharp",0,0);
                } else {
                        UserInt::updateStatusBar("",0,0);
                }
        break;
//        case HT_FACE:
//            if(HighItem.f==NULL)
//            UserInt::updateStatusBar("NULL returned, odd...",0,0);
//            break;
        default:
            UserInt::updateStatusBar("",0,0);
            break;
        }

        delete []selbuff;
}

short Model::checkGadget() {
    Array<GLuint> selbuff(GADBUFFSIZE);
    GLint vp[4];
    short retval=MM_GADNULL;

    if(!selbuff) {
        UserInt::updateStatusBar("Selection buffer creation failed.",0,0);
        return retval;
    }

// ensure that all objects are drawn correctly
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

// set up select mode
// prepare the selection buffer
    glSelectBuffer(GADBUFFSIZE,selbuff);

// get the viewport
    glGetIntegerv(GL_VIEWPORT,vp);

// now go into projection mode
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
// set up clipping region
    glLoadIdentity();
    gluPickMatrix((GLdouble)lX,(GLdouble)(vp[3]+vp[1]-lY+cView.Screen.top),(GLdouble)5,(GLdouble)5,vp);
    cView.loadIntoOpenGL(vp[2],vp[3]);

// go into selection mode
    glRenderMode(GL_SELECT);
// prepare name stack
    glInitNames();
    glPushName((GLuint)MM_GADNULL);
// render the gadget
    if(gadget) gadget->select();

    int hits=glRenderMode(GL_RENDER);

    if(hits) {
        int skip;
        unsigned int cmin,cmax;
        unsigned int hmin=0xffffffff;
        unsigned int hmax=0xffffffff;

        int j=0;
        while(hits>0 && j<GADBUFFSIZE) {
            skip=selbuff[j++];  // get count of names
            cmin=selbuff[j++];  // get shallow end of hit
            cmax=selbuff[j++];  // get deep end of hit
            while(skip) {
                if(cmin < hmin ) { // closer item found
                    retval=(short)selbuff[j];
                    hmin=cmin;
                    hmax=cmax;
                }
                j++;
                skip--;
            }
            hits--;
        }
    }

// done with the name stack
    glPopName();

// restore old viewing matrix
    glMatrixMode(GL_PROJECTION);
    glPopMatrix();

    return retval;
}

void Model::checkSelectBox(RECT *box,short sel_flags) {
        GLuint *selbuff;
        GLint vp[4];
        MYPTR m;
        bool showsubd=modelFlags&MF_SHOW_SUBD;

// make space for the select buffer
        selbuff=new GLuint[MINIBUFFSIZE];

        if(selbuff==NULL) {
                UserInt::updateStatusBar("Selection buffer creation failed.",0,0);
                return;
        }

// set up select mode
// prepare the selection buffer
        glSelectBuffer(MINIBUFFSIZE,selbuff);

// get the viewport
        glGetIntegerv(GL_VIEWPORT,vp);

// now go into projection mode
        glMatrixMode(GL_PROJECTION);
        glPushMatrix();
// set up clipping region
        glLoadIdentity();
        gluPickMatrix(
                ((GLdouble)box->left+(GLdouble)box->right)*.5,
                (GLdouble)vp[3]+(GLdouble)cView.Screen.top+(GLdouble)vp[1]-((GLdouble)box->bottom+(GLdouble)box->top)*.5,
                (GLdouble)(box->right-box->left),
                (GLdouble)(box->bottom-box->top),
                vp);
        cView.loadIntoOpenGL(vp[2],vp[3]);

// search vertices
        if(((sel_flags&HT_VERT)!=0) && ((modelFlags&MF_SHOW_VERTS)!=0)) {
                STEPVERTS(this,m.v)
//              if((m.v->flags&TF_INVALID)==0)
//                if((m.v->flags&TF_HIDDEN)==0)
                if(m.v!=drag.v) {
                    if(m.v->flags&TF_DRAGGED) {
                        m.v->flags|=TF_MARKED;
                    } else {
                        m.v->flags&=~TF_MARKED;
                    }
// go into selection mode
                    glRenderMode(GL_SELECT);
// prepare name stack
//                      glInitNames();
// used by all objects
//                      glPushName((GLuint)0);
// render the vertex
            glBegin(GL_POINTS);
            m.v->draw(showsubd);
            glEnd();
//      done
//                      glPopName();
                    if(glRenderMode(GL_RENDER)) m.v->flags^=TF_MARKED;
                }
        }
/*
        if((flags&HT_UVPT)!=0) {
                STEPUVPTS(this,m.u)
                if(m.u!=drag.u) {
                        glLoadName((GLuint)m.u);
// render the uv point
            glBegin(GL_POINTS);
                        glVertex2fv((GLfloat *)&m.u->Spot);
                        glEnd();
                }
        }
*/
        glMatrixMode(GL_PROJECTION);
        glPopMatrix();
        delete [](selbuff);

        MarkType=HT_VERT;
        applyMarkLogic();
}

void Model::clearAll() {
        MYPTR m;

        STEPVERTS(this,m.v)
        m.v->flags&=~TF_MARKED;

        STEPEDGES(this,m.e)
        m.e->flags&=~TF_MARKED;

        STEPFACES(this,m.f)
        m.f->flags&=~TF_MARKED;

        STEPBONES(this,m.b)
        m.b->flags&=~TF_MARKED;

        MarkType=HT_NOTHING;

        applyMarkLogic();
        InvalidateView();
}

void Model::clearAllDragging() {
        MYPTR mp;

        STEPVERTS(this,mp.v) mp.v->flags&=~TF_DRAGGED;
}

void Model::clearHighlight() {
        HighItem.b=(Bone*)0;
        HighType=HT_NOTHING;
}

void Model::convertPartModel() {
        Part*   p;
        Bone*   b;
        Bone*   e;

UserInt::informUser("Converting the old skeleton to the new type.");

        baseParts();

        STEPPARTS(this,p) {
                b=p->BaseBone;
                e=p->EndBone;

                b->data.Name=p->Name;
                b->data.Name+="_Base";

                e->data.Name=p->Name;
                e->data.Name+="_End";
        }

//UserInt::informUser("Bone names built");

// establish parentage
        STEPPARTS(this,p) {
// Parent of end Bone is base Bone
                p->EndBone->data.Parent=p->BaseBone;
// Parent of base Bone is End Bone of Parent, if one
                if(p->Parent)
                p->BaseBone->data.Parent=p->Parent->EndBone;
        }
// now all parentage is established

//UserInt::informUser("Parentage established");

// establish Spot of new Bones
        STEPPARTS(this,p) {
// Base Bone gets Base Knob Spot
            p->BaseBone->Spot=p->Base->Spot;

// End Bone gets halfway between Base Knob and End Knob
            ((p->EndBone->Spot=p->Base->Spot)+=p->End->Spot)*=.5;
        }

//UserInt::informUser("Bones spotted.");

// now we calculate the forward matrices for the new Bones
        Vec3 vZ;
        Vec3 vX;
        Vec3 vY;

        STEPPARTS(this,p) {
// The forward vector (Z row) is the same for both
// It is parallel to the vector from the Base Knob to the End Knob
        (vZ=p->End->Spot)-=p->Base->Spot;
        vZ.normalize();

        p->BaseBone->notScaled.fore.z=
        p->EndBone->notScaled.fore.z=vZ;

// the next two vectors apply only to the End Bone

// the right vector (X row) is perp to both the Z-row and the vector
// from the Base Knob to the Rise Knob
        (vY=p->Rise->Spot)-=p->Base->Spot;

        vX.cross(vY,vZ);
        vX.normalize();

        p->EndBone->notScaled.fore.x=vX;

// the up vector (Y row) is perp to the X-row and the Z-row
        vY.cross(vZ,vX);
        vY.normalize(); // shouldn't be necessary
        p->EndBone->notScaled.fore.y=vY;

// and get the final translate point in there as well
        p->EndBone->notScaled.fore.l=p->EndBone->Spot;
    }

//UserInt::informUser("End matrices calculated.");

    Matrix t;
    float fX,fY,fZ;

// the forward matrices for the End Bones are done
// the the forward vector of the Base Bones is done
//
// now for the right and up vectors for the Base Bones
//
// this is easy; we calculate the right, and then cross with the
// forward to get the up vector
    STEPPARTS(this,p) {

// get the inherited matrix
        if(p->Parent) {
            t=p->Parent->EndBone->notScaled.fore;
        } else {
            t.identity();
        }

        switch(p->flags&PT_MASK) {
// hinge joints bend only along the right axis, so the right axis is always
// the same as the one inherited from the parent
            case PT_HINGE:
                vX=t.x;
                break;

            case PT_UNI:
// for uni joints, the right axis is always perp to both the inherited up axis
// and the new forward axis, therefore right is the cross product of old up and
// new fore

                vX.cross(t.y,p->BaseBone->notScaled.fore.z);
// if inherited fore and current fore are parallel or anti-parallel,
// the new right vector is the same as the old
                if(vX.normalize()==0.0) {
                    vX=t.x;
                }
                break;
            case PT_BALL:
// get cosine of bending angle
                fZ=t.z.dot(p->BaseBone->notScaled.fore.z);
                if(fZ==1.0) {
                    vX=t.x;
                }
                else if(fZ==-1.0) {
                    (vX=t.x)*=-1.0f;
                }
                else {
                    Vec3 axisU,axisV;
// cross old fore with new fore to get U axis of bending
                    axisU.cross(t.z,p->BaseBone->notScaled.fore.z);
                    axisU.normalize();

// cross U axis with old fore to get old V axis
                    axisV.cross(axisU,t.z);
                    axisV.normalize(); // probably not needed

// get dot product of old X with U and V
                    fX=t.x.dot(axisU);
                    fY=t.x.dot(axisV);

// cross U axis with new fore to get new V axis
                    axisV.cross(axisU,p->BaseBone->notScaled.fore.z);
                    axisV.normalize(); // probably not needed

                    
// combine U axis with new V to get the new X axis
                    axisV*=fY;
                    ((vX=axisU)*=fX)+=axisV;
                }
                break;
        } // end of switch

// now cross the base forward with the base right to get the new base up
        p->BaseBone->notScaled.fore.y.cross(p->BaseBone->notScaled.fore.z,vX);
// and put the new right vector in place
        p->BaseBone->notScaled.fore.x=vX;

// and also put the location vector in there
        p->BaseBone->notScaled.fore.l=p->BaseBone->Spot;
    }

//UserInt::informUser("Base matrices calculated.");

// now the notScaled transforms have all of their info for the forward transform
    STEPPARTS(this,p) {
// tell each transform to build the inverse from the foreward
        p->BaseBone->notScaled.buildInverse();
        p->EndBone->notScaled.buildInverse();
// and copy the info to the Scaled transforms as well
        p->BaseBone->Scaled=p->BaseBone->notScaled;
        p->EndBone->Scaled=p->EndBone->notScaled;
    }

//UserInt::informUser("Inverses calculated.");

// and now force calculation of the rotation angles and the translate values
    STEPBONES(this,b)
    b->reparent();

//UserInt::informUser("Reparenting done.");

// now destroy all of the Part objects
    while(Parts.first()) nukePart((Part*)(Parts.first()));

    while(Knobs.first()) nukeKnob((Knob*)(Knobs.first()));

//        UserInt::informUser("Conversion complete.");
}

Face* Model::copyFace(Face* of) {
        Face* nf;

        nf = new Face;
        if(nf!=NULL) {
                Faces.addTail(nf);
//              flags|=MF_REBUILD_NORMS;   // force rebuild of normals
                // establish basic geometry
                nf->data.sides=of->data.sides;
                nf->data.verts=new Vert*[nf->data.sides+1];
                for(int i=0;i<=nf->data.sides;i++)
                nf->data.verts[i]=of->data.verts[i]->Copy;

                // set other data to defaults
                if(of->data.uvpts) {
                        nf->data.uvpts=new UVPt*[nf->data.sides];
                        for(int i=0;i<nf->data.sides;i++)
                        nf->data.uvpts[i]=copyUVPt(of->data.uvpts[i]);
                } else {
                        nf->data.uvpts=(UVPt**)0;
                }

                nf->edges=new Edge*[nf->data.sides];
                nf->data.Texture=of->data.Texture;
                nf->flags=0;
                nf->index=NextFaceIndex++;
//              nf->Extra=NULL;

                for(int i=0;i<nf->data.sides;i++)
                nf->edges[i]=getEdge(nf->data.verts[i],nf->data.verts[i+1]);
        }
        return nf;
}

void Model::copyMarked() {
        Vert *v,*nv;
        Edge *e,*ne;
        Face *f,*nf;

// copy vertices
        STEPVERTS(this,v)
        if(v->flags&TF_MARKED) {
                v->Copy=nv=newVert(0,0);
                nv->data.Spot=v->data.Spot;
                nv->flags=TF_NEW;
        } else {
                v->Copy=NULL;
                v->flags&=~TF_NEW;
        }

// copy marked edges
        STEPEDGES(this,e)
        if(e->flags & TF_MARKED) {
                ne=newEdge(e->data.VertS->Copy,e->data.VertE->Copy);
                ne->setSharpLevel(e->getSharpLevel());
        }

// copy marked faces
        STEPFACES(this,f) {
                if(f->flags & TF_MARKED)
                nf=copyFace(f);
        }

        STEPVERTS(this,v) v->flags&=~TF_MARKED;

        STEPVERTS(this,v)
        if(v->Copy)
        v->Copy->flags|=TF_MARKED;

        applyMarkLogic();

        validate();
        validateMapping();
        takeSnap();
        modelFlags|=MF_BUILD_SUBD;
        InvalidateView();
}

UVPt* Model::copyUVPt(UVPt*u) {
        UVPt *c;

        if(u==NULL) return NULL;

        c=newUVPt();
        if(c) c->data=u->data;

        return c;
}

char *Model::createBoneName() {
        char tbuff[20];
        char *name;
        int i;

        for(i=1,name=NULL;name==NULL;i++) {
                sprintf(tbuff,"Bone%d",i);
                if(isUniqueBoneName(tbuff,(Bone*)NULL)) name=CloneString(tbuff);
        }
        return name;
}

Model*  Model::createModel(HWND hwnd) {
        Model*  model;
        Dialex *dx;
        char*   newName;
        fReader file;

        dx=new Dialex;

        newName=createModelName();
        model=(Model*)0;

        dx->setDialogString(EB_NAME,newName);
        if(newName) { delete []newName; newName=(char*)0; }
        dx->setDialogString(EB_MESHFILE,(char*)0);
        dx->setDialogString(EB_TEXTUREFILE,(char*)0);
//      dx->setDialogString(EB_SKELETONFILE,(char*)0);
        if(DialogBoxParam(UserInt::hInst,MAKEINTRESOURCE(DLG_NEWMODEL),hwnd,
                CreateModelDlg,(LPARAM)dx)==IDOK) {
                model=new Model;
                if(model) {
                        model->name.own(dx->getDialogString(EB_NAME));

// load the skeleton file
//                      newName=dx->getDialogString(EB_SKELETONFILE);
//                      if(newName) {
//                              file.open(newName);
//                              if(file) {
//                                      model->readBONFile(file);
//                              } else {
//                                      UserInt::informUser("Could not open skeleton file");
//                              }
//                              delete []newName;
//                              newName=(char*)0;
//                      }

// load the mtl file
                        newName=dx->getDialogString(EB_TEXTUREFILE);
                        if(newName) {
                                file.open(newName);
                                if(file) {
                                        model->readMTLFile(file);
                                } else {
                                        UserInt::informUser("Could not open texture file");
                                }
                                delete[]newName;
                                newName=(char*)0;
                        }

// load the mesh file
                        newName=dx->getDialogString(EB_MESHFILE);
                        if(newName) {
                                file.open(newName);
                                if(file) {
                                        model->readOBJFile(file);
                                } else {
                                        UserInt::informUser("Could not open mesh file.");
                                }
                                delete []newName;
                                newName=(char*)0;
                        }
                        model->projectFlags|=PF_NOTDEFAULTNAME;
                        model->activate();
                        Project::redoDisplay(model);
                        model->zoomToFit();
                }
        }
        delete dx;

        return model;
}

char* Model::createModelName() {
    char tbuff[20];
    char *newName;
    int i;

    for(i=1,newName=NULL;newName==NULL;i++) {
        sprintf(tbuff,"Model%d",i);
        if(Project::isUniqueProjectName(tbuff)) newName=CloneString(tbuff);
    }
    return newName;
}

char* Model::createPartName() {
        char tbuff[20];
        char *name;
        int i;

        for(i=1,name=NULL;name==NULL;i++) {
                sprintf(tbuff,"Part%d",i);
                if(isUniquePartName(tbuff,(Part*)NULL)) name=CloneString(tbuff);
        }
        return name;
}

char* Model::createTintName() {
    char tbuff[20];
    char *name;
    int i;

    for(i=1,name=NULL;name==NULL;i++) {
        sprintf(tbuff,"Texture%d",i);
        if(isUniqueTintName(tbuff,NULL)) name=CloneString(tbuff);
    }
    return name;
}

char* Model::createRootName() {
    char tbuff[20];
    char root[]="Root";
    char*   name;
    int i;

    if(isUniqueBoneName(root,NULL)) {
        name=CloneString(root);
    } else {
        for(i=1,name=NULL;name==NULL;i++) {
            sprintf(tbuff,"Root%d",i);
            if(isUniqueBoneName(tbuff,NULL)) name=CloneString(tbuff);
        }
    }
    return name;
}

