// model_d.cpp -- implements Model member functions starting with 'd'

#include <windows.h>
#include <math.h>
#include <gl/glu.h>
//#include <gl/glext.h>
#include "userint.h"
#include "appdata.h"
#include "bound.h"
#include "model.h"
#include "bone.h"
#include "edge.h"
#include "face.h"
#include "vert.h"
#include "uvpt.h"
#include "funcs.h"
#include "mouse.h"
#include "pen.h"
#include "calc.h"
#include "patch.h"

// extern HWND AppData::hwndMain;

//const GLubyte markk[18]= { 0x08,0x00, 0x14,0x00, 0x22,0x00, 0x41,0x00,
//                0x80,0x80, 0x41,0x00, 0x22,0x00, 0x14,0x00, 0x08,0x00,};
//const GLubyte kbox[5]= { 0x20,0x50,0x88,0x50,0x20 };

void Model::decSelection() {
        Vert* v;

        STEPVERTS(this,v)
        if(v->Order)
        v->Order--;
}

// this is for edit deletes during the project's existence
void Model::deleteBone(Bone* b) {
        bool areDependents=false;

        Vert*   v;

        STEPVERTS(this,v)
        if(v->data.Owner==b)
        areDependents=true;

        Bone*   db;
        STEPBONES(this,db)
        if(db->data.Parent==b)
        areDependents=true;

        Bone*   pb=b->data.Parent;

        if(areDependents) {
            int answer;
            if(pb) { // has parent
                answer=MessageBoxEx(UserInt::hwndMain,
                    "This Bone has other objects attached to it.  Attach them to this Bone's parent?",
                    "Bone deletion",MB_YESNOCANCEL,LANG_NEUTRAL);
            } else { // didn't have parent
                answer=MessageBoxEx(UserInt::hwndMain,
                    "This Bone has other objects attached to it.  Deleting it will make them unattached.",
                    "Bone deletion",MB_OKCANCEL,LANG_NEUTRAL);
            }

            if(answer==IDCANCEL) return;
            if(answer==IDNO) pb=(Bone*)0;

            STEPVERTS(this,v)
            if(v->data.Owner==b)
            v->data.Owner=pb;

            STEPBONES(this,db)
            if(db->data.Parent==b) {
                db->data.Parent=pb;
                db->reparent();
            }
        }

        b->remove();
        delBones.addTail(b);
        b->flags|=TF_INVALID;

        if(HighItem.b==b) clearHighlight();
}

void    Model::deleteEdge(Edge* e) {
        Face*   f;

        e->flags|=TF_INVALID;

        f=FIRSTFACE(this);
        while(f) {
                Face* nf=(Face*)f->next();

                if(f->edges) {
                    bool kill=false;

                    for(int i=0;i<f->data.sides;i++)
                        if(f->edges[i]==e)
                            kill=true;

                    if(kill) deleteFace(f);
                }
                f=nf;
        }

        if(HighItem.e==e) clearHighlight();

        e->data.VertS->remEdge(e);
        e->data.VertE->remEdge(e);

//        e->subd=0;

        e->remove();
        delEdges.addTail(e);
        modelFlags|=MF_BUILD_SUBD;
}

void Model::deleteEdges() {
        Edge* e;

        e=FIRSTEDGE(this);
        while(e) {
            Edge* ne=(Edge*)e->next();
            if (e->flags & TF_MARKED) deleteEdge(e);
            e=ne;
        }
        validate();
        InvalidateView();
        validateMapping();
        takeSnap();
}

void Model::deleteFace(Face* f) {
//        Edge* e;

        if(f==NULL) return;

        f->flags|=TF_INVALID;

        f->norms=(Norm**)0;

        if(HighItem.f==f) clearHighlight();

        if(f->edges)
        for(int i=0;i<f->data.sides;i++) {
            f->edges[i]->replaceFace(f,NULL);
        }

        f->removeEdgeXRef();

//        f->subd=0;
//        f->usubd=0;

        f->remove();
        delFaces.addTail(f);

        modelFlags|=MF_BUILD_SUBD;
}

void Model::deleteFaces() {
        Face *f,*nf;

        f=FIRSTFACE(this);
        while(f) {
                nf=(Face*)f->next();
                if(f->flags & TF_MARKED) deleteFace(f);
                f=nf;
        }
        validate();
        validateMapping();
        InvalidateView();
        takeSnap();
}

void Model::deleteHighlight() {
    switch(HighType) {
        case HT_VERT:
            deleteVert(HighItem.v);
        break;
        case HT_BONE:
        case HT_KNOB:
            deleteBone(HighItem.b);
        break;
        case HT_EDGE:
            deleteEdge(HighItem.e);
        break;
        case HT_FACE:
            deleteFace(HighItem.f);
        break;
    }
    HighItem.e=NULL;
    HighType=HT_NOTHING;
    validate();
    validateMapping();
    InvalidateView();
    takeSnap();
}

void Model::deleteTint(Tint* t) {
    Face* f;

    if(t==NULL) return;

    STEPFACES(this,f)
    if(f->data.Texture==t) {
        f->data.Texture=&AppData::defTint;
    }

    if(t==CurTint) CurTint=&AppData::defTint;

    t->remove();
    delTints.addTail(t);
    t->flags|=TF_INVALID;
}

void Model::deleteUVPt(UVPt* u) {
    Face *f;
    bool kill;

    STEPFACES(this,f) {
        kill=false;
        if(f->data.uvpts)
        for(int i=0;i<f->data.sides;i++)
        if(f->data.uvpts[i]==u) kill=true;

        if(kill) {
//            delete []f->data.uvpts;
            f->data.uvpts=(UVPt**)NULL;
        }
    }

    u->remove();
    delUVPts.addTail(u);
    u->flags|=TF_INVALID;
}

void Model::deleteVert(Vert* v) {
    Edge *e,*ne;
    Face *f,*nf;
    bool kill;

    if(v==NULL) {
        UserInt::informUser("NULL pointer passed to DeleteVertex()");
        return;
    }

    v->flags|=TF_INVALID;

    f=FIRSTFACE(this);
    while(f) {
        nf=(Face*)(f->next());
        kill=false;
        for(int i=0;i<f->data.sides;i++)
        if(f->data.verts[i]==v)kill=true;
        if(kill) deleteFace(f);
        f=nf;
    }

    e=FIRSTEDGE(this);
    while(e) {
        ne=(Edge*)(e->next());
        if(e->data.VertS==v || e->data.VertE==v)
        deleteEdge(e);
        e=ne;
    }

    if(HighItem.v==v) clearHighlight();

//    v->subd=0;

// put the vertex into the deleted vertex chain
    v->remove();
    delVerts.addTail(v);

    modelFlags|=MF_BUILD_SUBD;
}

void Model::deleteVerts() {
        Vert* v,*nv;

        v=FIRSTVERT(this);
        while(v) {
                nv=(Vert*)(v->next());
                if (v->flags&TF_MARKED)
                deleteVert(v);
                v=nv;
        }

        validate();
        validateMapping();
        InvalidateView();
        takeSnap();
}

// function to move a Bone's end and everything attached to it
void Model::dragBone(Bone *b,Vec3& dVec) {
        Bone *db;
        Vert* v;

// move Bone location
        (b->Spot=b->Save)+=dVec;

// set translate value to represent new location
        b->data.translate=b->Spot;
        if(b->data.Parent) {
            b->data.translate /= b->data.Parent->Scaled;
        }

// rebuild matrices
        b->buildTransforms();

// drag the dependent vertices
        STEPVERTS(this,v)
        if(v->data.Owner==b) {
            (v->data.Spot=v->Save)+=dVec;
        }
// force recalculation of the subdivision preview
        modelFlags|=MF_CALC_SUBD;

// drag the dependent bones
        STEPBONES(this,db)
        if(db->data.Parent==b)
        dragBone(db,dVec);
}

void Model::draw() {
    COLORREF cr;
        switch(EditMode) {
            case EM_MESHEDIT:
                if(projectFlags&PF_MODIFIED) validate();

// turn on hidden object removal
                glEnable(GL_DEPTH_TEST);

// and render them
// was here
                glEnable(GL_LIGHTING);
                glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,GL_TRUE);

                if(modelFlags&MF_SHOW_SUBD) {
                    buildSubData();
                    calcSubData();
                }

                if(modelFlags&MF_SHOW_FACES)
                if(modelFlags&MF_SHOW_SUBD) {
                    drawSubdivision();
                } else {
                    drawFaces();
                }

                if(modelFlags&MF_SHOW_PARTS) drawBones();

                glDisable(GL_LIGHTING);
                glDisable(GL_LIGHTING);
                if(modelFlags&MF_SHOW_VERTS) drawVertices();

                if(modelFlags&MF_SHOW_EDGES) drawEdges();

                if(projectFlags&PF_SHOWAXES) cView.drawAxes(true);

                glLoadIdentity();

                drawMarking();
                drawHighlight();
                break;
            case EM_MAPEDIT:
                glDisable(GL_DEPTH_TEST);
// clear the screen
                glClear(GL_DEPTH_BUFFER_BIT);
                paintMapBackground();
                if(projectFlags&PF_SHOWAXES) cView.drawAxes(false);
                cr=CurTint->data.color;
// big dots in opposite color
                glDisable(GL_DEPTH_TEST);
                glEnable(GL_LINE_STIPPLE);
                glColor3d(floor(1.5-GetRValue(cr)/255.0),floor(1.5-GetGValue(cr)/255.0),
                    floor(1.5-GetBValue(cr)/255.0));
                glLineWidth(1.0);
                glPointSize(5.0);
                glLineStipple(1,0x4444);
                drawUVEdges();
                drawUVPts();
// small lines and matching color
                glColor3d(floor(GetRValue(cr)/255.0+.5),floor(GetGValue(cr)/255.0+.5),
                    floor(GetBValue(cr)/255.0+.5));
                glPointSize(3.0);
                glLineStipple(1,0x1111);
                drawUVEdges();
                drawUVPts();
                drawHighlight();
                glDisable(GL_LINE_STIPPLE);
                break;
        }
        if(gadget) gadget->draw();
}

void Model::drawBones() {
    Bone* b;

    glLineWidth(1.0);
    STEPBONES(this,b) {
        glEnable(GL_LIGHTING);
        glShadeModel(GL_SMOOTH);
//        if(b==defBone) {
//            glColor3f(.75,.95,.85);
//        } else
        if(HighType==HT_BONE || HighType==HT_KNOB) {
            if(b==HighItem.b) {
                glColor3f(.95,.95,.55);
            }
            else if(b->isParent(HighItem.b)) {
                glColor3f(.75,.85,.95);
            }
            else if(HighItem.b->isParent(b)) {
                glColor3f(.8,.8,.8);
            }
            else {
                glColor3f(.6,.6,.6);
            }
        } else {
            glColor3f(.75,.75,.75);
        }
        b->draw(getBoneLength(*b));
        if(b->data.Parent) {
            glLoadIdentity();
            glDisable(GL_LIGHTING);
            glShadeModel(GL_FLAT);
            glEnable(GL_LINE_STIPPLE);
            glLineStipple(1,0xacf0);
            glColor3f(1.0,1.0,1.0);
            glBegin(GL_LINES);
            b->Spot.glPlot();
            b->data.Parent->Spot.glPlot();
            glEnd();

            glLineStipple(1,0x530f);
            glColor3f(0.0,0.0,0.0);
            glBegin(GL_LINES);
            b->Spot.glPlot();
            b->data.Parent->Spot.glPlot();
            glEnd();
        }
    }
    glLoadIdentity();
}

void Model::drawEdges(void) {
    Edge *e;
    int subd=((modelFlags&MF_SHOW_SUBD)==0)?0:subPreviewLevel;

    SelectGLColor(AppData::PenColor[EDGEPEN]);

    STEPEDGES(this,e)
    if((e->flags&TF_HIDDEN)==0)
    e->draw(subd);
}

void Model::drawFaces() {
    Face* f;
    Tint* t;

    Vec3 norm;
//        GLfloat cent[3];
    GLfloat minu,minv,q;

    glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
    glEnable(GL_POLYGON_OFFSET_FILL);
    glPolygonOffset(.5,.5);
    glShadeModel(GL_FLAT);

// draw faces
//      if(modelFlags&MF_REBUILD_NORMS)  buildNorms();
//      if(modelFlags&MF_RECALC_NORMS)   calcNorms();
    STEPFACES(this,f)
    if((f->flags&TF_HIDDEN)==0) {
        f->calcNormal(norm);
        norm.normalize();
//        norm*=-1.0f;
        norm.glNormal();

        t=f->data.Texture;
        if(t==NULL) t=&AppData::defTint;
        if(f->data.uvpts==NULL) {
            SelectGLColor(t->data.color);
            glDisable(GL_TEXTURE_2D);
            glBegin(GL_POLYGON);
            for(int i=0;i<=f->data.sides;i++) {
                f->data.verts[i]->data.Spot.glPlot();
            }
            glEnd();
        } else {
            if(t->bmBits==NULL) t->getTintBitmap();
            glEnable(GL_TEXTURE_2D);
            SelectGLColor((COLORREF)0xffffff);
            t->selectIntoOpenGL();

            minu=f->data.uvpts[0]->data.Spot.u;
            minv=f->data.uvpts[0]->data.Spot.v;
            for(int i=1;i<f->data.sides;i++) {
                q=f->data.uvpts[i]->data.Spot.u;
                if(minu>q) minu=q;
                q=f->data.uvpts[i]->data.Spot.v;
                if(minv>q) minv=q;
            }

            minu=floorf(minu);
            minv=floorf(minv);

            glBegin(GL_POLYGON);
            for(int i=0;i<f->data.sides;i++) {
                glTexCoord2f(f->data.uvpts[i]->data.Spot.u-minu,
                    f->data.uvpts[i]->data.Spot.v-minv);
                f->data.verts[i]->data.Spot.glPlot();
            }
            glEnd();
        }
    }
    glDisable(GL_TEXTURE_2D);
}

void Model::drawHighlight() {
//        MYPTR   m;
        Vec3 vMin,vMax,veca,vecb;
        Bound   bound;
//        float mag;
    bool showsubd=modelFlags&MF_SHOW_SUBD;
#ifdef FACE_CURL
        float ctrlP[15];
        int order;
        float iU;
#endif

//        glMatrixMode(GL_MODELVIEW);
//        glLoadIdentity();
        glShadeModel(GL_FLAT);
        glDisable(GL_DEPTH_TEST);

        switch(MouseMode) {
            case MM_ADDBONE:
                if(drag.b!=0) {
                    cView.mapScr2XYZ(lX,lY,vMin);

//                    glEnable(GL_BLEND);
                    glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
                    SelectGLColorAlpha(AppData::PenColor[NEWPEN],.5);
                    glLineWidth(3.0);
                    glBegin(GL_LINES);
                    vMin.glPlot();
                    drag.b->Spot.glPlot();
                    glEnd();
//                    glDisable(GL_BLEND);

                    glLineWidth(1.0);
                    glEnable(GL_LINE_STIPPLE);

                    glLineStipple(1,0x6969);
                    glColor3f(1.0,1.0,1.0);
                    glBegin(GL_LINES);
                    vMin.glPlot();
                    drag.b->Spot.glPlot();
                    glEnd();

                    glLineStipple(1,0x9696);
                    glColor3f(0.0,0.0,0.0);
                    glBegin(GL_LINES);
                    vMin.glPlot();
                    drag.b->Spot.glPlot();
                    glEnd();

                    glDisable(GL_LINE_STIPPLE);                    
                }
            break;
        }
        
        glEnable(GL_COLOR_LOGIC_OP);
        glLogicOp(GL_XOR);
        SelectGLColor(AppData::PenColor[HIGHPEN]);

        switch(HighType) {
//              case HT_NOTHING:
//              break;
            case HT_UVPT:
                glPointSize(5.0f);
                glBegin(GL_POINTS);
                HighItem.u->data.Spot.glPlot();
                glEnd();
                break;
            case HT_VERT:
                glPointSize(5.0f);
                glBegin(GL_POINTS);
                HighItem.v->draw(showsubd);
                glEnd();
                break;
            case HT_EDGE:
                glLineWidth(3.0f);
                HighItem.e->trace(((modelFlags&MF_SHOW_SUBD)==0)?0:subPreviewLevel);
                glLineWidth(1.0f);
                break;
            case HT_FACE:
//                if(HighItem.f==NULL) {
//                    UserInt::updateStatusBar("Odd, Highitem.f==NULL",0,0);
//                }

//                if((HighItem.f->flags&TF_HIDDEN)==0) {
                    glLineWidth(3.0f);
                    for(int i=0;i<HighItem.f->data.sides;i++)
                    HighItem.f->edges[i]->trace(((modelFlags&MF_SHOW_SUBD)==0)?0:subPreviewLevel);
//                    glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
//                    glBegin(GL_POLYGON);
//                    glEnd();
//                    glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
                    glLineWidth(1.0f);
//                }
                break;
            case HT_BONE:
                        if(HighItem.b->data.Parent) {
                            glLineWidth(3.0);
                            glBegin(GL_LINES);
                            HighItem.b->data.Parent->Spot.glPlot();
                            HighItem.b->Spot.glPlot();
                            glEnd();
                        }

                        glDisable(GL_COLOR_LOGIC_OP);
                        if(!gadget) HighItem.b->drawTranslation(cView.Scale,0,vMin);
//                        glEnable(GL_BLEND);
                        glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);

                        getBoneBound(*HighItem.b,bound);

                        vMin.x=bound.get_minx()-.05f*cView.Scale;
                        vMin.y=bound.get_miny()-.05f*cView.Scale;
                        vMin.z=bound.get_minz()-.05f*cView.Scale;
                        vMax.x=bound.get_maxx()+.05f*cView.Scale;
                        vMax.y=bound.get_maxy()+.05f*cView.Scale;
                        vMax.z=bound.get_maxz()+.05f*cView.Scale;

                        HighItem.b->notScaled.loadIntoOpenGL();

                        glLineWidth(1.0);
                        glBegin(GL_LINE_STRIP);
// start the box
                        veca=vMin;
                        veca.glPlot();
//0,0,0 - 1,0,0
                        veca.x=vMax.x;
                        SelectGLColorAlpha(AppData::PenColor[XAXISPEN],.5);
                        veca.glPlot();
//1,0,0-1,1,0
                        veca.y=vMax.y;
                        SelectGLColorAlpha(AppData::PenColor[YAXISPEN],.5);
                        veca.glPlot();
//1,1,0-0,1,0
                        veca.x=vMin.x;
                        SelectGLColorAlpha(AppData::PenColor[XAXISPEN],.5);
                        veca.glPlot();

//0,1,0-0,0,0
                        veca.y=vMin.y;
                        SelectGLColorAlpha(AppData::PenColor[YAXISPEN],.5);
                        veca.glPlot();

//0,0,0-0,0,1
                        veca.z=vMax.z;
                        SelectGLColorAlpha(AppData::PenColor[ZAXISPEN],.5);
                        veca.glPlot();

//0,0,1 - 1,0,1
                        veca.x=vMax.x;
                        SelectGLColorAlpha(AppData::PenColor[XAXISPEN],.5);
                        veca.glPlot();
//1,0,1-1,1,1
                        veca.y=vMax.y;
                        SelectGLColorAlpha(AppData::PenColor[YAXISPEN],.5);
                        veca.glPlot();
//1,1,1-0,1,1
                        veca.x=vMin.x;
                        SelectGLColorAlpha(AppData::PenColor[XAXISPEN],.5);
                        veca.glPlot();

//0,1,1-0,0,1
                        veca.y=vMin.y;
                        SelectGLColorAlpha(AppData::PenColor[YAXISPEN],.5);
                        veca.glPlot();
                        glEnd();

                        glBegin(GL_LINES);
                        SelectGLColorAlpha(AppData::PenColor[ZAXISPEN],.5);
//0,1,1 - 0,1,0
                        veca.y=vMax.y;
                        veca.glPlot();
                        veca.z=vMin.z;
                        veca.glPlot();
//1,1,0 - 1,1,1
                        veca.x=vMax.x;
                        veca.glPlot();
                        veca.z=vMax.z;
                        veca.glPlot();
//1,0,1 - 1,0,0
                        veca.y=vMin.y;
                        veca.glPlot();
                        veca.z=vMin.z;
                        veca.glPlot();

                        glEnd();
                break;
                case HT_KNOB:
                    if(!gadget) {
                        glDisable(GL_COLOR_LOGIC_OP);
                        glEnable(GL_DEPTH_TEST);
                        HighItem.b->drawRotation2(cView.Scale,MM_GADNULL,cView.Orient.z);
                    }
                    break;
//                case HT_SCALE:
//                    glDisable(GL_COLOR_LOGIC_OP);
//                    HighItem.b->drawScaleGadget(cView.Scale*.5);
//                    break;
        }
        glDisable(GL_COLOR_LOGIC_OP);
        glEnable(GL_DEPTH_TEST);
}

void Model::drawMarking() {
    Face*   f;
    Edge*   e;
    Vert*   v;
    int subd=((modelFlags&MF_SHOW_SUBD)==0)?0:subPreviewLevel;

//        glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
    SelectGLColorAlpha(AppData::PenColor[MARKPEN],.5);

    bool showsubd=modelFlags&MF_SHOW_SUBD;
    if(modelFlags&MF_SHOW_VERTS) {
        glPointSize(5.0);
        glBegin(GL_POINTS);
        STEPVERTS(this,v)
        if(v->flags&TF_MARKED)
        if((v->flags&TF_HIDDEN)==0)
        v->draw(showsubd);
        glEnd();
    }

    if(modelFlags&MF_SHOW_EDGES) {
        glLineWidth(3.0);
        glDisable(GL_LINE_STIPPLE);
        STEPEDGES(this,e) {
            if((e->flags&TF_HIDDEN)==0)
            if(e->flags & TF_MARKED) {
                SelectGLColorAlpha(AppData::PenColor[MARKPEN],.5);
                e->trace(subd);
            }
            else if(e->data.VertS->Order>1
                && (e->data.VertE->Order==e->data.VertS->Order-1 || e->data.VertE->Order==e->data.VertS->Order+1)) {
                SelectGLColorAlpha(AppData::PenColor[NEWPEN],.5);
                e->trace(0);
            }
        }
        glLineWidth(1.0);
    }

    if(modelFlags&MF_SHOW_FACES) {
        SelectGLColorAlpha(AppData::PenColor[MARKPEN],.5);
        glDisable(GL_DEPTH_TEST);
        if(modelFlags&MF_SHOW_SUBD) {
            for(Patch* p=patches;p;p=p->next) {
                if(p->f->flags&TF_MARKED) {
                    p->draw(subd,false);
                }
            }
        } else {
            STEPFACES(this,f)
            if((f->flags&TF_HIDDEN)==0)
            if(f->flags&TF_MARKED) {
                glBegin(GL_POLYGON);
                for(int i=0;i<=f->data.sides;i++)
                f->data.verts[i]->data.Spot.glPlot();
                glEnd();
            }
        }
        glEnable(GL_DEPTH_TEST);
    }

//        glDisable(GL_BLEND);
}

void Model::drawSubdivision() {
    Patch*p;

    for(p=patches;p;p=p->next)
    p->draw(subPreviewLevel,true);
    glDisable(GL_TEXTURE_2D);
}

void Model::drawUVEdges() {
        Face* f;

        glBegin(GL_LINES);

        STEPFACES(this,f)
        if(f->data.uvpts)
        if(f->data.Texture==CurTint) {
            for(int i=0;i<f->data.sides-1;i++)
            f->data.uvpts[i]->drawEdge(f->data.uvpts[i+1]);
            f->data.uvpts[0]->drawEdge(f->data.uvpts[f->data.sides-1]);
        }
        glEnd();
}

void Model::drawUVPts() {
        Face* f;

        glBegin(GL_POINTS);
        STEPFACES(this,f)
        if(f->data.uvpts)
        if(f->data.Texture==CurTint)
        for(int i=0;i<f->data.sides;i++)
        f->data.uvpts[i]->data.Spot.glPlot();

        glEnd();
}

void Model::drawVertices(void) {
        Vert* v;
//      bool flag,blank;

//      UpdateStatusBar("Drawing vertices.",0,0);

        bool showsubd=modelFlags&MF_SHOW_SUBD;
        SelectGLColor((COLORREF)0xffffff);
        glPointSize(3.0);
        glBegin(GL_POINTS);
//      flag=blank=false;
        STEPVERTS(this,v)
            if((v->flags&TF_HIDDEN)==0)
                v->draw(showsubd);
        glEnd();

//      if(flag==false) UpdateStatusBar("No verts drawn",0,0);
}
