// model_g.cpp -- implements member functions of class Model that start with 'g'

#include "userint.h"
#include "appdata.h"
#include "bound.h"
#include "model.h"
#include "edge.h"
#include "join.h"
#include "vert.h"
#include "funcs.h"
#include "mystr.h"
#include "lsmres.h"
#include "calc.h"

// function to find the dimensions of a Bone's immediate dependents

void Model::getBoneBound(Bone& b,Bound& bound) {
    Vec3    temp;

    bound.empty();

    temp.set(0.0f);
    bound.include(temp);

    Bone*   ob;
    STEPBONES(this,ob)
    if(ob->data.Parent==&b) {
        (temp=ob->Spot)/=b.notScaled.fore;
        bound.include(temp);
    }

    Vert*   v;
    STEPVERTS(this,v)
    if(v->data.Owner==&b) {
        (temp=v->data.Spot)/=b.notScaled.fore;
        bound.include(temp);
    }
}

// function to find the length that a Bone should be
float Model::getBoneLength(Bone& b) {
    Bound bound;
    float temp;
    float result=cView.Scale*.75f;

    getBoneBound(b,bound);
    temp=bound.size();
    if(temp>result) result=temp;

    if(result>cView.Scale*1.25f) result=cView.Scale*1.25f;

    return result;
}
   

// this function finds the edge with the given start and end points, and creates
// one if it doesn't already exist
// This needs to be rewritten, because it leads to an O(n^2) performance issue
// when it is executed to find all of the edges for every face in a model
Edge* Model::getEdge(Vert* a,Vert* b) {
        Edge* e;

        if (a==b) return (Edge*)0;

        e=a->findEdge(b);
        if(e==(Edge*)0) e=newEdge(a,b);
        return e;
}

Join* Model::getJoin(UVPt* a,UVPt* b) {
    Join*   j=0;
    if(a==b) return j;
    if(a==0 || b==0) return j;

    j=a->findJoin(b);
    if(j==0) j=newJoin(a,b);
    return j;
}

HMENU Model::getPopupMenu() {
    MENUITEMINFO qmii;

    memset(&qmii,0,sizeof(MENUITEMINFO));
    qmii.cbSize=sizeof(MENUITEMINFO);
    qmii.fMask=MIIM_STATE;

    HMENU menu=LoadMenu(UserInt::hInst,MAKEINTRESOURCE(IDPOPUPMENU));

    qmii.fState=(modelFlags&MF_LOCK_X)?MFS_CHECKED:MFS_UNCHECKED;
    SetMenuItemInfo(menu,IDM_LOCK_X,FALSE,&qmii);

    qmii.fState=(modelFlags&MF_LOCK_Y)?MFS_CHECKED:MFS_UNCHECKED;
    SetMenuItemInfo(menu,IDM_LOCK_Y,FALSE,&qmii);

    qmii.fState=(modelFlags&MF_LOCK_Z)?MFS_CHECKED:MFS_UNCHECKED;
    SetMenuItemInfo(menu,IDM_LOCK_Z,FALSE,&qmii);

    return menu;
}

Tint* Model::getTintByName(Mystr& tn) {
    Tint*   t;

    if(tn.len()==0) return &AppData::defTint;

    t=findTintByName(tn);
    if(t) return t;

    t=newTint(0xa0a0a0);
    if(t) {
        t->data.Name=tn;
    } else {
        t=&AppData::defTint;
    }
    return t;
}

