// model_i.cpp -- implements member functions of class Model that start with 'i'

#include <windows.h>
#include "model.h"
#include "bone.h"
#include "norm.h"
#include "part.h"
#include "tint.h"
#include "uvpt.h"
#include "vert.h"
#include "funcs.h"
#include "appdata.h"
#include "mouse.h"
#include "obvec.h"


void Model::incSelection() {
        Vert* v;

        STEPVERTS(this,v)
        if(v->Order)
        v->Order++;
}

Bone *Model::indexBone(INDEX i) {
        return (Bone*)Bones.index(i);
}

Knob *Model::indexKnob(INDEX i) {
        return (Knob*)Knobs.index(i);
}

Norm *Model::indexNorm(INDEX i) {
        Norm*   n;

        STEPNORMS(this,n)
        if(n->index==i)
        return n;

        return (Norm*)0;
}

Part *Model::indexPart(INDEX i) {
        return (Part*)Parts.index(i);
}

Tint *Model::indexTint(INDEX i) {
    Tint *t;
    t=Tints.index(i);
    if(t==NULL) return &AppData::defTint;
    return t;
}

UVPt* Model::indexUVPt(INDEX i) {
        UVPt*   u;
        if(!uvpttable) {
            uvpttable=new UVPt*[NextUVPtIndex];
            if(uvpttable) {
                for (int j=0;j<NextUVPtIndex;j++)
                uvpttable[j]=(UVPt*)0;
                STEPUVPTS(this,u)
                uvpttable[u->index]=u;
            } else {
                return (UVPt*)UVPts.index(i);
            }
        }
        if(i>=0)
        return uvpttable[i];
        return (UVPt*)0;
}

Vert* Model::indexVert(INDEX i) {
        Vert*   v;

        if(!verttable) {
            verttable=new Vert*[NextVertIndex];
            if(verttable) {
                for (int j=0;j<NextVertIndex;j++)
                verttable[j]=(Vert*)0;
                STEPVERTS(this,v)
                verttable[v->index]=v;
            } else {
                return (Vert*)Vertices.index(i);
            }
        }
        if(i>=0) return verttable[i];
        return (Vert*)0;
}

void Model::insertCube() {
    clearAll();

    Array<Vert*> newverts;
    newverts=new Vert*[8];

    for(int i=0;i<8;i++) {
        newverts[i]=newVert(0,0);
        newverts[i]->data.Spot.set(1.0f);
        newverts[i]->flags|=TF_MARKED;
    }

    for(int i=0;i<4;i++) newverts[i]->data.Spot.x*=-1;
    for(int i=0;i<8;) {
        newverts[i]->data.Spot.y*=-1;
        i++;
        newverts[i]->data.Spot.y*=-1;
        i+=3;
    }
    for(int i=0;i<8;i+=2) newverts[i]->data.Spot.z*=-1;

    newFace(newverts[0],newverts[1],newverts[3],newverts[2]);
    newFace(newverts[4],newverts[5],newverts[7],newverts[6]);
    newFace(newverts[0],newverts[2],newverts[6],newverts[4]);
    newFace(newverts[1],newverts[3],newverts[7],newverts[5]);
    newFace(newverts[0],newverts[4],newverts[5],newverts[1]);
    newFace(newverts[2],newverts[6],newverts[7],newverts[3]);

    modelFlags|=MF_BUILD_SUBD;
    validate();
    takeSnap();
    MarkType=HT_VERT;
    applyMarkLogic();
}

void Model::insetHigh() {
    Face* f=(Face*)0;
    if(HighType!=HT_FACE) return;

    f=HighItem.f;
    clearAll();

    f->flags|=TF_MARKED;

    MarkType=HT_FACE;
    applyMarkLogic();
    extrudeMarked();

    Vec3 c;
    f->calcCenter(c);

    for(int i=0;i<f->data.sides;i++) ((f->data.verts[i]->data.Spot-=c)*=.5)+=c;

    gadget=new UniScaleMeshGadget(this,f);
}

bool Model::isUniqueBoneName(char *n,Bone* ex) {
        Bone* b;

        STEPBONES(this,b)
        if(b!=ex)
        if(b->data.Name==n) return false;

    return true;
}

bool Model::isUniquePartName(char *n,Part* ex) {
        Part* p;

        STEPPARTS(this,p)
        if(p!=ex)
        if(p->Name==n) return false;

    return true;
}

bool Model::isUniqueTintName(char *n,Tint *ex) {
        Tint *t;

        STEPTINTS(this,t)
        if(t!=ex)
        if(t->data.Name==n) return false;

    return true;
}
