// model_m.cpp -- implements member functions of class Model that start with 'm'

#include "userint.h"
#include <math.h>
#include "appdata.h"
#include "model.h"
#include "dialex.h"
#include "vert.h"
#include "face.h"
#include "edge.h"
#include "lsmtypes.h"
#include "lsmres.h"
#include "funcs.h"
#include "mouse.h"
#include "calc.h"

void Model::makeLathe(HWND hwnd) {
        MYPTR mp;
        int i,j,k,count;
        Vert **varray;
        float c,s;
        Matrix mat;
        Vec3 axis(0.0f);
        int div;
        Dialex *dx;

        dx=new Dialex;
        dx->setDialogFlag(CB_LOCKX,TRUE);
        dx->setDialogFlag(CB_LOCKY,FALSE);
        dx->setDialogFlag(CB_LOCKZ,FALSE);
        dx->setDialogInteger(EB_VECX,LatheSteps);

        if(DialogBoxParam(UserInt::hInst,MAKEINTRESOURCE(DLG_LATHE),hwnd,
                LatheDlg,(int)dx)!=IDOK) {
                delete dx;
                return;
        }

        LatheSteps=div=dx->getDialogInteger(EB_VECX);
//        axis.set(0.0f);

        if(dx->getDialogFlag(CB_LOCKX)) axis.x=1.0;
        if(dx->getDialogFlag(CB_LOCKY)) axis.y=1.0;
        if(dx->getDialogFlag(CB_LOCKZ)) axis.z=1.0;

        delete dx;

        STEPEDGES(this,mp.e)
        if(mp.e->flags & TF_MARKED)
        if(mp.e->FaceL!=NULL || mp.e->FaceR!=NULL) {
                UserInt::informUser("One or more of the marked edges has a face bordering on it.");
                return;
        }

        count=0;
        STEPVERTS(this,mp.v)
        if(mp.v->flags&TF_MARKED)
        count++;

        if(count==0) {
                UserInt::informUser("No vertices marked!");
                return;
        }

        varray=new Vert*[div*count];

        if(varray==NULL) {
                UserInt::informUser("Memory allocation for vertex table failed.");
                return;
        }

        for(i=0;i<div*count;i++) varray[i]=(Vert*)0;

        i=0;
        STEPVERTS(this,mp.v)
        if(mp.v->flags&TF_MARKED) {
                varray[i]=mp.v;
                i++;
        }

        for(i=count;i<count*div;i++) {
                varray[i]=newVert(0,0);
                if(varray[i]==NULL) {
                        for(j=count;j<count*div;j++)
                        if(varray[j])
                        deleteVert(varray[j]);
                        UserInt::informUser("Could not allocate memory to create needed vertices.");
                        return;
                }
        }

    mat.l.set(0.0f);

        for(i=1;i<div;i++) {
                c=2.0*PI*(float)i/(float)div;
                s=sinf(c);
                c=cosf(c);
                mat.setXVec(axis.x*axis.x*(1.0-c)+c,
                        axis.x*axis.y*(1.0-c)+s*axis.z,
                        axis.x*axis.z*(1.0-c)-s*axis.y);

                mat.setYVec(axis.y*axis.x*(1.0-c)-s*axis.z,
                        axis.y*axis.y*(1.0-c)+c,
                        axis.y*axis.z*(1.0-c)+s*axis.x);

                mat.setZVec(axis.z*axis.x*(1.0-c)+s*axis.y,
                        axis.z*axis.y*(1.0-c)-s*axis.x,
                        axis.z*axis.z*(1.0-c)+c);
                for(j=0;j<count;j++)
                (varray[i*count+j]->data.Spot = varray[j]->data.Spot)*=mat;
        }

// make circuits
        for(i=0;i<count;i++) {
                for(j=0;j<div-1;j++)
                newEdge(varray[i+j*count],varray[i+count+j*count]);
                newEdge(varray[i+(div-1)*count],varray[i]);
        }

// make faces
        STEPEDGES(this,mp.e)
        if(mp.e->flags & TF_MARKED) {
                for(k=0;k<count;k++) {
                        if(varray[k]==mp.e->data.VertS) i=k;
                        if(varray[k]==mp.e->data.VertE) j=k;
                }
                for(k=0;k<div-1;k++)
                newFace(varray[i+k*count],varray[j+k*count],varray[j+k*count+count],varray[i+k*count+count]);
                newFace(varray[i+k*count],varray[j+k*count],varray[j],varray[i]);
        }

        for(i=0;i<count*div;i++)
        varray[i]->flags|=TF_MARKED;

        delete []varray;
        validate();
        modelFlags|=MF_BUILD_SUBD;
        applyMarkLogic();
        takeSnap();
}

void Model::markAll() {
        Vert* v;

        STEPVERTS(this,v)
        v->flags|=TF_MARKED;

        MarkType=HT_VERT;

        applyMarkLogic();

        InvalidateView();
}

void Model::markContiguous() {
        Vert*   v;
        Edge*   e;
        Face*   f;
        FLAGS   result;
        bool    repeat;

        if(HighItem.e==NULL) return;

        result=(~HighItem.e->flags)&TF_MARKED;

        switch(HighType) {
                case HT_VERT:
                        STEPVERTS(this,v) v->flags&=~TF_NEW;
                        HighItem.v->flags|=TF_NEW;
                        do {
                                repeat=false;
                                STEPEDGES(this,e)
                                if(e->data.VertS->flags&TF_NEW) {
                                        if((e->data.VertE->flags&TF_NEW)==0) {
                                                e->data.VertE->flags|=TF_NEW;
                                                repeat=true;
                                        }
                                } else {
                                        if(e->data.VertE->flags&TF_NEW) {
                                                e->data.VertS->flags|=TF_NEW;
                                                repeat=true;
                                        }
                                }
                        } while (repeat);
                        STEPVERTS(this,v)
                        if(v->flags&TF_NEW)
                        v->flags=(v->flags & ~TF_MARKED) | result;
                break;
                case HT_FACE:
                        STEPFACES(this,f) f->flags&=~TF_NEW;
                        HighItem.f->flags|=TF_NEW;
                        do {
                                repeat=false;
                                STEPEDGES(this,e)
                                if(e->FaceL)
                                if(e->FaceR)
                                if(e->FaceL->flags&TF_NEW) {
                                        if((e->FaceR->flags&TF_NEW)==0) {
                                                e->FaceR->flags|=TF_NEW;
                                                repeat=true;
                                        }
                                } else {
                                        if(e->FaceR->flags&TF_NEW) {
                                                e->FaceL->flags|=TF_NEW;
                                                repeat=true;
                                        }
                                }
                        } while (repeat);

                        STEPFACES(this,f)
                        if(f->flags&TF_NEW)
                        f->flags=(f->flags & ~TF_MARKED) | result;
                break;
                case HT_EDGE:
                        STEPEDGES(this,e) e->flags &= ~TF_NEW;
                        HighItem.e->flags|=TF_NEW;
                        do {
                                repeat=false;
                                STEPEDGES(this,e)
                                if(e->flags&TF_NEW) {
                                        e->data.VertS->flags|=TF_NEW;
                                        e->data.VertE->flags|=TF_NEW;
                                }
                                STEPEDGES(this,e)
                                if((e->flags&TF_NEW)==0)
                                if((e->data.VertS->flags|e->data.VertE->flags)&TF_NEW) {
                                        e->flags|=TF_NEW;
                                        repeat=true;
                                }
                        } while (repeat);
                        STEPEDGES(this,e)
                        if(e->flags&TF_NEW)
                        e->flags=(e->flags & ~TF_MARKED) | result;
                break;
        }
        MarkType=HighType;
        applyMarkLogic();
}

void Model::matrixMarked(HWND hwnd,short action) {
        Vert* v;
        Dialex *dx;

        if(this==NULL) return;
        dx=new Dialex;

        dx->setDialogLabel(0,"Matrix transform");
        dx->setDialogMatrix(EB_VECXX,&UserMatrix);

        if(DialogBoxParam(UserInt::hInst,MAKEINTRESOURCE(DLG_MATRIX),hwnd,
                InputMatrixDlg,(int)dx)==IDOK) {
                switch(action) {
                        case EA_EXTRUDE: extrudeMarked(); break;
                        case EA_COPY: copyMarked(); break;
                }

                STEPVERTS(this,v)
                if(v->flags&TF_MARKED) {
                    v->data.Spot*=UserMatrix;
                }
                InvalidateView();
        }

        modelFlags|=MF_CALC_SUBD;
        delete dx;
        takeSnap();
}

void Model::mergeVerts(Vert*va,Vert*vb) {
    Edge*   e;
    Edge*   ne;
    Face*   f;
    Face*   nf;

    STEPEDGES(this,e) {
// mark all orphan edges as already being such
        e->flags&=~TF_NEW;
        if(e->FaceL==NULL && e->FaceR==NULL) e->flags|=TF_NEW;
        if(e->data.VertS==va) e->data.VertS=vb;
        if(e->data.VertE==va) e->data.VertE=vb;
    }

    STEPFACES(this,f) {
        for(int i=0;i<=f->data.sides;i++) {
            if(f->data.verts[i]==va) f->data.verts[i]=vb;
        }
        f->removeEdgeXRef();
    }

    STEPFACES(this,f) {
        int i1,i2;
        do {
            i1=i2=-1;
            for(int i=0;i<f->data.sides;i++)
            if(f->data.verts[i]==vb) {
                if(i1==-1) {
                    i1=i;
                } else {
                    if(i2==-1) i2=i;
                }
            }
            if(i2!=-1) {
// process face
// UV map data is always invalid when vertices are modified
    f->removeMapping();

// and now we process different situations of face modifications
// case A: joined vertices are adjacent in the vertex array
// solution: shift all indices, from i2 on up, down one position
    if(i2-i1==1) {
        for(int i=i1;i<f->data.sides;i++)
        f->data.verts[i]=f->data.verts[i+1];
        f->data.sides--;
    }
// case B: joined vertices are separated by one other vertex in the array
// solution: shift all indices, from i1+2 on up, down two positions
    else if(i2-i1==2) {
        for(int i=i1;i<f->data.sides-1;i++)
        f->data.verts[i]=f->data.verts[i+2];
        f->data.sides-=2;
    }
// case C: joined vertices are at position 0 and n-1 or n-2 of array
// solution: no moving, but sides count is adjusted
    else if(i1==0 &&(i2==f->data.sides-1 || i2==f->data.sides-2) ) { f->data.sides=i2; }
// case D: joined vertices are at position 1 and n-1 of array
// solution:  move all indices, from i1 to i2, down one position
    else if(i1==1 and i2==f->data.sides-1) {
        for(int i=0;i<f->data.sides-1;i++)
        f->data.verts[i]=f->data.verts[i+1];
        f->data.sides-=2;
    }
// case E:  joined vertices are separated by more than three positions
// solution:  make a new face
    else {
        nf= new Face(i2-i1);
        nf->data.Texture=f->data.Texture;

        for(int i=i1;i<=i2;i++)
        nf->data.verts[i-i1]=f->data.verts[i];

        Faces.addTail(nf);

        for(int i=i2;i<=f->data.sides;i++)
        f->data.verts[i-i2+i1]=f->data.verts[i];

        f->data.sides-=i2-i1;
    }
// end of face processing
            }
        } while (i2!=-1);
    }
    
// remove all degenerate edges
    e=FIRSTEDGE(this);
    while(e) {
        ne=(Edge*)e->next();
        if(e->data.VertS==e->data.VertE) deleteEdge(e);
        e=ne;
    }

// remove all degenerate faces
    f=FIRSTFACE(this);
    while(f) {
        nf=(Face*)f->next();
        if(f->data.sides < 3) deleteFace(f);
        f=nf;
    }

    deleteVert(va);

    validate();
    validateMapping();

    e=FIRSTEDGE(this);
    while(e) {
        ne=(Edge*)e->next();
        if((e->flags&TF_NEW)==0) // was not marked as orphan before
        if(e->FaceL==NULL && e->FaceR==NULL) // but is now an orphan
        deleteEdge(e);  // remove the edge

        e=ne;
    }

    modelFlags|=MF_BUILD_SUBD;
}

void Model::mergeAtEdge(Edge* e) {
    int i;
    if(e==(Edge*)0) return;

    Face*   lf=e->FaceL;
    if(lf==(Face*)0) return;

    Face*   rf=e->FaceR;
    if(rf==(Face*)0) return;

    Vert*   vs=e->data.VertS;
    Vert*   ve=e->data.VertE;

    int il=-1;
    int sl=lf->data.sides;
    for(i=0;i<sl;i++) {
        if(lf->data.verts[i]==vs && lf->data.verts[i+1]==ve) il=i;
        if(lf->data.verts[i]==ve && lf->data.verts[i+1]==vs) {
            il=i;
            Vert* qv=vs;vs=ve;ve=qv;
        }
    }

    int ir=-1;
//    for(i=0;i<sl;i++)
//    if(lf->edges[i]==e) ir=i;

//    if(ir!=il) UpdateStatusBar("Differing results for left side",0,0);
        
    if(il<0) return;    // shouldn't happen, but you never know

    bool flip=false;
    int sr=rf->data.sides;

    for(i=0 /*,ir=-1 */;i<sr;i++) {
        if(rf->data.verts[i]==ve && rf->data.verts[i+1]==vs) ir=i;
        if(rf->data.verts[i]==vs && rf->data.verts[i+1]==ve) {
            ir=i;
            flip=true;
        }
    }

    if(ir<0) return;    // shouldn't happen, but you never know

//    int it=-1;
//    for(i=0;i<sr;i++)
//    if(rf->edges[i]==e) it=i;

//    if(ir!=it) UpdateStatusBar("Differing results for right side",0,0);

    if(flip) {
        rf->reverseWinding();
        ir=sr-ir-1;
    }

    Face*   nf=new Face(sl+sr-2);

    if(nf==(Face*)0) return;

    Faces.addTail(nf);

    nf->data.Texture=lf->data.Texture;

    int j=0;
   
// 0 to il gets copied directly into new Face
    for(i=0;i<il;i++)
    nf->data.verts[j++]=lf->data.verts[i];

    for(i=ir+1;i<sr;i++)
    nf->data.verts[j++]=rf->data.verts[i];

    for(i=0;i<ir;i++)
    nf->data.verts[j++]=rf->data.verts[i];

    for(i=il+1;i<=sl;i++)
    nf->data.verts[j++]=lf->data.verts[i];

    deleteFace(lf);
    deleteFace(rf);
    deleteEdge(e);
//    modelFlags|=MF_BUILD_SUBD;
    validateMapping();
    validate();
    takeSnap();
}

