// model_n.cpp -- implementes member functions of class Model that start with 'n'

#include "userint.h"
#include <math.h>
#include "model.h"
#include "vert.h"
#include "face.h"
#include "bone.h"
#include "edge.h"
#include "part.h"
#include "norm.h"
#include "funcs.h"
#include "uvpt.h"
#include "appdata.h"

Bone* Model::newBone(short pX,short pY) {
        Bone* b;

        b=new Bone;
        Bones.addTail(b);
        cView.mapScr2XYZ(pX,pY,b->Spot);
        b->data.translate=b->Spot;
        b->data.Name.own(createBoneName());
        b->index=NextBoneIndex++;
        b->buildTransforms();

        return b;
}

Edge* Model::newEdge(Vert* a,Vert* b) {
        Edge* e;

        e=new Edge;
        if(e!=NULL) {
// inherited stuff
            Edges.addTail(e);
            modelFlags|=MF_BUILD_SUBD;
            e->index=NextEdgeIndex++;
            e->flags=EF_NO_LEFT|EF_NO_RIGHT;
// primary data
            e->data.VertS=a;
            e->data.VertE=b;
//              e->data.sharpness=0;
// derivative stuff
            a->addEdge(e);
            b->addEdge(e);
        }
        return e;
}

Face* Model::newFace(Vert* a,Vert* b,Vert* c,Vert* d) {
        Face* f;
        Edge *ee /*,*ef,*eg,*eh */ ;

        if(d==NULL) {
                if(a==b || a==c || b==c) return NULL;
        } else {
                if(b==d || a==c) return NULL;
                if(a==b) { b=c; c=d; d=NULL; }
                if(b==c) { c=d; d=NULL; }
                if(c==d) d=NULL;
        }

        ee=a->findEdge(b);
        if(ee!=NULL)
        if(ee->FaceL!=NULL && ee->FaceR!=NULL) {
//            UserInt::informUser("Face creation failed:  First Edge already has two bordering Faces");
            return NULL;
        }

        ee=b->findEdge(c);
        if(ee!=NULL)
        if(ee->FaceL!=NULL && ee->FaceR!=NULL) {
//            UserInt::informUser("Face creation failed:  Second Edge already has two bordering Faces");
            return NULL;
        }
        if(d==NULL) {
                ee=c->findEdge(a);
                if(ee!=NULL)
                if(ee->FaceL!=NULL && ee->FaceR!=NULL) return NULL;
        } else {
                ee=c->findEdge(d);
                if(ee!=NULL)
                if(ee->FaceL!=NULL && ee->FaceR!=NULL) return NULL;

                ee=d->findEdge(a);
                if(ee!=NULL)
                if(ee->FaceL!=NULL && ee->FaceR!=NULL) return NULL;
        }

        f=new Face;
        if(f!=NULL) {
                Faces.addTail(f);
                modelFlags|=MF_BUILD_SUBD;
                // establish basic geometry
                f->data.sides=((d==NULL)?3:4);
                f->data.verts=new Vert*[f->data.sides+1];
                f->data.verts[0]=a;
                f->data.verts[1]=b;
                f->data.verts[2]=c;
                if(d) {
                        f->data.verts[3]=d;
                        f->data.verts[4]=a;
                } else {
                        f->data.verts[3]=a;
                }

                // set other data to defaults
//              f->data.uvpts=NULL;
                f->edges=new Edge*[f->data.sides];
                f->data.Texture=CurTint;
                f->flags=0;
                f->index=NextFaceIndex++;

                for(int i=0;i<f->data.sides;i++)
                f->edges[i]=getEdge(f->data.verts[i],f->data.verts[i+1]);
        }
        return f;
}

Join*   Model::newJoin(UVPt* ua,UVPt*ub) {
        Join* j;

        j=new Join(ua,ub);
        if(j!=NULL) {
// inherited stuff
            Joins.addTail(j);
            modelFlags|=MF_BUILD_SUBD;
        }
        return j;
}

Knob* Model::newKnob() {
    Knob *k;

    k=new Knob;

    if(k!=NULL) {
        Knobs.addTail(k);
        k->flags=0;
        k->index=NextKnobIndex++;
        k->Owner=(Part*)0;
    }
    return k;
}

Norm* Model::newNorm() {
    Norm*   n;

    n=new Norm;
    if(n) {
        Norms.addTail(n);
    }
    return n;
}

Part* Model::newPart(Knob *b,Knob *e,Knob *r) {
    Part* p;
//        bool    newRiser;

    p= new Part;
    if(p!=NULL) {
        Parts.addTail(p);
        p->Base=b;
        p->End=e;
        p->Rise=r;
        p->flags=0;
        p->index=NextPartIndex++;
        p->Name.own(createPartName());
        p->BaseBone=newBone(0,0);
        p->EndBone=newBone(0,0);
    }
    return p;
}

Tint* Model::newTint(COLORREF c) {
    Tint* t;

    t=new Tint;
    if(t) {
        Tints.addTail(t);
        t->data.color=c;
        t->data.Name.own(createTintName());
        t->index=NextTintIndex++;
    }
    return t;
}

UVPt* Model::newUVPt() {
        UVPt* u;

        u=new UVPt;
        if(u) {
                UVPts.addTail(u);
                // uvpttable is no longer valid, nuke it
                uvpttable=0;
                u->index=NextUVPtIndex++;
        } else {
                UserInt::informUser("UVPt creation failed.");
        }
        return u;
}

// used when creating a new vertex in edit mode
Vert* Model::newVert(short pX,short pY) {
        Vert* v;

        v=new Vert;
        if(v) {
            modelFlags|=MF_BUILD_SUBD;
            Vertices.addTail(v);
// verttable is no longer valid, nuke it
            verttable=(Vert**)0;
// inherited stuff
            v->index=NextVertIndex++;
// the primary data
            cView.mapScr2XYZ(pX,pY,v->data.Spot);
//              v->data.Owner=NULL; // set in constructor
//              v->data.flags=0; // set in constructor
//                v->Order=0; // moved to constructor
//                v->Sharp=0; // moved to constructor
//              v->Copy=NULL; // moved to constructor
//              v->Extra=NULL;  // moved to constructor
//              v->flags=0; // moved to constructor
        } else {
//              UserInt::informUser("Vertex creation failed.");
        }
        return v;
}

// this is for the final destruction of the bone at project end
// or when it becomes part of an orphan change
void Model::nukeBone(Bone* b) {
        delete b;
}

void Model::nukeEdge(Edge *e) {
        delete e;
}

void Model::nukeFace(Face* f) {
//        Edge* e;

        if(f==NULL) return;

        if(HighItem.f==f) clearHighlight();

        delete f;
}

void Model::nukeJoin(Join* j) {
    delete j;
}

void Model::nukeKnob(Knob *k) {
//        Part *p,*np;

        if(k==NULL) return;

        delete k;
}

void Model::nukeNorm(Norm* n) {
        delete n;
}

void Model::nukePart(Part* p) {
        Part* p1;
        Knob *k;

        if(p==NULL) return;

        STEPPARTS(this,p1)
        if(p1->Parent==p) p1->Parent=NULL;

        STEPKNOBS(this,k)
        if(k->Owner==p) {
                k->Owner=NULL;
        }

        k=p->Rise;

        delete p;

        nukeKnob(k);
}

void Model::nukeTint(Tint* t) {
    Face* f;

    if(t==NULL) return;

    STEPFACES(this,f)
    if(f->data.Texture==t) {
        f->data.Texture=&AppData::defTint;
    }

    if(t==CurTint) CurTint=&AppData::defTint;

    t->deleteInfo();

    delete t;
}

void Model::nukeUVPt(UVPt* u) {
    u->destroyHistory();
    delete u;
    // uvpttable is no longer valid, nuke it
    uvpttable=(UVPt**)0;
}

void Model::nukeVertex(Vert* v) {
//        Edge *e,*ne;
//        Face *f,*nf;

        if(v==NULL) {
                UserInt::informUser("NULL pointer passed to nukeVertex()");
                return;
        }
#if 0
  else {
                UserInt::informUser("Vertex destroyed");
        }
#endif

        if(HighItem.v==v) clearHighlight();

        delete v;
//        modelFlags|=MF_RECALC_SUBD;

        // verttable is no longer valid, nuke it
        verttable=(Vert**)0;
}
