// model_o.cpp -- implements member functions of class Model that start with 'o'

#include <windows.h>
#include <math.h>
#include "model.h"
#include "edge.h"
#include "face.h"
#include "funcs.h"

// this function is to get the vertices of all faces going in a counter-
// clockwise direction.  This is needful for making smoothing normals, and
// export to OBJ files, and probably for finding the final normals for
// surface subdivision.

// this should work for any mesh that has two sides (ie, is not a moebius strip
// or some similar type surface).

void Model::orientWindings() {
    Face*   f;
    Face*   tf;
    bool drop;
    Edge*   e;
    float dist;
    float td;
    Vec3 norm,cent;
    bool out;
//        bool gooddrop;

// use the TF_NEW flag to indicate processed faces and edges
    STEPFACES(this,f) f->flags&=(~TF_NEW);

    STEPEDGES(this,e)
    if(e->FaceL==NULL || e->FaceR==NULL) {
        e->flags|=TF_NEW;
    } else {
        e->flags&=(~TF_NEW);
    }

//      gooddrop=false;

    modelFlags|=MF_BUILD_SUBD|MF_RECALC_NORMS;
    modelFlags&=~MF_REORIENT;

    do {    // process sets of faces
// get the unprocessed face lying in the plane farthest from the origin
                f=NULL;
                dist=-1.0;
                out=true;
                STEPFACES(this,tf)
                if((tf->flags&TF_NEW)==0) {
                        tf->calcNormal(norm);
                        tf->calcCenter(cent);
                        norm.normalize();
                        td=norm.dot(cent);
                        if(fabs(td)>dist) {
                                dist=fabs(td);
                                f=tf;
                                out=(td>=0.0);
                        }
                }
// if there is no farthest face, done; happens when all faces are processed
                if(f==NULL) {
//                      if(gooddrop==true) ReportSystemError("At least one face processed!");
                        return;
                }

//              gooddrop=true;

// make sure that it's set to go counter-clockwise
                if(out==false) f->reverseWinding();

// mark face as processed
                f->flags|=TF_NEW;

                do { // process this set of faces
                        drop=true;
                        STEPEDGES(this,e)
                        if((e->flags&TF_NEW)==0) {
                                if(e->checkWindings()) drop=false;
            }
        } while(drop==false);
    } while (true);
}
