// model_r.cpp -- implements member functions of class Model that start with 'r'

#include "userint"
#include <math.h>
#include "appdata.h"
#include "fhand.h"
#include "model.h"
#include "bone.h"
#include "edge.h"
#include "face.h"
#include "norm.h"
#include "tint.h"
#include "uvpt.h"
#include "vert.h"
#include "funcs.h"
#include "dialex.h"
#include "parser.h"
#include "lsmres.h"
#include "lsmtypes.h"
#include "calc.h"

void Model::readMTLFile(fReader& file) {
    Mystr   buffer;
//    MYPTR   m;
//    Bone*   usebone;
    Tint*   usetint;
    float r,g,b;
//    Face*   f;

    usetint=(Tint*)0;

    while(file) {
        file.getOBJLine(buffer);
        Parser p(buffer);
        if(p.isValid()) {
            if(p.compareToken("newmtl ")) { // start of new material
// skip the token
                p.skipToken();
// and make the new item
                Mystr newName=p.getString();
// the OBJ spec for MTL files states that earlier statements override later
// ones.  Therefore, if the material already exists, then this statement is
// ignored.
                usetint=findTintByName(newName);
                if(usetint==(Tint*)0) {
// doesn't exist, create it
                    usetint=newTint(0xffffff);
                    usetint->data.Name=newName;
                } else {
// already exists, ignore all specs for this statement
                    usetint=(Tint*)0;
                }
//                if(newName) delete[]newName;
            }
            if(p.compareToken("Kd ")) { // diffuse values
//                COLORREF color;
                p.skipToken();
                if(usetint) {
                    r=p.getFloat();
                    g=p.getFloat();
                    b=p.getFloat();
                    usetint->data.color=RGB((unsigned char)(r*255.0),
                        (unsigned char)(g*255.0),
                        (unsigned char)(b*255.0));
                }
            }
        // other code for MTL specs will go here
        }
    }
}

void Model::readOBJFile(fReader& file) {
    Mystr   buffer;
    MYPTR   m;
    Bone*   usebone;
    Tint*   usetint;
    Face*   f;

    usebone=(Bone*)0;
    usetint=&AppData::defTint;

    while(file) {
        file.getOBJLine(buffer);
        Parser p(buffer);

        if(p.isValid()) {
            if(p.compareToken("vt ")) { // uv-mapping point
// skip the token
                p.skipToken();
// and make the new item
                m.u=newUVPt();
                m.u->data.Spot.u=p.getFloat();
                m.u->data.Spot.v=p.getFloat();
            }
            if(p.compareToken("v ")) { // a vertex
                p.skipToken();
                m.v=newVert(0,0);
                m.v->data.Spot.x=p.getFloat();
                m.v->data.Spot.y=p.getFloat();
                m.v->data.Spot.z=p.getFloat();
            }
            if(p.compareToken("vn ")) { // a normal
//                 p.skipToken();
//                 m.n=model->newNorm();
//                 m.n->dir[X]=p.getFloat();
//                 m.n->dir[Y]=p.getFloat();
//                 m.n->dir[Z]=p.getFloat();
            }
            if(p.compareToken("usemtl ")) { // a texture spec
                p.skipToken();
                Mystr tintName=p.getString(); // get the name
                usetint=getTintByName(tintName);
//                if(tintName) delete []tintName;
            }
            if(p.compareToken("g ")) { // a group spec
//                ReportSystemError("Reading a group statement");
                p.skipToken();
//                ReportSystemError("Returned from p.skipToken()");
                Mystr boneName=p.getString(); // get the name
//                ReportSystemError("Returned from p.getString");
                usebone=findBoneByName(boneName);
//                ReportSystemError("Returned from model->findBoneByName()");
//                if(boneName) delete []boneName;
//                ReportSystemError("Done processing group statement.");
            }
            if(p.compareToken("mtllib ")) { // specifies an MTL library
                p.skipToken();
                Mystr mtlFileName=p.getString();
                fReader mtlFile;
                mtlFile.open(mtlFileName);
                readMTLFile(mtlFile);                    
            }
            if(p.compareToken("f ")) { // a face object
                INDEX iv,in,iu;
//                iv=0;
                int s=0 /* ,i */ ;
                p.skipToken();
                f=new Face;
                f->data.Texture=usetint;
                f->Owner=usebone;
                Faces.addTail(f);
                while(p.isValid()) {
                    p.parseOBJCorner(iv,iu,in);
                    if(iv) {
                        if(iv<0) iv+=NextVertIndex+1;
                        if(iu<0) iu+=NextUVPtIndex+1;
                            Vert* v=indexVert(iv-1);
                            UVPt* u=indexUVPt(iu-1);
                            Norm* n=indexNorm(in-1);
                            f->addCorner(v,u,n);
                        }
                    }
                    if(f->data.sides<3) {
                        nukeFace(f);
                        f=(Face*)0;
                    }
                }
            }
        }
    updateSkeleton(false);
    assignVerts();

    validate();
    validateMapping();
    destroyAllHistory();
    currentVersion=-1;
    takeSnap();
}

void Model::reindexObjects() {
    INDEX i;
    MYPTR mp;

    STEPBONES(this,mp.b) mp.b->flags|=TF_NEW;

    bool redo;
    i=0;

    do {
        redo=false;
        STEPBONES(this,mp.b)
        if(mp.b->flags&TF_NEW)
        if(mp.b->data.Parent) {
            if((mp.b->data.Parent->flags&TF_NEW)==0) {
                mp.b->index=i++;
                mp.b->flags&=~TF_NEW;
                redo=true;
            }
        } else {
            mp.b->index=i++;
            mp.b->flags&=~TF_NEW;
            redo=true;
        }
    } while(redo);
    NextBoneIndex=i;

    i=0;
    STEPVERTS(this,mp.v) mp.v->index=i++;
    NextVertIndex=i;

    i=0;
    STEPEDGES(this,mp.e) mp.e->index=i++;
    NextEdgeIndex=i;

    i=0;
    STEPFACES(this,mp.f) mp.f->index=i++;
    NextFaceIndex=i;

    i=0;
    STEPTINTS(this,mp.t) mp.t->index=i++;
    NextTintIndex=i;
    AppData::defTint.index=(INDEX)-1;

    i=0;
    STEPUVPTS(this,mp.u) mp.u->index=i++;
    NextUVPtIndex=i;

    i=0;
    STEPNORMS(this,mp.n) mp.n->index=i++;
//    NextUVPtIndex=i;
}

void Model::removeUnusedUVPts() {
    MYPTR m,nm;

    STEPUVPTS(this,m.u)
    m.u->flags&=~TF_NEW;

    STEPFACES(this,m.f)
    if(m.f->data.uvpts)
    for(int i=0;i<m.f->data.sides;i++)
    m.f->data.uvpts[i]->flags|=TF_NEW;

    m.u=FIRSTUVPT(this);
    while(m.u) {
        nm.u=(UVPt*)m.u->next();

        if((m.u->flags&TF_NEW)==0) deleteUVPt(m.u);

        m.u=nm.u;
    }
}

// this establishes whether to create a new Bone object to serve as
// the root of all other objects.
// If there is any vertex with no owner, or if there is more than
// one unparented Bone, then then create a new Bone and parent all
// unparented items to the new Bone

// if there are no Bones at all, then create a new Bone.
#if 0
void Model::rootCheck() {
    Vert*   v;
    bool empty=true;
    bool make=false;

    STEPVERTS(this,v)  {
        empty=false;
        if(v->data.Owner==NULL) make=true;
    }

    int orphan=0;
    Bone*   b;
    Bone*   oldroot=0;
    STEPBONES(this,b) {
        if(b->data.Parent==NULL) {
            oldroot=b;
            orphan++;
        }
        empty=false;
    }

    Bone*   rb=(Bone*)0;
    if(empty || (orphan>1) || make) {
        rb=newBone(0,0);
        defBone=rb;
    } else {
        defBone=oldroot;
    }
// set Bone name here to ModelRoot, counting up from there if taken
    if(rb) {
        rb->data.Name.own(createRootName());
        rb->data.translate.set(0);
        rb->Spot.set(0);
        STEPVERTS(this,v) {
            if(v->data.Owner==NULL) {
                v->data.Owner=rb;
            }
        }
        STEPBONES(this,b) {
            if(b!=rb) {
                if(b->data.Parent==NULL) {
                    b->data.Parent=rb;
                }
            }
        }
        updateSkeleton(true);
    }
}
#endif

void Model::rotateMarked(HWND hwnd,short action) {
    Vert* v;
    Dialex *dx;
    float sinx,siny,sinz;
    float cosx,cosy,cosz;
    Matrix qm;

    dx=new Dialex;

    dx->setDialogLabel(0,"Rotate by");
    dx->setDialogVector(EB_VECX,UserRotate);

    if(DialogBoxParam(UserInt::hInst,MAKEINTRESOURCE(DLG_VECTOR),hwnd,
        InputMatrixDlg,(int)dx)==IDOK) {
        switch(action) {
            case EA_EXTRUDE: extrudeMarked(); break;
            case EA_COPY: copyMarked(); break;
        }
        sinx=UserRotate.x*DEG2RAD;
        siny=UserRotate.y*DEG2RAD;
        sinz=UserRotate.z*DEG2RAD;
        cosx=cosf(sinx);
        sinx=sinf(sinx);
        cosy=cosf(siny);
        siny=sinf(siny);
        cosz=cosf(sinz);
        sinz=sinf(sinz);

        qm.setXVec(cosy*cosz,cosy*sinz,-siny);
        qm.setYVec(sinx*siny*cosz-cosx*sinz,sinx*siny*sinz+cosx*cosz,sinx*cosy);
        qm.setZVec(cosx*siny*cosz+sinx*sinz,cosx*siny*sinz-sinx*cosz,cosx*cosy);
        qm.setLVec(0.0,0.0,0.0);

        STEPVERTS(this,v)
        if(v->flags&TF_MARKED) {
            v->data.Spot*=qm;
        }
        InvalidateView();
    }

    modelFlags|=MF_CALC_SUBD;
    delete dx;
    takeSnap();
}

