// model_s.cpp -- implements member functions of class Model that start with 's'

#include "userint.h"
#include "appdata.h"
#include "mystr.h"
#include "model.h"
#include "dialex.h"
#include "uvpt.h"
#include "vert.h"
#include "edge.h"
#include "face.h"
#include "lsmtypes.h"
#include "lsmres.h"
#include "mouse.h"
#include "funcs.h"

void Model::saveView() {
    switch(EditMode) {
        case EM_MESHEDIT:
            mView=cView;
        break;
        case EM_MAPEDIT:
            uView=cView;
        break;
    }
}

void Model::scaleMarked(HWND hwnd,short action) {
        Vert *v;
        Dialex *dx;

        dx=new Dialex;

        dx->setDialogLabel(0,"Scale by");
        dx->setDialogVector(EB_VECX,UserScale);

        if(DialogBoxParam(UserInt::hInst,MAKEINTRESOURCE(DLG_VECTOR),hwnd,
                InputMatrixDlg,(long int)dx)==IDOK) {
                switch(action) {
                        case EA_EXTRUDE: extrudeMarked(); break;
                        case EA_COPY: copyMarked(); break;
                }
                STEPVERTS(this,v)
                if(v->flags&TF_MARKED) {
                        v->data.Spot.x*=UserScale.x;
                        v->data.Spot.y*=UserScale.y;
                        v->data.Spot.z*=UserScale.z;
                }
                InvalidateView();
        }

        modelFlags|=MF_CALC_SUBD;
        delete dx;
        takeSnap();
}

Vert* Model::selectVert(ULONG i) {
        Vert* v;

        STEPVERTS(this,v)
        if(v->Order==i) return v;

        return NULL;
}

void Model::setFineLevel(HWND hwnd) {
    Dialex *dx;

    dx=new Dialex;
    dx->setDialogInteger(EB_VECX,subPreviewLevel);
    dx->setDialogLabel(0,"Preview refinement");
    dx->setDialogLabel(EB_VECX,"Level:");

    if(DialogBoxParam(UserInt::hInst,MAKEINTRESOURCE(DLG_SNAPLOCK),hwnd,
        InputMatrixDlg,(int)dx)==IDOK) {
        subPreviewLevel=dx->getDialogInteger(EB_VECX);
        InvalidateView();
    }

    delete dx;
}

void Model::setGridSnap(HWND hwnd) {
        Dialex *dx;

        dx=new Dialex;
        switch(EditMode) {
                case EM_MESHEDIT:
                        dx->setDialogFloat(EB_VECX,MeshSnap);
                break;
                case EM_MAPEDIT:
                        dx->setDialogFloat(EB_VECX,MapSnap);
                break;
        }

        if(DialogBoxParam(UserInt::hInst,MAKEINTRESOURCE(DLG_SNAPLOCK),hwnd,
                InputMatrixDlg,(int)dx)==IDOK)
        switch(EditMode) {
                case EM_MESHEDIT:
                        MeshSnap=dx->getDialogFloat(EB_VECX);
                break;
                case EM_MAPEDIT:
                        MapSnap=dx->getDialogFloat(EB_VECX);
                break;
        }

        delete dx;
}

void Model::setWindowTitle(HWND hwnd) {
        Mystr title("The LionSnake Modeler - Model: ");

        title+=name;
        SetWindowText(hwnd,title);
}

// this function splits an Edge
void Model::splitEdge(Edge* e,bool use_subd) {

// identify any Joins that lie on the same side of each bordering face as this Edge
    Join*   jl=0;
    Face* fl=e->FaceL;
    if(fl)
    if(fl->data.uvpts) {
        bool bad=true;
        for(int i=0;i<fl->data.sides;i++) {
            if(e==fl->edges[i]) jl=fl->joins[i];
            bad=false;
        }
//        if(bad) SetWindowText(UserInt::hwndMain,"bad");
//        if(jl==0) SetWindowText(UserInt::hwndMain,"jl==0");
    }

    UVPt*   ulm=0;
    if(jl) {
        ulm=newUVPt();
        if(use_subd) {
            ulm->data.Spot=jl->getSub1(); // use result of first subdivision
        } else {
            ulm->data.Spot.u=(jl->UVPtS->data.Spot.u+jl->UVPtE->data.Spot.u)*.5;
            ulm->data.Spot.v=(jl->UVPtS->data.Spot.v+jl->UVPtE->data.Spot.v)*.5;
        }
        newJoin(ulm,jl->UVPtS);
        newJoin(ulm,jl->UVPtE);
    }

    Join*   jr=0;
    Face* fr=e->FaceR;
    if(fr)
    if(fr->data.uvpts) {
        for(int i=0;i<fr->data.sides;i++) {
            if(e==fr->edges[i]) jr=fr->joins[i];
        }
//        if(jr==0) SetWindowText(AppData::hwndMain,"fr has no jr");
    }

    UVPt*   urm=0;
    if(jr) {
        if(jr!=jl) {
            urm=newUVPt();
            if(use_subd) {
                urm->data.Spot=jr->getSub1(); // use result of first subdivision
            } else {
                urm->data.Spot.u=(jr->UVPtS->data.Spot.u+jr->UVPtE->data.Spot.u)*.5;
                urm->data.Spot.v=(jr->UVPtS->data.Spot.v+jr->UVPtE->data.Spot.v)*.5;
            }
            newJoin(ulm,jr->UVPtS);
            newJoin(ulm,jr->UVPtE);
        } else {
            urm=ulm;
        }
    }

// create the new vertex at the middle of the edge
    Vert* vm=newVert(0,0);
    if(use_subd) {
        vm->data.Spot=e->getSub1();
    } else {
        ((vm->data.Spot=e->data.VertS->data.Spot)+=e->data.VertE->data.Spot)*=.5;
    }
    vm->flags=TF_NEW;

// create the new edges that connect it in place
    Edge* es=newEdge(e->data.VertS,vm);
    es->flags=TF_NEW;
    es->setSharpLevel(e->getSharpLevel());

    Edge* ee=newEdge(vm,e->data.VertE);
    ee->flags=TF_NEW;
    ee->setSharpLevel(e->getSharpLevel());

    es->FaceL=ee->FaceL=fl;
    es->FaceR=ee->FaceR=fr;

// process the face on the left side
    if(fl) fl->insertCorner(e,vm,ulm);

// and now process the face on the right side
    if(fr) fr->insertCorner(e,vm,urm);

    e->FaceL=e->FaceR=0;
    deleteEdge(e);

    if(jl) {
        jl->FaceL=jl->FaceR=0;
        nukeJoin(jl);
    }

    if(jr && jr!=jl) {
        jr->FaceL=jr->FaceR=0;
        nukeJoin(jr);
    }

    modelFlags|=MF_BUILD_SUBD; //|MF_REBUILD_NORMS;
    InvalidateView();
}

void Model::splitFace(Face* f,bool use_subd) {
        int i,ip;
        Face*   nf;
        UVPt*   um=0;
        Vert*   vm;
        Edge*   et;
        Edge*   ef;

        vm=newVert(0,0);
        if(use_subd) {
            vm->data.Spot=f->getSub1();
        } else {
            f->calcCenter(vm->data.Spot);
        }
        vm->flags|=TF_NEW;

        if(f->data.uvpts) {
            um=newUVPt();
            if(use_subd) {
                um->data.Spot=f->getUSub1();
            } else {
                f->calcMapCenter(um->data.Spot);
            }
        }

        ef=
        et=newEdge(vm,f->data.verts[f->data.sides-1]);
        et->flags|=TF_NEW;

        for(i=0;i<f->data.sides;i+=2) {
                ip=i-1;
                if(i==0) ip=f->data.sides-1;
                nf=new Face(4);
                Faces.addTail(nf);
                nf->edges=new Edge*[4];
                nf->flags|=TF_NEW;
// set new face vertices
                nf->data.verts[0]=
                nf->data.verts[4]=f->data.verts[i];
                nf->data.verts[1]=f->data.verts[i+1];
                nf->data.verts[2]=vm;
                nf->data.verts[3]=f->data.verts[ip];
// and assign some edges
// first, make face use these edges
                nf->edges[0]=f->edges[i];
                nf->edges[3]=f->edges[ip];
                nf->edges[2]=et;    // keep this before reassignment below!
                if(i==f->data.sides-2) {
                    nf->edges[1]=ef;
                } else {
                    et=
                    nf->edges[1]=newEdge(vm,f->data.verts[i+1]);
                    et->flags|=TF_NEW;
                }
// then make edges aware that face borders on them
                nf->edges[0]->replaceFace(f,nf);
                nf->edges[1]->replaceFace(NULL,nf);
                nf->edges[2]->replaceFace(NULL,nf);
                nf->edges[3]->replaceFace(f,nf);

// copy UV-mapping data
                if(f->data.uvpts) {
                        nf->data.uvpts=new UVPt*[4];
                        nf->data.uvpts[0]=f->data.uvpts[i];
                        nf->data.uvpts[1]=f->data.uvpts[i+1];
                        nf->data.uvpts[2]=um;
                        nf->data.uvpts[3]=f->data.uvpts[ip];
                }

// ensure children inherit some of the parent stuff
                nf->data.Texture=f->data.Texture;
        }
        deleteFace(f);
        modelFlags|=MF_BUILD_SUBD;//|MF_REBUILD_NORMS;
    InvalidateView();
}

void Model::splitHigh(bool use_subd) {
    clearAll();

    switch(HighType) {
        case HT_EDGE:
            splitEdge(HighItem.e,false);
            validateMapping();
            validate();
            takeSnap();
        break;
        case HT_FACE:
            for(int i=0;i<HighItem.f->data.sides;i++)
            HighItem.f->edges[i]->flags|=TF_MARKED;

            HighItem.f->flags|=TF_MARKED;

            splitMarked(use_subd);
        break;
    }
    clearAll();
}

void Model::splitMarked(bool use_subd) {
    Edge*   e;
    Edge*   ne;
    Face*   f;
    Face*   nf;
    Vert*   v;
    UVPt*   u;

    if(use_subd) {
        modelFlags|=MF_BUILD_SUBD;
        buildSubData();
        calcSubData();
    }

// indicate that marked vertices are to retain marking
    STEPVERTS(this,v) {
        v->flags&=~TF_NEW;
        if(v->flags&TF_MARKED)
        v->flags|=TF_NEW;
    }

// clear marking of texture mapping points
    STEPUVPTS(this,u) u->flags&=~TF_MARKED;

// make sure that unaffected edges don't gain marking
    STEPEDGES(this,e) e->flags&=~TF_NEW;

// make sure that unaffected faces don't gain marking
    STEPFACES(this,f) {
        f->flags &= ~TF_NEW;

        if(f->data.uvpts)
        for(int i=0;i<f->data.sides;i++)
        if(f->data.verts[i]->flags&TF_MARKED)
        f->data.uvpts[i]->flags|=TF_MARKED;
    }

    e=FIRSTEDGE(this);
    while(e) {
        ne=(Edge*)(e->next());
        if(e->flags&TF_MARKED)
        splitEdge(e,use_subd);
        e=ne;
    }

    f=FIRSTFACE(this);
    while(f) {
        nf=(Face*)(f->next());
        if(f->flags & TF_MARKED)
        splitFace(f,use_subd);
        f=nf;
    }

    if(use_subd) {
      STEPVERTS(this,v)
      if(v->flags&TF_MARKED) {
          v->data.Spot=v->getSub1();
      }
      STEPUVPTS(this,u)
      if(u->flags&TF_MARKED) {
          u->data.Spot=u->getSub1();
      }
    }

    STEPVERTS(this,v) {
        if(v->flags&TF_NEW)
        v->flags|=TF_MARKED;
    }

    STEPEDGES(this,e) {
        if(e->flags&TF_NEW)
        e->flags|=TF_MARKED;
    }

    STEPFACES(this,f) {
        if(f->flags&TF_NEW)
        f->flags|=TF_MARKED;
    }

    validateMapping();
    validate();
    takeSnap();
//    modelFlags|=MF_BUILD_SUBD;//|MF_REBUILD_NORMS;
}
