// model_u.cpp -- implements member functions of class Model that start with 'u'
#include "userint.h"
#include "appdata.h"
#include "model.h"
#include "bone.h"
#include "vert.h"
#include "edge.h"
#include "face.h"
#include "lsmres.h"
#include "mouse.h"

void Model::unhideAll() {
    Vert*   v;
    Edge*   e;
    Face*   f;

    STEPVERTS(this,v)
    v->flags&=~TF_HIDDEN;
    
    STEPEDGES(this,e)
    e->flags&=~TF_HIDDEN;

    STEPFACES(this,f)
    f->flags&=~TF_HIDDEN;
}

void Model::updateMenu() {
        MENUITEMINFO qmii;

        memset(&qmii,0,sizeof(MENUITEMINFO));

        qmii.cbSize=sizeof(MENUITEMINFO);
        qmii.fMask=MIIM_STATE;

        qmii.fState=(modelFlags&MF_SHOW_PARTS)?MFS_CHECKED:MFS_UNCHECKED;
        SetMenuItemInfo(UserInt::ModelMenu,IDM_SHOWPARTS,FALSE,&qmii);

        qmii.fState=(modelFlags&MF_SHOW_VERTS)?MFS_CHECKED:MFS_UNCHECKED;
        SetMenuItemInfo(UserInt::ModelMenu,IDM_SHOWVERTS,FALSE,&qmii);

        qmii.fState=(modelFlags&MF_SHOW_EDGES)?MFS_CHECKED:MFS_UNCHECKED;
        SetMenuItemInfo(UserInt::ModelMenu,IDM_SHOWEDGES,FALSE,&qmii);

        qmii.fState=(modelFlags&MF_SHOW_FACES)?MFS_CHECKED:MFS_UNCHECKED;
        SetMenuItemInfo(UserInt::ModelMenu,IDM_SHOWFACES,FALSE,&qmii);

        qmii.fState=(modelFlags&MF_SHOW_SUBD)?MFS_CHECKED:MFS_UNCHECKED;
        SetMenuItemInfo(UserInt::ModelMenu,IDM_SHOWSUBD,FALSE,&qmii);

        qmii.fState=(projectFlags&PF_SHOWAXES)?MFS_CHECKED:MFS_UNCHECKED;
        SetMenuItemInfo(UserInt::ModelMenu,IDM_SHOWAXES,FALSE,&qmii);

        qmii.fState=(modelFlags&MF_LOCK_X)?MFS_CHECKED:MFS_UNCHECKED;
        SetMenuItemInfo(UserInt::ModelMenu,IDM_LOCK_X,FALSE,&qmii);

        qmii.fState=(modelFlags&MF_LOCK_Y)?MFS_CHECKED:MFS_UNCHECKED;
        SetMenuItemInfo(UserInt::ModelMenu,IDM_LOCK_Y,FALSE,&qmii);

        qmii.fState=(modelFlags&MF_LOCK_Z)?MFS_CHECKED:MFS_UNCHECKED;
        SetMenuItemInfo(UserInt::ModelMenu,IDM_LOCK_Z,FALSE,&qmii);

    qmii.fState=(EditMode==EM_MESHEDIT)?MFS_CHECKED:MFS_UNCHECKED;
    SetMenuItemInfo(UserInt::ModelMenu,IDM_MESHMODE,FALSE,&qmii);

    qmii.fState=(EditMode==EM_MAPEDIT)?MFS_CHECKED:MFS_UNCHECKED;
    SetMenuItemInfo(UserInt::ModelMenu,IDM_MAPMODE,FALSE,&qmii);

    qmii.fState=(MouseMode==MM_ADDMESH0)?MFS_CHECKED:MFS_UNCHECKED;
    SetMenuItemInfo(UserInt::ModelMenu,IDM_ADDMESH,FALSE,&qmii);

    qmii.fState=(MouseMode==MM_ADDVERT)?MFS_CHECKED:MFS_UNCHECKED;
    SetMenuItemInfo(UserInt::ModelMenu,IDM_ADDVERT,FALSE,&qmii);

    qmii.fState=(MouseMode==MM_ADDBONE)?MFS_CHECKED:MFS_UNCHECKED;
    SetMenuItemInfo(UserInt::ModelMenu,IDM_ADDBONE,FALSE,&qmii);
}

void Model::updateSkeleton(bool and_verts) {
        Vert*   v;
        Bone*   b;
        bool redo;

        if(and_verts)
        STEPVERTS(this,v)
        if(v->data.Owner)
        (v->Save = v->data.Spot)/=v->data.Owner->Scaled;

        STEPBONES(this,b)
        b->flags|=TF_NEW;

        do {
            redo=false;

            STEPBONES(this,b)
            if(b->flags&TF_NEW) {
                if(b->data.Parent)
                if(b->data.Parent->flags&TF_NEW)
                continue;

                b->buildTransforms();
                b->flags&=~TF_NEW;
                redo=true;
            }
        } while(redo);

        if(and_verts) {
            modelFlags|=MF_CALC_SUBD;

            STEPVERTS(this,v)
            if(v->data.Owner)
            (v->data.Spot = v->Save)*=v->data.Owner->Scaled;
        }
}

