// model_v.cpp -- implements Model member functions starting with 'v'

#include "model.h"
#include "vert.h"
#include "uvpt.h"
#include "join.h"
#include "edge.h"
#include "face.h"
#include "funcs.h"

void Model::validate() {
    Face* f;
    Face*   nf;
    Edge* e;
    Vert* v;

    STEPVERTS(this,v) v->Edges.clear();

    STEPEDGES(this,e) {
        e->FaceL=e->FaceR=NULL;
        e->data.VertS->addEdge(e);
        e->data.VertE->addEdge(e);
    }

//      ReportSystemError("Vert-Edge cross-referencing rebuilt.");

    STEPFACES(this,f) {
        f->flags&=~TF_NEW;
        if(!(f->edges)) f->edges=new Edge*[f->data.sides];

        for(int i=0;i<f->data.sides;i++) {
            f->edges[i]=(Edge*)0;
            e=getEdge(f->data.verts[i],f->data.verts[i+1]);

            f->edges[i]=e;
//              if(e==NULL) ReportSystemError("About to boom");
            if(e->FaceL==NULL) {
                e->FaceL=f;
            } else {
                if(e->FaceR!=NULL) {
                    f->flags|=TF_NEW;
                } else {
                    e->FaceR=f;
                }
            }
        }
    }

    f=FIRSTFACE(this);
    while(f) {
        nf=(Face*)f->next();

        if(f->flags&TF_NEW) {
            deleteFace(f);
//                UpdateStatusBar("Deleted a face during validation",0,0);
        }
        f=nf;
    }

//      ReportSystemError("Face-Edge cross-referencing rebuilt.");

    STEPEDGES(this,e) {
        e->flags&=~(EF_NO_LEFT|EF_NO_RIGHT);
        if(e->FaceL==NULL) e->flags|=EF_NO_LEFT;
        if(e->FaceR==NULL) e->flags|=EF_NO_RIGHT;
    }

    projectFlags&=~PF_MODIFIED;
    modelFlags|=MF_REORIENT;
}

void Model::validateMapping() {
    UVPt*   u;
    Join*   j;
    Face*   f;
    int i,ii;

// flag all mapping points as unused, remove all uvpt-join referencing
    STEPUVPTS(this,u) {
        u->joins.clear();
        u->flags&=~TF_NEW;
    }

// empty all face-join cross-referencing, rebuild uvpt-join referencing
    STEPJOINS(this,j) {
        j->FaceL=j->FaceR=0;
        j->UVPtS->addJoin(j);
        j->UVPtE->addJoin(j);
    }

    STEPFACES(this,f)
    if(f->data.uvpts) {
        if(f->joins==0) f->joins=new Join*[f->data.sides];
        for(i=0;i<f->data.sides;i++) {
            u=f->data.uvpts[i];
            ii=i+1; if(ii==f->data.sides) ii=0;
            j=getJoin(u,f->data.uvpts[ii]);
            f->joins[i]=j;
            if(j->FaceL==0) {
                j->FaceL=f;
            } else {
                j->FaceR=f;
            }
            u->flags|=TF_NEW;
        }
    }

    Join*   nj;
    j=FIRSTJOIN(this);
    while(j) {
        nj=(Join*)j->next();
        if(j->FaceL==0 && j->FaceR==0) nukeJoin(j);
        j=nj;
    }

    UVPt*   nu;
    u=FIRSTUVPT(this);
    while(u) {
        nu=(UVPt*)u->next();
        if((u->flags&TF_NEW)==0)
        deleteUVPt(u);
        u=nu;
    }
}
    
