// model_w.cpp -- implements Model functions starting with 'w'

#include "fhand.h"
#include "model.h"
#include "fhand.h"
#include "bone.h"
#include "uvpt.h"
#include "tint.h"
#include "edge.h"
#include "face.h"
#include "vert.h"
#include "funcs.h"
#include "appdata.h"

void Model::writeINCFile(fWriter& file) {
        Bone* b;
        Vert* v;
        Face* f;
        Edge* e;
        Tint *t;
        UVPt*   u;
        char *c;
        int    i;
        BOOL drop;
        Vec3 redux;

        removeUnusedUVPts();
        destroyAllHistory();
        currentVersion=(INDEX)-1;
        takeSnap();
        reindexObjects();

        file << "#include \42transforms.inc\42\15\12#include \42c2s3.inc\42\15\12";

        STEPBONES(this,b) {
                c=b->data.Name;
// write ifndef block for scaling
                file << "#ifndef(" << c << "_s) #local " << c << "_s=<"
                     << b->data.scale.x << ','
                     << b->data.scale.y << ','
                     << b->data.scale.z << ">; #end\15\12";
                file << "#ifdef(" << c << "_ds) #local " << c << "_s=" << c << "_s * " << c
                     << "_ds; #end\15\12";

// write ifndef block for twisting
                file << "#ifndef(" << c << "_az) #local " << c << "_az="
                     << b->data.twistAngle << "; #end\15\12";
                file << "#ifdef(" << c << "_dz) #local " << c << "_az=" << c << "_az + " << c
                     << "_dz; #end\15\12";

// write ifndef block for bending
                file << "#ifndef(" << c << "_ax) #local " << c << "_ax="
                     << b->data.bendAngle << "; #end\15\12";
                file << "#ifdef(" << c << "_dx) #local " << c << "_ax=" << c << "_ax + " << c
                     << "_dx; #end\15\12";

// write ifndef block for turning
                file << "#ifndef(" << c << "_ay) #local " << c << "_ay="
                     << b->data.turnAngle << "; #end\15\12";
                file << "#ifdef(" << c << "_dy) #local " << c << "_ay=" << c << "_ay + " << c
                     << "_dy; #end\15\12";

// write ifndef block for translate
                file << "#ifndef(" << c << "_t) #local " << c << "_t=<"
                     << b->data.translate.x << ','
                     << b->data.translate.y << ','
                     << b->data.translate.z << ">; #end\15\12";
                file << "#ifdef(" << c << "_dt) #local " << c << "_t=" << c << "_t + " << c
                     << "_dt; #end\15\12";

                file << "\15\12";
        }

// write part build for Bones
        for(i=0;i<NextBoneIndex;i++) {
                b=indexBone(i);
                if(b) b->writeINCBuild(file);
        }

        file << "\15\12";

        i=0;
        STEPVERTS(this,v) {
                i++;
                file << "#local pV";
                file.writeHexString(v->index);
                file << "=";
                if(v->data.Owner==NULL) {
                        file.writeVecString(v->data.Spot);
                        file << ";\15\12";
                } else {
                        c=v->data.Owner->data.Name;
                        file << "vtransform(";
                        (redux = v->data.Spot)/=v->data.Owner->Scaled;
                        file.writeVecString(redux);
                        file << "," << c << "_tfs);\15\12";
                }
        }

        if(NextVertIndex) {
//      write the vertices array
                file << "\15\12#local apV=array[";
                file << NextVertIndex;
                file << "] {";

                i=0;
                STEPVERTS(this,v) {
                        if(i) file << ",";
                        if(i==17 || i==0) { file << "\15\12"; i=1; }
                        file << "pV";
                        file.writeHexString(v->index);
                        i++;
                }
                file << "\15\12}\15\12";
        }

        if(NextFaceIndex) {
            i=NextFaceIndex;
            STEPFACES(this,f) i+=f->data.sides;

// write the faces array
                file << "\15\12#local aiF=array[";
                file << i;
                file << "] {\15\12";

                drop=false;
                STEPFACES(this,f) {
                        if(drop) file << ",\15\12";
                        drop=true;
                        file << f->data.sides;
                        for(i=0;i<f->data.sides;i++) {
                            file << ", ";
                            file << f->data.verts[i]->index;
                        }
                }
                file << "\15\12}\15\12\15\12c2s3_ImportBasicMesh(apV,aiF)\15\12";
        }

        i=0;
        STEPEDGES(this,e)
        if(e->getSharpLevel())
        i++;

        if(i) {
                drop=FALSE;
                file << "\15\12#local aiS=array[";
                file << i;
                file << "][3] {";
                STEPEDGES(this,e)
                if(e->getSharpLevel()) {
                        if(drop) file << ',';
                        drop=TRUE;
                        file << "\15\12{ ";
                        file << e->data.VertS->index;
                        file << ',';
                        file << e->data.VertE->index ;
                        file << ',';
                        file << e->getSharpLevel();
                        file << '}';
                }
                file << "\15\12}\15\12\15\12c2s3_AddSharpData(aiS)\15\12\15\12";
        }

// write uv-mapping data
        i=0;
        STEPFACES(this,f)
        if(f->data.uvpts)
        i+=f->data.sides;

        if(i) { // yes, there's uv-mapping
// write the array of uv points first
            file << "#local aU=array[";
            file << NextUVPtIndex;
            file << "] {\15\12";
            drop=false;
            STEPUVPTS(this,u) {
                if(drop) file << ",\15\12";
                file << u->data.Spot;
                drop=true;
            }
            file << "}\15\12\15\12";
// and then write the mapping index array
            drop=false;
                file << "\15\12#local aiM=array[";
                file << i+NextFaceIndex;
                file << "] {\15\12";

                STEPFACES(this,f) {
                        if(drop) file << ",\15\12";
                        drop=true;
                        if(f->data.uvpts) {
                                file << f->data.sides;
                            for(i=0;i<f->data.sides;i++)
                            file << ',' << f->data.uvpts[i]->index;
                        } else {
                                file << "0";
                        }
                }
                file << "\15\12}\15\12\15\12c2s3_AddTextureMapping(aU,aiM)\15\12";
        }

// now write texturing data
        i=0;
        STEPFACES(this,f)
        if(f->data.Texture!=&AppData::defTint)
        i++;

        if(i) { // yes, there was a face with texturing data
                i=0;
                file << "\15\12#local aiT=array[";
                file << NextFaceIndex;
                file << "] {";
                STEPFACES(this,f) {
                        if(i) file << ",";
                        if(i==17 || i==0) { file << "\15\12"; i=1; }
                        i++;
                        if(f->data.Texture!=&AppData::defTint) {
                                file << f->data.Texture->index;
                        } else {
                                file << "-1";
                        }
                }
                file << "\15\12}\15\12\15\12c2s3_AddTextures(aiT)\15\12\15\12";
// write comment lines that state which texture name goes with which texture in the array
                STEPTINTS(this,t)
                if(t!=&AppData::defTint) {
                        file << "// texture " << t->index
                             << " is named " << t->data.Name
                             << " in the model.\15\12";
                }
        }
}

void Model::writeMTLFile(fWriter& file) {
//        Face*   f;
        Tint*   t;

// write a comment
        file << "# created by LionSnake v 1.7\n" ;

        STEPTINTS(this,t) {
                file << "newmtl " << t->data.Name << '\n' ;
                file << "Kd " ;
                file << GetRValue(t->data.color)/255.0f ;
                file << ' ' ;
                file << GetGValue(t->data.color)/255.0f ;
                file << ' ' ;
                file << GetBValue(t->data.color)/255.0f ;
                file << '\n' ;

                if(t->data.File.len())
                file << "map_Kd \"" << t->data.File << "\"\n" ;
        }
}

/* MTL file stuff
newmtl name
    Start the definition of a new named material property.

Ka float float float
    Ambient color.

Kd float float float
    Diffuse color.

Ks float float float
    Specular color.

illum (0, 1, or 2)
        0 Color on and Ambient off
        1 Color on and Ambient on
        2 Highlight on
        3 Reflection on and Ray trace on
        4 Transparency: Glass on
          Reflection: Ray trace on
        5 Reflection: Fresnel on and Ray trace on
        6 Transparency: Refraction on
          Reflection: Fresnel off and Ray trace on
        7 Transparency: Refraction on
          Reflection: Fresnel on and Ray trace on
        8 Reflection on and Ray trace off
        9 Transparency: Glass on
          Reflection: Ray trace off
        10 Casts shadows onto invisible surfaces

map_Kd filename
    Texture map. Supported file types are .gif, .jpg, .xbm, .rgb, .rgba, .int, .inta, .sgi, and .bw.

d a
Alpha value (0.0 == transparent, 1.0==fully opaque)

Ns v
specular reflection.  v is inverse of POV-Ray roughness for specular.

Ni 1
Refraction index. Values range from 1 upwards. A value of 1 will cause no
refraction. A higher value will cause refraction.

Tf r g b
transmission filter (light passing is scaled by r,g,b)

*/

void Model::writeOBJFile(fWriter& file,char* libName) {
        Vert*   v;
        Face*   f;
        Tint*   t;
        UVPt*   u;
//        Edge*   e;
//      Norm*   n;
        Bone*   b;
        bool write;
//        short type;

        orientWindings();
        assignFaces();
//buildNorms();
//      calcNorms();
//      STEPNORMS(this,n) Normalize(n->dir,NULL);

// write a comment
        file << "# created by LionSnake v 1.7\n";

// write MTLLIB statement
        if(libName)
        file << "mtllib " << libName << '\n';

// write the vertex data
        STEPVERTS(this,v) {
                file << "v "
                /* file */ << v->data.Spot.x;
                file << ' ';
                file << v->data.Spot.y;
                file << ' ';
                file << v->data.Spot.z;
                file << '\n';
        }

// write the uv-mapping data
        STEPUVPTS(this,u) {
                file << "vt "
                         << frac(u->data.Spot.u)
                         << ' '
                         << frac(u->data.Spot.v)
                         << '\n';
        }

// write the normal data
//      STEPNORMS(this,n) {
//              WriteString(fhand,"vn ");
//              WriteFloatString(fhand,n->dir[X]);
//              WriteSpace(fhand);
//              WriteFloatString(fhand,n->dir[Y]);
//              WriteSpace(fhand);
//              WriteFloatString(fhand,n->dir[Z]);
//              WriteCR(fhand);
//      }

// write the faces without any group info
// starting with the faces that have no textures
        STEPFACES(this,f)
        if(f->Owner==NULL)
        if(f->data.Texture==&AppData::defTint)
        f->writeToOBJFile(file);

// write the textured faces
        STEPTINTS(this,t) {
                write=false;
                STEPFACES(this,f)
                if(f->Owner==NULL)
                if(f->data.Texture==t)
                write=true;

                if(write) {
                        file << "usemtl "
                                 << t->data.Name
                                 << '\n' ;
                        STEPFACES(this,f)
                        if(f->Owner==NULL)
                        if(f->data.Texture==t)
                        f->writeToOBJFile(file);
                }
        }

// and now do the same for the grouped stuff
        STEPBONES(this,b) {
                file << "g " << b->data.Name << '\n' ;

// write untextured stuff first
                write=false;
                STEPFACES(this,f)
                if(f->Owner==b)
                if(f->data.Texture==NULL)
                write=true;

                if(write) {
                        file << "usemtl\n";
                        STEPFACES(this,f)
                        if(f->Owner==b)
                        if(f->data.Texture==&AppData::defTint)
                        f->writeToOBJFile(file);
                }
// and now the textured ones
                STEPTINTS(this,t) {
                        write=false;
                        STEPFACES(this,f)
                        if(f->Owner==b)
                        if(f->data.Texture==t)
                        write=true;

                        if(write) {
                                file << "usemtl " << t->data.Name << '\n' ;
                                STEPFACES(this,f)
                                if(f->Owner==b)
                                if(f->data.Texture==t)
                                f->writeToOBJFile(file);
                        }
                }
        }

// now write the sharp edges
// needed for edges with defined sharpness and orphan edges
//      STEPEDGES(this,e) {
//              if(e->data.sharpness!=0 || e->sharpType()==-3) {
//                      file << "#e ";
//                      file << e->data.VertS->index+1 ;
//                      file << ' ';
//                      file << e->data.VertE->index+1 ;
//                      file << ' ';
//                      file << e->data.sharpness;
//                      file << '\n';
//              }
//      }
}
