// parser.cpp -- implements the Parser object
#include <windows.h>
#include <string.h>
#include "parser.h"

Parser::Parser(char* use_buffer) {
        buff=use_buffer;
        ptr=0;

        if(buff) {
                len=strlen(buff);
                good=true;
        } else {
                len=0;
                good=false;
        }
}

// Parser::~Parser() { }

bool Parser::compareToken(char* t) {
        if(t==NULL) return false;

        if(*t=='\0') return false;

        for(int i=0;t[i]!='\0';i++)
        if(t[i]!=buff[ptr+i]) return false;

        return true;
}

void Parser::reset() { ptr=0; good=true; }

void Parser::skipToken() {
        if(good==false) return;

        while(true)
        switch(buff[ptr]) {
                case '\0':
                        good=false;
                        return;
                case ' ':
                case '\011':
                        ptr++;
                        return;
                default:
                        ptr++;
        }
}

void Parser::skipWhiteSpace() {
    if(good==false) return;

    bool repeat=true;
    while(repeat) {
        switch(buff[ptr]) {
            case ' ':
            case '\011':
                ptr++;
                break;
            case '\0':
                good=false;
            default:
                repeat=false;
        }
    }
}

char Parser::getChar() {
        char result;

        if(good==false) return '\0';

        result=buff[ptr++];

        if(ptr>=len) good=false;

        return result;
}

INDEX Parser::getIndex() {
        int i;

        if(good==false) return (INDEX)-1;

        skipWhiteSpace();

        i=atoi(buff+ptr);

        skipToken();

        return (INDEX)i;
}

bool Parser::getBool() {
        return (bool)getIndex();
}

float Parser::getFloat() {
        float f;

        if(good==false) return 0.0;

        skipWhiteSpace();

        f=atof(buff+ptr);

        skipToken();

        return f;
}

char* Parser::getString() {
        char*   result;
        char* end;
        char delim;
        int length;
//        int skip;

        if(good==false) return (char*)0;

        skipWhiteSpace();

        switch(buff[ptr]) {
                case '\'':
                        delim = '\'';
                        ptr++;
                break;
                case '\"':
                        delim = '\"';
                        ptr++;
                break;
                default:
                delim=' ';
        }

        end=strchr(buff+ptr,delim);
        if(end==NULL) {
                end=strchr(buff+ptr,'\0');
        }

        if(end==NULL) return (char*)0;

        length=end-buff-ptr;

        delim=*end;
        *end='\0';

        if(length) {
                result=new char[length+1];
                for(int i=0;i<length;i++)
                result[i]=buff[i+ptr];
                result[length]='\0';
        } else {
                result=(char*)0;
        }

        *end=delim;

        ptr=end-buff;

        if(buff[ptr]=='\'') ptr++;
        if(buff[ptr]=='\"') ptr++;

        while(buff[ptr]==' ') ptr++;

        return result;
}

#if(0)
bool Parser::readFloat(float& result) {
        double  temp;

        if(fscanf((FILE*)fhand,"%f",&temp)==EOF) EOFyet=true;

        result=(float)temp;

        return !EOFyet;
}

char* Parser::getString() {

        if(fscanf((FILE*)fhand,"\"%s\"",buff)==EOF) EOFyet=true;

        if(EOFyet) return (char*)0;

        return CloneString(buff);
}
#endif

void Parser::parseOBJCorner(INDEX& iv,INDEX& it,INDEX& in) {
        int val;
        char c;
        bool loop;

        iv=it=in=0;

        skipWhiteSpace();

        for(int i=0;i<3;i++) {
                val=0;
                do {
                        loop=false;
                        c=getChar();
                        if(c>='0' && c<='9') {
                                val*=10;
                                val+=c&0xf;
                                loop=true;
                        }
                } while(loop);
                switch(i) {
                        case 0: iv=val; break;
                        case 1: it=val; break;
                        case 2: in=val; break;
                }
                if(c!='/') return;
        }
}
