// patch.cpp -- implements the Patch class

#include <windows.h>
#include <gl/glu.h>
//#include <gl/glext.h>
#include "lsmtypes.h"
#include "patch.h"
#include "funcs.h"
#include "vert.h"
#include "edge.h"
#include "face.h"
#include "tint.h"
#include "join.h"
#include "uvpt.h"

void Patch::calcSubf() {
    Vec3    va=v->getSub1();
    Vec3    vd=en->getSub1();
    Vec3    vc=f->getSub1();
    Vec3    vb=ep->getSub1();

    ((((((v_v=va)*=2.0f)+=vd)+=vb)*=2.0f)+=vc)*=.111111111f;
    ((((((v_n=vd)*=2.0f)+=va)+=vc)*=2.0f)+=vb)*=.111111111f;
    ((((((v_f=vc)*=2.0f)+=vb)+=vd)*=2.0f)+=va)*=.111111111f;
    ((((((v_p=vb)*=2.0f)+=va)+=vc)*=2.0f)+=vd)*=.111111111f;

    en->getSubf()+=v_n;
    ep->getSubf()+=v_p;
    f->getSubf()+=v_f;
    v->getSubf()+=v_v;
}

void Patch::draw(int sub_preview_level,bool useTexture) {
    static float gvec[48];
    static float tvec[8];
    Vec3    va,vb,vc,vd;
    Vec3    ve,vf; //,vg,vh;
//    Vec3    vi;

//    Vec2    ta,tb,tc,td;
//    Vec2    te,tf,tg,th;
//    Vec2    ti;

    Tint*   t;

    if(f->flags&TF_HIDDEN) return;

    t=f->data.Texture;
    if(t==0) return;

    v->getSubf().put(gvec);
    en->getSubf().put(gvec+9);
    ep->getSubf().put(gvec+36);
    f->getSubf().put(gvec+45);

    va=v->getSub1();
    vd=en->getSub1();
    vc=f->getSub1();
    vb=ep->getSub1();

    v_v.put(gvec+15);
    v_n.put(gvec+18);
    v_f.put(gvec+30);
    v_p.put(gvec+27);

    if(n0) { // if there's a neighbor 0
        ((ve=n0->v_v)+=v_v)*=.5;
        ((vf=n0->v_p)+=v_n)*=.5;
    } else {
        (((ve=va)*=2.0f)+=vd)*=.33333333f;
        (((vf=vd)*=2.0f)+=va)*=.33333333f;
    }
    ve.put(gvec+3);
    vf.put(gvec+6);

    ((ve=n1->v_p)+=v_n)*=.5f;
    ((vf=n1->v_f)+=v_f)*=.5f;
    ve.put(gvec+21);
    vf.put(gvec+33);

    ((ve=n2->v_f)+=v_f)*=.5f;
    ((vf=n2->v_n)+=v_p)*=.5f;
    ve.put(gvec+42);
    vf.put(gvec+39);

    if(n3) { // if there's a neighbor 3
        ((ve=n3->v_n)+=v_p)*=.5f;
        ((vf=n3->v_v)+=v_v)*=.5f;
    } else {
        (((ve=vb)*=2.0f)+=va)*=.33333333f;
        (((vf=va)*=2.0f)+=vb)*=.33333333f;
    }
    ve.put(gvec+24);
    vf.put(gvec+12);

    glMap2f(GL_MAP2_VERTEX_3,0.0f,1.0f,3,4,0.0f,1.0f,12,4,gvec);
    glEnable(GL_MAP2_VERTEX_3);

    if(u==NULL && useTexture==true)
    SelectGLColor(f->data.Texture->data.color);

    if(u==NULL || useTexture==false) {
        glDisable(GL_TEXTURE_2D);
        glDisable(GL_MAP2_TEXTURE_COORD_2);
    } else {
        u->getSub1().put(tvec);
        jn->getSub1().put(tvec+2);
        jp->getSub1().put(tvec+4);
        f->getUSub1().put(tvec+6);
        SelectGLColor((COLORREF)0xffffff);
        if(t->bmBits==NULL) t->getTintBitmap();
        t->selectIntoOpenGL();
        glEnable(GL_TEXTURE_2D);
        glMap2f(GL_MAP2_TEXTURE_COORD_2,0.0f,1.0f,2,2,0.0f,1.0f,4,2,tvec);
        glEnable(GL_MAP2_TEXTURE_COORD_2);

    }

    if(useTexture) {
        glEnable(GL_AUTO_NORMAL);
        glShadeModel(GL_SMOOTH);
    } else {
        glDisable(GL_AUTO_NORMAL);
        glShadeModel(GL_FLAT);
    }

    glMapGrid2f(sub_preview_level,0.0f,1.0f,sub_preview_level,1.0f,0.0f);
    glEvalMesh2(GL_FILL,0,sub_preview_level,0,sub_preview_level);
}

Patch::Patch(Face* uf,int i) : f(uf) {
    int j=((i==0)?uf->data.sides-1:i-1);

    v=uf->data.verts[i];
    en=uf->edges[i];
//    F=uf;
    ep=uf->edges[j];

    if(uf->data.uvpts) {
        u=uf->data.uvpts[i];
        jn=uf->joins[i];
        jp=uf->joins[j];
    }

    if(v==en->data.VertS) {
        en->patchSL=this;
    } else {
        en->patchER=this;
    }
    if(v==ep->data.VertS) {
        ep->patchSR=this;
    } else {
        ep->patchEL=this;
    }
}
    
