// patch.h -- declares the Patch class

#ifndef PATCH_H
#define PATCH_H

#include "pointer.h"
//#include "wsum.h"
#include "calc.h"

// the data contains mostly cross-references and interior vertices

struct Patch {
    protected:
        Vec3    v_v;    // the vertex interior control point
        Vec3    v_n;    // the trailing edge interior control point
        Vec3    v_f;    // the face interior control point
        Vec3    v_p;    // the leading edge interior control point
        Vec2    u_v;    // the vertex interior control point
        Vec2    u_n;    // the trailing edge interior control point
        Vec2    u_f;    // the face interior control point
        Vec2    u_p;    // the leading edge interior control point
        Ref<Patch>  n0; //  opposite the e0 edge
        Ref<Patch>  n1; //  opposite the e1 edge
        Ref<Patch>  n2; //  opposite the e2 edge
        Ref<Patch>  n3; //  opposite the e3 edge

    public:
        Ref<Patch>  next;   // linked lists, ahoy!
        Ref<Vert> v;  // the vertex point
        Ref<Edge> en; // the trailing edge point
        Ref<Face> f;  // the face point
        Ref<Edge> ep; // the leading edge point
        Ref<UVPt> u;  // the uv-point
        Ref<Join> jn; // the trailing join
        Ref<Join> jp; // the leading join

        Patch(class Face*,int); // construct from a Face and an index into that Face
        void draw(int,bool); // draw this patch in OpenGL
        void calcSubf();    // calculate final points

    friend class Model;
    friend class Edge;
};

#endif

