// pointer.h -- creates two classes for encapsulating pointers
// this does no error-checking, except for reducing memory leaks
// the Ref class doesn't even do this, but is here for symmetry

#ifndef POINTER_H
#define POINTER_H

// This is used to track objects that are owned by the object having the pointer.
// It's main purposed is to ensure the destruction of T when necessary.
template<class T>
class Ptr {
    T*  p;
    public:
    Ptr() { p=(T*)0; }  // construct
    ~Ptr() { if(p) delete p; }  // automatic destruction
    T&  operator*() { return *p; }  // reference
    T*  operator->() const { return p; }    // member reference
    operator bool() const { return p!=0; }  // test for zero
    T*  operator=(T* n) { if(p) delete p; return (p=n); }   // set
    operator T*() const { return p; }   // access item for comparisons & stuff
};

// This is used to track objects that are not owned by the object having the pointer.
template<class T>
class Ref {
    T*  p;
    public:
    Ref() { p=(T*)0; }  // construct without argument
    Ref(T* ip) { p=ip; } // construct from argument
//    ~Ref() { }  // automatic destruction
    T&  operator*() { return *p; }  // reference
    T*  operator->() const { return p; }    // member reference
    operator bool() const { return p!=0; }  // test for zero
    T*  operator=(T* n) { return (p=n); }   // set
    operator T*() const { return p; }   // access item for comparisons & stuff
};

#endif

