#ifndef PROJECT_H
#define PROJECT_H

#include <windows.h>
#include <stdio.h>
#include "link.h"
#include "list.h"
#include "view.h"
#include "undo.h"
#include "mystr.h"
#include "pointer.h"
#include "gadget.h"

class Project : public Link {
        public:
        static List allProjects;
        static Project* activeProject;
        FLAGS   projectFlags;  // we have flags
        Mystr   name;
        char    *SaveTitle;
        char    *SavePath;
        char    *ExportTitle;
        char    *ExportPath;
        lsmView cView;  // in use now
        lsmView sView;  // for saving during interactive changes
        short   EditMode;               // the editing mode in force in the project
        short   MouseMode;              // the mouse mode in force for the project
        Vec3     UserScale,UserRotate,UserTranslate;
        Matrix  UserMatrix;
        int   currentVersion; // current redo/undo version
        int   highVersion;    // the farthest that redo ops may go
        RECT    focusRect;
        bool    focusRectValid;
        TYPE    MarkType;           // the type of the marked items
        TYPE    HighType;           // the type of the highlighted item
        MYPTR   HighItem;
        short   cX,cY,lX,lY,oX,oY,rX,rY; // mouse click points
        Ptr<class Gadget>   gadget;

// implemented in files.cpp
        void virtual saveProject(bool)=0;

// implemented in mouse.cpp
        void    setMouseMode(int);

// implemented elsewhere
        Project();
        virtual ~Project();
        void    abortMode();
        void virtual activate() =0 ;
        virtual void    clearHighlight()=0;
        void virtual deleteHighlight(/*bool*/)=0;
//    void virtual    destroyHistory()=0;
//    void virtual detachHighlight()=0;
        virtual void    doCommands(HWND,int,HWND,UINT)=0;
        void virtual draw()=0;
        void drawFocusRect();
        void virtual drawHighlight()=0;
        void virtual drawMarking()=0;
        void virtual    editHighProperties(HWND)=0;
        void virtual    editMarkedProperties(HWND)=0;
        void virtual exportFile()=0;
        HMENU virtual   getPopupMenu();
        virtual void    leftClick(HWND,WPARAM,LPARAM)=0;
        virtual void    leftDoubleClick(HWND,WPARAM,LPARAM)=0;
        virtual void    leftRelease(HWND,WPARAM,LPARAM)=0;
        virtual void    mouseMove(HWND,WPARAM,LPARAM)=0;
        virtual void    rightClick(HWND,WPARAM,LPARAM)=0;
        virtual void    rightRelease(HWND,WPARAM,LPARAM)=0;
        virtual void setWindowTitle(HWND)=0;
        void    swapHandedness();
        void virtual enterMode(short)=0;
        void virtual    saveView();
        void virtual updateMenu();
        void virtual updateToolBar();
        void virtual zoomToFit()=0;

        static void     activateFirst();
        static void activateProject(Project*);
        static void activateNext();
        static void activatePrev();
        static void closeActiveProject(void);
        static bool destroyAllProjects(void);
        static void     redoDisplay(Project*);
        static void setWindowTitle(HWND,Project*);
        static bool isUniqueProjectName(char *);

//  undo/redo functions and support
    virtual void        destroyAllHistory()=0;
//      virtual void    takeSnap()=0;
//      virtual void    undoChanges()=0;
//      virtual void    redoChanges()=0;
};

#define PF_MODIFIED             1       // modified since last load
#define PF_SHOWAXES             2       // axes will be visible
#define PF_NOTDEFAULTNAME       4

//#define PT_MODEL      0
//#define       PT_SCENE        1

#endif

