// scene.h -- declares the Scene class

#ifndef SCENE_H
#define SCENE_H

#include <windows.h>
#include "project.h"
#include "list.h"

class Scene : public Project {
//        Chain   Basics; // the base models for objects
//        Chain   Objects;    // the objects

        public:

// implemented in scene_a.cpp
    void    activate();

// implemented in scene_c.cpp
    void    clearAll();
    static Scene*   createScene(HWND);
    static char*    createSceneName();

// implemented in scene_d.cpp
    void    draw();

// implemented in display.cpp
        void    drawHighlight();
        void    drawMarking();

// implemented in edit.cpp
        void    deleteHighlight();
        void    editHighProperties(HWND);
        void    editMarkedProperties(HWND);

// implemented in files.cpp
        void    exportFile();
        void    saveProject(bool);

// implemented in main.cpp
    void        doCommands(HWND,int,HWND,UINT);

// implemented in mouse_s.cpp
        void    leftClick(HWND,WPARAM,LPARAM);
        void    leftDoubleClick(HWND,WPARAM,LPARAM);
        void    leftRelease(HWND,WPARAM,LPARAM);
        void    mouseMove(HWND,WPARAM,LPARAM);
        void    rightClick(HWND,WPARAM,LPARAM);
        void    rightRelease(HWND,WPARAM,LPARAM);

// implemented in project.cpp
        void    enterMode(short);

// implemented in scene.cpp
        Scene();
        ~Scene();
        void    clearHighlight();
        void    setWindowTitle(HWND);
        void    zoomToFit();

// implemented in string.cpp
        char*   createName();

// implemented in undo.cpp
        void    destroyAllHistory();
};

class BaseModel : public Link {
        char*   FilePath;   // the file that was loaded to make this model
        float*  baseSettings;   // the default settings for the Bones (9 per Bone)
        float*  baseUVPts;  // the uv values for the uv-mapping points (2 per point)
        float*  baseVerts;  // the local xyz values for each vertex
        INDEX*  faceVerts;  // an array of indices to the vertices (4 per face)
        INDEX*  faceTints;  // an array of indices to the textures (1 per face)
        INDEX*  faceUVPts;  // an array of indices to the uv maps (4 per face)
//        Chain   Tints;  // the textures
};

struct Setting {
        float   time;   // the time this setting applies
        Vec3    values; // the values applied as of the specified time
};

struct Pose {
        unsigned int    count;  // the number of settings for this Pose
        Setting*   values; // an array of Setting objects
};

class PosedModel : public Link {
        BaseModel*  usedModel;  // the base model used here
        Pose*           poses;          // the settings for the Bones (1 per bone)
        float*      vertCache;  // the vertex data for this pose (3 per vertex)
};

#endif
