// script.h -- declares the various classes for the scripting objects

// scripts objects will require the following functions:

// constructor -- gets called when the object is created.
// destructor -- gets called when the object is destroyed.
// on modification -- gets called when the object is modified.
// Display (OpenGL drawing functions for Render mode) -- gets called for screen redrawing.
// Select (OpenGL drawing functions for Select mode) -- gets called for mouse-over detection.

#ifndef SCRIPT_H
#define SCRIPT_H

#include "link.h" // get my linked list stuff
#include "list.h"

typedef short ScriptWord;

enum ObjDataType {
    SDT_FLOAT,
    SDT_VEC2,
    SDT_VEC3,
    SDT_RGB,
    SDT_RGBA,
    SDT_STRING,
    SDT_REF,
    SDT_

class Script : public Link {
    int             scriptID;
    ScriptWords*    instructions;
};

class ModelObject : public Link {
    ModelObjectType*   type;
    ObjectData    current;
    Chain   history;
};

class ObjectType : public Link {
    int object_type_id;
    int field_count;    // number of defined fields
    bool*   fields_are_arrays;  // whether individual fields are arrays
    ObjDataType*    field_types;
    int*    ref_types;  // the type of user-defined object referenced by reference types
    Script* scripts;    // the scripts that this object uses
};

class ObjectData {
    unsigned short*    sizes; // the array sizes for this particular object
    unsigned short*    starts; // the starting point in the respective array of the field
    float*   f_array;    // floating-point data
    int*     i_array;    // integer data
    class Mystr*   s_array;    // strings
    ScriptObject**  o_array;    // references to other script objects
};

#endif


