// subd.cpp -- implements everything having to do with the subdivision preview
#include <math.h>
#include "model.h"
#include "vert.h"
#include "edge.h"
#include "face.h"
#include "funcs.h"
#include "patch.h"

void Model::deleteSubData() {
//    MYPTR m;

    Patch*  p;
    while(patches) {
        p=patches;
        patches=p->next;
        delete p;
    }
}

void Model::buildSubData() {
    MYPTR m;
    Patch*  p;

    // bail out if not flagged for this
    if((modelFlags&MF_BUILD_SUBD)==0) {
//        UpdateStatusBar("Early return from Model::buildSubData()",0,0);
        return;
    }

// get rid of the old patches
    while(patches) {
        p=patches;
        patches=p->next;
        delete p;
    }

// set up geometry vertices for subdivision
    STEPVERTS(this,m.v) m.v->prepSub();

// set up texture vertices for subdivision
    STEPUVPTS(this,m.u) m.u->prepSub();

// set up the edges, and adjust the sharp count of each vertex
    STEPEDGES(this,m.e) {
//        m.e->prepSub();

// increment the border count if this edge is a border
        if(m.e->sharpType()==-2) {
            m.e->data.VertS->border()++;
            m.e->data.VertE->border()++;
        }
// but the following happen depending on the edge sharpness
        if(m.e->sharpType()) { // this is a sharp edge
            m.e->data.VertS->sharp1()++;
            m.e->data.VertE->sharp1()++;
        }
    }

// and now we force patch corners to be treated like regular corners
    STEPVERTS(this,m.v) {
        if(m.v->border()==2 && m.v->Edges.pop()==2) {
            m.v->sharp1()=
//            m.v->sharp2()=
            3;
        }
    }

    STEPJOINS(this,m.j) {
// create the subdivision structure
//        m.j->prepSub();

// increment the border count if this edge is a border
        if(m.j->sharpType()==-2) {
            m.j->UVPtS->border()++;
            m.j->UVPtE->border()++;
        }
// but the following happen depending on the edge sharpness
        if(m.j->sharpType()) { // this is a sharp edge
            m.j->UVPtS->sharp1()++;
            m.j->UVPtE->sharp1()++;
        }

    }

// force texture mapping patch corners to stay put during subdivision
    STEPUVPTS(this,m.u) {
        if(m.u->border()==2 && m.u->joins.pop()==2) m.u->sharp1()=3;
    }

// all typing info is now in place, weighting can be done now

    STEPFACES(this,m.f) {
        Patch*  fp=NULL;
        for(int i=0;i<m.f->data.sides;i++) {
            int j=i-1; if(i==0) j=m.f->data.sides-1;
// create the patch, which handles cross-referencing and deeper subdivision
            p=new Patch(m.f,i);
            if(i==0) {
                fp=p;
            } else {
                patches->n1=p;
                p->n2=patches;
            }
            p->next=patches;
            patches=p;

        }
        fp->n2=p;
        p->n1=fp;
    }

    STEPEDGES(this,m.e) {
        if(m.e->sharpType()==0) {
            if(m.e->patchSL)
            if(m.e->patchSR) {
                m.e->patchSL->n0=m.e->patchSR;
                m.e->patchSR->n3=m.e->patchSL;
            }

            if(m.e->patchEL)
            if(m.e->patchER) {
                m.e->patchEL->n3=m.e->patchER;
                m.e->patchER->n0=m.e->patchEL;
            }
        }
    }

    modelFlags&=~MF_BUILD_SUBD; // no longer need to rebuild
    modelFlags|=MF_CALC_SUBD; // but now we need to recalculate
}

void Model::calcSubData() {
    float factor;
    
    if((modelFlags&MF_CALC_SUBD)==0) {
//        UpdateStatusBar("Early return from Model::calcSubData()",0,0);
        return;
    }
    
// recalc level one first

    Vert*   v;
    STEPVERTS(this,v) {
        v->setupCalc();
        v->getSubf().set(0);
    }

    UVPt*   u;
    STEPUVPTS(this,u) {
        u->setupCalc();
        u->getSubf().set(0);
    }

//    UpdateStatusBar("Step 1",0,0);

    Face*   f;
    STEPFACES(this,f) {
        f->calcSub1();
        for(int i=0;i<f->data.sides;i++) {
            if(f->data.verts[i]->sharp1()<2) {
                f->data.verts[i]->weightSub1(f->getSub1());
            }
            if(f->data.uvpts) {
                if(f->data.uvpts[i]->sharp1()<2) {
                    f->data.uvpts[i]->weightSub1(f->getUSub1());
                }
            }
        }
    }

//    UpdateStatusBar("Step 2",0,0);

    Edge*   e;
    STEPEDGES(this,e) {
        e->calcSub1();

        if(e->data.VertS->sharp1()<2 || (e->data.VertS->sharp1()==2 && e->sharpType()!=0) ) {
            e->data.VertS->weightSub1(e->data.VertE->data.Spot);
        }

        if(e->data.VertE->sharp1()<2 || (e->data.VertE->sharp1()==2 && e->sharpType()!=0) ) {
            e->data.VertE->weightSub1(e->data.VertS->data.Spot);
        }
    }

//    UpdateStatusBar("Step 3",0,0);

    STEPVERTS(this,v) {
// third, each vertex gets a weighting depending on its sharp count
        factor=6.0f;
        if(v->sharp1()<2) factor=(v->Edges.pop()-2)*v->Edges.pop();
        if(v->sharp1()>2) factor=1.0;
// V1 by V0
        v->weightSub1(v->data.Spot,factor);
        v->calcSub1();
    }

    Join*   j;
    STEPJOINS(this,j) {
        j->calcSub1();
        if(j->UVPtS->sharp1()<2 || (j->UVPtS->sharp1()==2 && j->sharpType()!=0) ) {
            j->UVPtS->weightSub1(j->UVPtE->data.Spot);
        }

        if(j->UVPtE->sharp1()<2 || (j->UVPtE->sharp1()==2 && j->sharpType()!=0) ) {
            j->UVPtE->weightSub1(j->UVPtS->data.Spot);
        }
    }

    STEPUVPTS(this,u) {
        factor=6.0f;
        if(u->sharp1()<2) factor=(u->joins.pop()-2)*u->joins.pop();
        if(u->sharp1()>2) factor=1.0;
        u->weightSub1(u->data.Spot,factor);
        u->calcSub1();
    }

// now recalc level two
// the level two faces first

    Patch*  p=(Patch*)0;
    p=patches;
    while(p) {
        p->calcSubf();
        p=p->next;
    }

//    UpdateStatusBar("Step 5",0,0);

// now the final level points for the edges, faces, and vertices
    STEPEDGES(this,e) e->calcSubf();
    STEPJOINS(this,j) j->calcSubf();
    STEPFACES(this,f) f->getSubf()*=1.0f/((float)f->data.sides);
    STEPVERTS(this,v) v->calcSubf();
    STEPUVPTS(this,u) u->calcSubf();

//    UpdateStatusBar("Step 6",0,0);

//    STEPJOINS(this,j) j->calcSubF();

    modelFlags&=~MF_CALC_SUBD;
}

