#include "userint.h"
#include <gl/gl.h>
//#include <gl/glext.h>
#include "tint.h"
#include "face.h"
#include "funcs.h"

// a lot of this needs to be changed in order to create a Linux or Mac port,
// because the BMP loading uses the Official Windows way of doing it, which
// is horribly highly OS-specific.

TintData::TintData() : color(0xffffff) { }

bool TintData::operator!=(TintData& ote) {
        if(color!=ote.color) return true;
        if(Name!=ote.Name) return true;
        if(File!=ote.File) return true;
        return false;
}

void TintData::operator=(TintData& ote) {
        color=ote.color;
        Name=ote.Name;
        File=ote.File;
}

TintData::~TintData() { }

// for troubleshooting--commented out for now
//bool Model::isValidTint(Tint* test) {
//      Tint* t;
//
//      STEPTINTS(this,t)
//      if(t==test) return true;
//      return false;
//}

Tint::Tint(bool isDefTint) : hbm(NULL), sX(0), sY(0), bmBits(NULL), glName(0) {
// ensure that there is an initial history entry
    TintEntry* te=new TintEntry;
    history.addTail(te);
    if(isDefTint) {
        index=-1;
        data.Name.copy("default");
        takeSnap(0);
    }
}

void Tint::deleteInfo() {
    bmBits=NULL;
    if(hbm) { DeleteObject(hbm); hbm=(HBITMAP)0; }
    if(glName) glDeleteTextures(1,&glName); glName=(GLuint)0;
}

Tint::~Tint() {
    destroyHistory();
}

void Tint::buildTintBitmap(HDC dc) {
    BITMAPINFO bmi;
    HBITMAP new_hbm;
    HBRUSH hbr;
    BITMAP bm;
    HDC pdc,adc;
    HPEN hpen;
    RECT dr;
    int nX,nY;
    GLubyte *nbuff;
    char r,g,b;

        if(this==NULL) return;

        if(hbm) DeleteObject(hbm);

// create a device context
        pdc=CreateCompatibleDC(dc);
        memset(&bmi,0,sizeof(BITMAPINFO));
        bmi.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
        bmi.bmiHeader.biPlanes=1;
        bmi.bmiHeader.biBitCount=24;
        bmi.bmiHeader.biCompression=BI_RGB;

        new_hbm=(HBITMAP)0;

        if(data.File.len()) { // load the image
            new_hbm=(HBITMAP)LoadImage(NULL,data.File.c_string(),IMAGE_BITMAP,0,0,LR_CREATEDIBSECTION|LR_LOADFROMFILE);
            if(new_hbm==NULL) UserInt::informUser("Failed to open texture file.");
        }

        if(new_hbm==(HBITMAP)0) {
// create a DIB section
            bmi.bmiHeader.biWidth=128;
            bmi.bmiHeader.biHeight=128;

            new_hbm=CreateDIBSection(pdc,&bmi,DIB_RGB_COLORS,(void **)&(bmBits),NULL,0);
            SelectObject(pdc,new_hbm);

            hbr=CreateSolidBrush(data.color);
            SelectObject(pdc,hbr);

            dr.left=dr.top=0;
            dr.right=dr.bottom=128;

            FillRect(pdc,&dr,hbr);

            SetROP2(pdc,R2_COPYPEN);

            r=GetRValue(data.color); r=(r&128)?0:255;
            g=GetGValue(data.color); g=(g&128)?0:255;
            b=GetBValue(data.color); b=(b&128)?0:255;
            hpen=CreatePen(PS_SOLID,1,RGB(r,g,b));
            SelectObject(pdc,hpen);

//  U
            MoveToEx(pdc,56,1,NULL);
            LineTo(pdc,56,33);
            LineTo(pdc,72,33);
            LineTo(pdc,72,0);
//      D
            MoveToEx(pdc,56,126,NULL);
            LineTo(pdc,56,94);
            LineTo(pdc,64,94);
            LineTo(pdc,72,102);
            LineTo(pdc,72,118);
            LineTo(pdc,64,126);
            LineTo(pdc,56,126);
//  L
            MoveToEx(pdc,1,48,NULL);
            LineTo(pdc,1,80);
            LineTo(pdc,18,80);
//  R
            MoveToEx(pdc,110,80,NULL);
            LineTo(pdc,110,48);
            LineTo(pdc,118,48);
            LineTo(pdc,126,56);
            LineTo(pdc,118,64);
            LineTo(pdc,126,72);
            LineTo(pdc,126,81);
            MoveToEx(pdc,111,64,NULL);
            LineTo(pdc,118,64);

            DeleteObject(hbr);
            DeleteObject(hpen);
        }
        hbm=new_hbm;

        GetObject(hbm,sizeof(BITMAP),&bm);

        sX=(short)bm.bmWidth;
        sY=(short)bm.bmHeight;

        nX=nearestPowerof2(sX);
        nY=nearestPowerof2(sY);

        if(nX!=sX || nY!=sY) {
                bmi.bmiHeader.biWidth=nX;
                bmi.bmiHeader.biHeight=nY;

                new_hbm=CreateDIBSection(dc,&bmi,DIB_RGB_COLORS,(void **)&nbuff,NULL,0);

                adc=CreateCompatibleDC(dc);
                if(new_hbm==(HBITMAP)0) UserInt::informUser("Bitmap creation failed");
                if(nbuff==0) UserInt::informUser("No bits reserved for bitmap");

                SelectObject(pdc,hbm);
                SelectObject(adc,new_hbm);

                SetStretchBltMode(adc,COLORONCOLOR);
                StretchBlt(adc,0,0,nX,nY,pdc,0,0,sX,sY,SRCCOPY);

                bmBits=nbuff;
                DeleteObject(hbm);
                hbm=new_hbm;
                DeleteDC(adc);
                sX=nX;
                sY=nY;
        }
        DeleteDC(pdc);
        deriveSolid();
}

void Tint::tintWindow(HWND hwnd) {
        RECT trect;
        HDC tdc,sdc;

//      if(this==NULL) this=&((Model*)ActiveProject)->defTint;

        GetClientRect(hwnd,&trect);
        tdc=GetDC(hwnd);
        sdc=CreateCompatibleDC(tdc);

        if(hbm==NULL) buildTintBitmap(tdc);

        SelectObject(sdc,hbm);

        SetStretchBltMode(tdc,COLORONCOLOR);
        StretchBlt(tdc,0,0,128,128,sdc,0,0,sX,sY,SRCCOPY);

        trect.left-=2;
        trect.top-=2;
        trect.right=130;
        trect.bottom=130;

        DrawEdge(tdc,&trect,EDGE_SUNKEN,BF_RECT);

        DeleteDC(sdc);
        ReleaseDC(hwnd,tdc);
}

void Tint::getTintBitmap() {
        BITMAP bm;
        HDC q_dc;

        q_dc=GetDC(NULL);

        if(hbm==NULL) {
            q_dc=GetDC(NULL);
            buildTintBitmap(q_dc);
            ReleaseDC(NULL,q_dc);
        }

        GetObject(hbm,sizeof(BITMAP),&bm);

        bmBits=(GLubyte *)bm.bmBits;
        sX=(short)bm.bmWidth;
        sY=(short)bm.bmHeight;
}

void Tint::deriveSolid() {
        int i,j;
        float r,g,b;
        GLubyte *p;

        if(data.File.len()==0) return;

        if(bmBits==(GLubyte*)0) getTintBitmap();

        p=bmBits;

        r=g=b=0.0;

        for (i=0;i<sX;i++)
        for (j=0;j<sY;j++) {
                b+=*(p++);
                g+=*(p++);
                r+=*(p++);
        }

        data.color=RGB(r/(float)(sX*sY),g/(float)(sX*sY),b/(float)(sX*sY));
}

void Tint::selectIntoOpenGL() {
//    if(bmBits==0) getTintBitmap();
    if(glName==0) {
        glGenTextures(1,&glName);
        glBindTexture(GL_TEXTURE_2D,glName);
        glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
        glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
        glTexImage2D(GL_TEXTURE_2D,0,GL_RGB,sX,sY,0,GL_BGR_EXT,GL_UNSIGNED_BYTE,bmBits);
    } else {
        glBindTexture(GL_TEXTURE_2D,glName);
    }
}

// undo/redo functions

bool Tint::takeSnap(INDEX newVersion) {
        TintEntry*      th;
        TintEntry*      lth;
        bool make;
        TYPE    newType;

        // clip all history equal to or greater than new version
        clipHistory(newVersion);

        // most recent entry should be last
        lth=(TintEntry*)(history.last());

/* behavior logic:
for takeSnap():

last entry      current object
NONE            invalid                 flag as TF_KILLME
et_nul          invalid                 flag as TF_KILLME
et_chg          invalid                 archive as et_del
et_del          invalid                 do nothing

NONE            valid                   archive as et_chg
et_nul          valid                   archive as et_chg
et_chg          valid                   compare and archive as et_chg
et_del          valid                   shouldn't happen
*/

        make=false; // default is don't make new entry
        newType=ET_CHG; // most common type is change

        if(lth==NULL) {
                if(flags&TF_INVALID) {
                        flags|=TF_KILLME;
                } else {
                        make=true;
                }
        } else {
                switch(lth->type) {
                        case ET_NUL:
                                if(flags&TF_INVALID) {
                                        flags|=TF_KILLME;
                                } else {
                                        make=true;
                                        newType=ET_CHG;
                                }
                                break;
                        case ET_CHG:
                                if(flags&TF_INVALID) {
                                        make=true;
                                        newType=ET_DEL;
                                } else {
                                        // make it if latest is different from current
                                        if (lth->data!=data) make=true;
                                }
                                break;
// last case is either do nothing or shouldn't happen
//                      case ET_DEL:
//                              break;
                }
        }

        if(make) {
                th=new TintEntry;
                history.addTail(th);
                th->version=newVersion;
                th->type=newType;
                th->data=data;
        }

        return make;
}

void    Tint::clipHistory(INDEX version) {
        TintEntry *te;
        TintEntry *nte;

        te=(TintEntry*)(history.first());
        while(te!=NULL) {
                nte=(TintEntry*)(te->next());
                if(te->version>=version) {
                        te->remove();
                        delete te;
                }
                te=nte;
        }
}

void Tint::destroyHistory() {
        TintEntry*  te;

        while(history.last()) {
                te=(TintEntry*)(history.last());
                te->remove();
                delete te;
        }
}

void Tint::restoreVersion(INDEX getVersion) {
        TintEntry*  te;

        te=(TintEntry*)(history.last());

// there is always an entry with version 0, so this should never back up
// such that ve==0
        while(te->version > getVersion)
        te=(TintEntry*)(te->prev());

        if(te->type==ET_CHG) {
                data=te->data;
                deleteInfo();   // force reloading of bitmap data
                flags&=~TF_INVALID;
        } else {
                flags|=TF_INVALID;
        }

}
