// transform.cpp -- implements the Transform structure

#include <math.h>
#include "transform.h"
#include "calc.h"

Matrix::Matrix() {
    identity();
}

void Matrix::identity() {
    x.x=y.y=z.z=1.0;

    x.y=x.z=
    y.x=y.z=
    z.x=z.y=
    l.x=l.y=l.z=0.0;
}

//float Matrix::det() {
//    return x.x * (y.y*z.z -y.z*z.y)
//         + x.y * (y.z*z.x -y.x*z.z)
//         + x.z * (y.x*z.y -y.y*z.x);
//}

void Matrix::invert(Matrix& i) {
//        Matrix t1,t2;

        i.x.x=x.x;
        i.x.y=y.x;
        i.x.z=z.x;

        i.y.x=x.y;
        i.y.y=y.y;
        i.y.z=z.y;

        i.z.x=x.z;
        i.z.y=y.z;
        i.z.z=z.z;

        i.l.x=-l.dot(x);
        i.l.y=-l.dot(y);
        i.l.z=-l.dot(z);
}

void Matrix::mult(Matrix& s,Matrix& d) {
        d.x.x = x.x*s.x.x + x.y*s.y.x + x.z*s.z.x;
        d.x.y = x.x*s.x.y + x.y*s.y.y + x.z*s.z.y;
        d.x.z = x.x*s.x.z + x.y*s.y.z + x.z*s.z.z;

        d.y.x = y.x*s.x.x + y.y*s.y.x + y.z*s.z.x;
        d.y.y = y.x*s.x.y + y.y*s.y.y + y.z*s.z.y;
        d.y.z = y.x*s.x.z + y.y*s.y.z + y.z*s.z.z;

        d.z.x = z.x*s.x.x + z.y*s.y.x + z.z*s.z.x;
        d.z.y = z.x*s.x.y + z.y*s.y.y + z.z*s.z.y;
        d.z.z = z.x*s.x.z + z.y*s.y.z + z.z*s.z.z;

        d.l.x = l.x*s.x.x + l.y*s.y.x + l.z*s.z.x + s.l.x;
        d.l.y = l.x*s.x.y + l.y*s.y.y + l.z*s.z.y + s.l.y;
        d.l.z = l.x*s.x.z + l.y*s.y.z + l.z*s.z.z + s.l.z;
}

void Matrix::mult(Matrix& s) {
        Matrix temp;

        mult(s,temp);

        *this=temp;
}

Matrix& Matrix::operator*=(Matrix& om) {
        Matrix temp;

        mult(om,temp);

        *this=temp;
        return *this;
}

// the divisor can have translation and rotation, but no scaling or shearing
// however the code that is commented out will support scaling as well
// proper matrix inversion is far more complicated
Matrix& Matrix::operator/=(Matrix& om) {
    Matrix temp;

//    Vec3    div;

//    div.x=1.0f/length(om.x);
//    div.y=1.0f/length(om.y);
//    div.z=1.0f/length(om.z);

    l-=om.l;
    temp.x.x = x.dot(om.x);//*div.x;
    temp.x.y = x.dot(om.y);//*div.y;
    temp.x.z = x.dot(om.z);//*div.z;

    temp.y.x = y.dot(om.x);//*div.x;
    temp.y.y = y.dot(om.y);//*div.y;
    temp.y.z = y.dot(om.z);//*div.z;

    temp.z.x = z.dot(om.x);//*div.x;
    temp.z.y = z.dot(om.y);//*div.y;
    temp.z.z = z.dot(om.z);//*div.z;

    temp.l.x = l.dot(om.x);//*div.x;
    temp.l.y = l.dot(om.y);//*div.y;
    temp.l.z = l.dot(om.z);//*div.z;

    *this=temp;
    return *this;
}

//void Matrix::transform(Vec3& s,Vec3& d) {
//        d.x=s.x*x.x + s.y*y.x + s.z*z.x + l.x;
//        d.y=s.x*x.y + s.y*y.y + s.z*z.y + l.y;
//        d.z=s.x*x.z + s.y*y.z + s.z*z.z + l.z;
//}

Vec3& Vec3::operator*=(Matrix& m) {
    Vec3 temp;

    temp.x=x*m.x.x + y*m.y.x + z*m.z.x + m.l.x;
    temp.y=x*m.x.y + y*m.y.y + z*m.z.y + m.l.y;
    temp.z=x*m.x.z + y*m.y.z + z*m.z.z + m.l.z;

    *this=temp;
    return *this;
}

Vec3& Vec3::operator*=(Transform& m) {
        (*this)*=m.fore;
        return *this;
}

Vec3& Vec3::operator/=(Transform& m) {
        (*this)*=m.back;
        return *this;
}

Vec3& Vec3::operator/=(Matrix& m) {
        Vec3 temp;

        (temp=*this)-=m.l;

        x=temp.dot(m.x);
        y=temp.dot(m.y);
        z=temp.dot(m.z);

        return *this;
}

void Matrix::invertUVector(Vec3& s,Vec2& d) {
        Vec3 src;

        (src=s)-=l;

        d.u=src.x*x.x + src.y*x.y + src.z*x.z;
        d.v=src.x*y.x + src.y*y.y + src.z*y.z;
}

// divides the argument by this
void Matrix::invertMatrix(Matrix& m) {
    Vec3 temp;

    m.l.x-=l.x;
    m.l.y-=l.y;
    m.l.z-=l.z;

    temp.x=m.x.dot(x);
    temp.y=m.x.dot(y);
    temp.z=m.x.dot(z);
    m.x=temp;

    temp.x=m.y.dot(x);
    temp.y=m.y.dot(y);
    temp.z=m.y.dot(z);
    m.y=temp;

    temp.x=m.z.dot(x);
    temp.y=m.z.dot(y);
    temp.z=m.z.dot(z);
    m.z=temp;

    temp.x=m.l.dot(x);
    temp.y=m.l.dot(y);
    temp.z=m.l.dot(z);
    m.l=temp;
}

// Transform requires no constructor -- all members have them

void Transform::identity() {
        fore.identity();
        back.identity();
}

#if(0)
void Transform::transformVector(Vec3& s,Vec3& d) {
        fore.transform(s,d);
}

void Transform::invertVector(Vec3& s,Vec3& d) {
        back.transform(s,d);
}
#endif

void Transform::scaleBy(Vec3& svec) {
        Matrix temp;

        temp.x.x=svec.x;
        temp.y.y=svec.y;
        temp.z.z=svec.z;

        fore.mult(temp);

        temp.x.x=1.0/svec.x;
        temp.y.y=1.0/svec.y;
        temp.z.z=1.0/svec.z;

        temp.mult(back);
        back=temp;
}

void Transform::rotateBy(Vec3& rvec) {
        Matrix temp;
        float cosx,cosy,cosz;
        float sinx,siny,sinz;

        cosx=rvec.x*PI/180.0;
        sinx=sinf(cosx);
        cosx=cosf(cosx);
        cosy=rvec.y*PI/180.0;
        siny=sinf(cosy);
        cosy=cosf(cosy);
        cosz=rvec.z*PI/180.0;
        sinz=sinf(cosz);
        cosz=cosf(cosz);

        temp.x.x=cosy*cosz;
        temp.x.y=cosy*sinz;
        temp.x.z=-siny;
        temp.y.x=sinx*siny*cosz-cosx*sinz;
        temp.y.y=sinx*siny*sinz+cosx*cosz;
        temp.y.z=sinx*cosy;
        temp.z.x=cosx*siny*cosz+sinx*sinz;
        temp.z.y=cosx*siny*sinz-sinx*cosz;
        temp.z.z=cosx*cosy;

        fore.mult(temp);

        float swap;

        swap=temp.x.y;
        temp.x.y=temp.y.x;
        temp.y.x=swap;

        swap=temp.x.z;
        temp.x.z=temp.z.x;
        temp.z.x=swap;

        swap=temp.y.z;
        temp.y.z=temp.z.y;
        temp.z.y=swap;

        temp.mult(back);
        back=temp;
}

void Transform::translateBy(Vec3& tvec) {
        Matrix temp;

        temp.l=tvec;

        fore.mult(temp);

        temp.l*=-1.0f;

        temp.mult(back);
        back=temp;
}

void Transform::clearTranslate() {
        fore.l.x=
        fore.l.y=
        fore.l.z=
        back.l.x=
        back.l.y=
        back.l.z=0;
}

void Transform::transformBy(Transform& b) {
        Matrix temp;
        fore.mult(b.fore);
        b.back.mult(back,temp);
        back=temp;
}

// the builds the inverse transform from the obverse
// this assumes that the matrix has no scaling or skewing element to it
// rotate and transform only
void Transform::buildInverse() {
    back.setXVec(fore.x.x,fore.y.x,fore.z.x);
    back.setYVec(fore.x.y,fore.y.y,fore.z.y);
    back.setZVec(fore.x.z,fore.y.z,fore.z.z);
    back.setLVec( -fore.x.dot(fore.l),-fore.y.dot(fore.l),-fore.z.dot(fore.l) );
}

