#include <windows.h>
#include <math.h>
#include "vert.h"
#include "patch.h"
#include "edge.h"
#include "face.h"
#include "bone.h"
#include "funcs.h"
#include "calc.h"
#include "lsmtypes.h"

bool VertData::operator!=(VertData& od) {
        if(Spot!=od.Spot) return true;
        if(Owner!=od.Owner) return true;
//      if (flags!=od.flags) return true;
        return false;
}

VertData& VertData::operator=(VertData& od) {
        Spot=od.Spot;
        Owner=od.Owner;
//      flags=od.flags;
        return *this;
}

VertData::VertData() : Owner((Bone*)0) { }

Vert::Vert()
    : Copy((Vert*)0)
    , Order(0)
{
    VertEntry* ve=new VertEntry;
    history.addTail(ve);
}

Vert::~Vert() {
    destroyHistory();
}

void Vert::addEdge(Edge* e) {
//    if(e->data.VertS!=this && e->data.VertE!=this)
//    ReportSystemError("Attempted edge add to a non-bordering vert");
    Edges.append(e);
}

void Vert::calcSubf() {
    switch(sharp1()) {
        case 0:
        case 1:
            subLevelf*=1.0f/((float)Edges.pop());
            break;
        case 2:
            if(Edges.pop()==2) {
                subLevelf=getSub1();
            } else {
                (subLevelf=getSub1())*=4.0;
                for(int i=0;i<Edges.pop();i++) {
                    if(Edges[i]->sharpType()) {
                        subLevelf+=Edges[i]->getSub1();
                    }
                }
                subLevelf*=.1666666667;
            }
            break;
        default:
            subLevelf=getSub1();
    }
}

void Vert::detach() { data.Owner=NULL; }

void Vert::draw(bool subd) {
//    if(flags&TF_HIDDEN) return;
    
    if(subd) {
        subLevelf.glPlot();
    } else {
        data.Spot.glPlot();
    }
}

Edge* Vert::findEdge(Vert* ov) {
    for(int i=0;i<Edges.pop();i++)
    if(Edges[i]->data.VertS==ov || Edges[i]->data.VertE==ov) return Edges[i];

    return (Edge*)0;
}

void Vert::prepSub() {
    sharpCount1=
//    sharpCount2=
    borderCount=0;
//    subLevel1.setup(Edges.pop()*2+1);
//    subLevel2.setup(Edges.pop()*2+1);
}

void Vert::remEdge(Edge* e) {
    Edges.remove(e);
}

void Vert::snapToGrid(float grid) {
    Vec3 tmp;


    tmp=data.Spot;
    if(data.Owner) {
        tmp/=data.Owner->Scaled;
    }

    tmp.x=floorf(tmp.x/grid+.5)*grid;
    tmp.y=floorf(tmp.y/grid+.5)*grid;
    tmp.z=floorf(tmp.z/grid+.5)*grid;

    data.Spot=tmp;
    if(data.Owner) {
        data.Spot*=data.Owner->Scaled;
    }
}

void    Vert::saveLoc() {
        Save=data.Spot;
}

// undo/redo functions

bool Vert::takeSnap(INDEX newVersion) {
        VertEntry*      vh;
        VertEntry*      lvh;
        bool make;
        TYPE    newType;

        // clip all history equal to or greater than new version
        clipHistory(newVersion);

        // most recent entry should be last
        lvh=(VertEntry*)(history.last());

/* behavior logic:
for takeSnap():

last entry      current object
NONE            invalid                 flag as TF_KILLME
et_nul          invalid                 flag as TF_KILLME
et_chg          invalid                 archive as et_del
et_del          invalid                 do nothing

NONE            valid                   archive as et_chg
et_nul          valid                   archive as et_chg
et_chg          valid                   compare and archive as et_chg
et_del          valid                   shouldn't happen
*/

        make=false; // default is don't make new entry
        newType=ET_CHG; // most common type is change

        if(lvh==NULL) {
                if(flags&TF_INVALID) {
                        flags|=TF_KILLME;
                } else {
                        make=true;
                }
        } else {
                switch(lvh->type) {
                        case ET_NUL:
                                if(flags&TF_INVALID) {
                                        flags|=TF_KILLME;
                                } else {
                                        make=true;
                                        newType=ET_CHG;
                                }
                                break;
                        case ET_CHG:
                                if(flags&TF_INVALID) {
                                        make=true;
                                        newType=ET_DEL;
                                } else {
                                        // make it if latest is different from current
                                        if (lvh->data!=data) make=true;
                                }
                                break;
// last case is either do nothing or shouldn't happen
//                      case ET_DEL:
//                              break;
                }
        }

        if(make) {
                vh=new VertEntry;
                history.addTail(vh);
                vh->version=newVersion;
                vh->type=newType;
                vh->data=data;
        }

        return make;
}

void Vert::clipHistory(INDEX version) {
        VertEntry *ve;
        VertEntry *nve;

        ve=(VertEntry*)(history.first());
        while(ve!=NULL) {
                nve=(VertEntry*)(ve->next());
                if(ve->version>=version) {
                        ve->remove();
                        delete ve;
                }
                ve=nve;
        }
}

void Vert::destroyHistory() {
        VertEntry*  ve;

        while(history.last()) {
                ve=(VertEntry*)(history.last());
                ve->remove();
                delete ve;
        }
}

void Vert::restoreVersion(INDEX getVersion) {
        VertEntry*  ve;

        ve=(VertEntry*)(history.last());

// there is always an entry with version 0, so this should never back up
// such that ve==0
        while(ve->version > getVersion)
        ve=(VertEntry*)(ve->prev());

        if(ve->type==ET_CHG) {
                data=ve->data;
                flags&=~TF_INVALID;
        } else {
                flags|=TF_INVALID;
        }
}

//Delta* Snap::createVertDelta(Vert *v,TYPE dtype) {
//      Delta* d;
//      LSMWORD ext[2];

//      d=createDelta(dtype,VERTDELT,0);
//      if(d==NULL) {
//      //              ReportSystemError("Creation of Vertex Delta object failed.");
//              return d;
//      }
//      d->index=v->index;

//      if(dtype==DT_DEL) return d;

//      d->extend(3,(LSMWORD *)&v->data.Spot);

//      if(v->data.Owner) {
//              ext[0].l=v->data.Owner->index;
//      } else {
//              ext[0].l=(INDEX)-1;
//      }
//      d->extend(1,ext);

//      if(v->data.flags&VF_WITHB) {
//              ext[0].l=ATBASEFLAG;
//              d->extend(1,ext);
//      }
//      if(v->data.flags&VF_LOCKED) {
//              ext[0].l=LOCKVFLAG;
//              d->extend(1,ext);
//      }

//      if(v->data.layer->index!=(INDEX)-1) {
//              ext[0].l=LAYERBLOCK;
//              ext[1].l=v->data.layer->index;
//              d->extend(2,ext);
//      }

//      return d;
//}
