#include <windows.h>
#include <stdio.h>
#include <gl/gl.h>
#include <math.h>
#include "view.h"
#include "funcs.h"
#include "pen.h"
#include "calc.h"
#include "bound.h"
#include "appdata.h"

float ScaleTable[27]={
        .001000,.002000,.005000,
        .010000,.020000,.050000,
        .100000,.200000,.500000,
        1.00000,2.00000,5.00000,
        10.0000,20.0000,50.0000,
        100.000,200.000,500.000,
        1000.00,2000.00,5000.00,
        10000.0,20000.0,50000.0,
        100000.,200000.,500000. };

lsmView::lsmView()
        : Scale(1.0)
        , Flags(VF_LEFTHANDED)
{ }

void lsmView::loadIntoOpenGL(int cX,int cY) {
        GLfloat m[16];

        m[ 1]=m[ 2]=m[ 3]=m[ 4]=m[ 6]=m[ 7]=m[ 8]=m[ 9]=0.0f;
        m[10]=0.001f/Scale;
        m[11]=m[12]=m[13]=0.0f;
        m[14]=0.5f;
        m[15]=1.0f;

        m[0]=128.0f/Scale/(float)cX;
        m[5]=128.0f/Scale/(float)cY;
        glMultMatrixf(m);

        m[ 0]=Orient.x.x;
        m[ 1]=Orient.y.x;
        m[ 2]=Orient.z.x;
        m[ 3]=0.0;
        m[ 4]=Orient.x.y;
        m[ 5]=Orient.y.y;
        m[ 6]=Orient.z.y;
        m[ 7]=0.0;
        m[ 8]=Orient.x.z;
        m[ 9]=Orient.y.z;
        m[10]=Orient.z.z;
        m[11]=
        m[12]=
        m[13]=
        m[14]=0.0f;
        m[15]=1.0f;
        glMultMatrixf(m);

        m[ 0]=1.0f;
        m[ 1]=m[ 2]=m[ 3]=m[ 4]=0.0f;
        m[ 5]=1.0f;
        m[ 6]=m[ 7]=m[ 8]=m[ 9]=0.0f;
        m[10]=1.0f;
        m[11]=0.0f;
        m[12]=-Orient.l.x;
        m[13]=-Orient.l.y;
        m[14]=-Orient.l.z;
        m[15]=1.0f;
        glMultMatrixf(m);
}

void lsmView::mapScr2XYZ(short pX,short pY,Vec3& d) {
    d.x=(float)((pX<<1)-sX)/128.0f*Scale;
    d.y=(float)(sY-((pY-Screen.top)<<1))/128.0f*Scale;
    d.z=0.0;
    d*=Orient;
}

void lsmView::mapScr2UV(short pX,short pY,Vec2& d) {
//    float fX=(float)((pX<<1)-sX)/(float)sM*Scale;
//    float fY=(float)(sY-(pY<<1))/(float)sM*Scale;
    float fX=(float)((pX<<1)-sX)/128.0f*Scale;
    float fY=(float)(sY-(pY<<1))/128.0f*Scale;
    d.u=Orient.l.x + Orient.x.x*fX + Orient.y.x*fY;
    d.v=Orient.l.y + Orient.x.y*fX + Orient.y.y*fY;
}

void lsmView::mapXYZ2UV(Vec3& s,Vec2& d) {
    Orient.invertUVector(s,d);

    d.u/=(Scale*2.0f);d.u+=.5;
    d.v/=(Scale*2.0f);d.v+=.5;
}

void lsmView::drawAxes(bool xyz) {
    float sC,pX,pY,pZ;
    float sW;
    float mX,mY,mZ;
    float nX,nY,nZ;
    float aX,aY,aZ;
    float fine;
    float coarse;
    int ratio;
    float fV;
    int ii,j;
    GLfloat m[16];

    glDisable(GL_LINE_STIPPLE);

    sC=Scale;

    if(sX>sY)
        sW=(Scale*sY)/160.0f;
    else
        sW=(Scale*sX)/160.0f;

    pX=Orient.l.x;
    pY=Orient.l.y;
    pZ=Orient.l.z;

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glTranslatef(pX,pY,pZ);
    glScalef(sW,sW,sW);

    if(xyz) {
        glLineWidth(1.5f);
        SelectGLColor(AppData::PenColor[XAXISPEN]);
        glCallList(AppData::GLObjBase+AD_OBJ_ARROW);

        SelectGLColor(AppData::PenColor[YAXISPEN]);
        glCallList(AppData::GLObjBase+AD_OBJ_FLIPXY);
        glCallList(AppData::GLObjBase+AD_OBJ_ARROW);

        SelectGLColor(AppData::PenColor[ZAXISPEN]);
        glCallList(AppData::GLObjBase+AD_OBJ_FLIPXY);
        glCallList(AppData::GLObjBase+AD_OBJ_FLIPXZ);
        glCallList(AppData::GLObjBase+AD_OBJ_ARROW);

        glCallList(AppData::GLObjBase+AD_OBJ_FLIPXZ);

        m[ 0]=Orient.x.x;    m[ 1]=Orient.x.y;    m[ 2]=Orient.x.z;
        m[ 4]=Orient.y.x;    m[ 5]=Orient.y.y;    m[ 6]=Orient.y.z;
        m[ 8]=Orient.z.x;    m[ 9]=Orient.z.y;    m[10]=Orient.z.z;
        m[ 3]=    m[ 7]=    m[11]=    m[12]=    m[13]=    m[14]=0.0f;
        m[15]=1.0f;

        for(int i=0;i<6;i++) {
            switch(i) {
                case 0: // positive X axis
                case 3: // negative X axis
                    SelectGLColor(AppData::PenColor[XAXISPEN]);
                    break;
                case 1: // positive Y axis
                case 4: // negative Y axis
                    SelectGLColor(AppData::PenColor[YAXISPEN]);
                    break;
                case 2: // positive Z axis
                case 5: // negative Z axis
                    SelectGLColor(AppData::PenColor[ZAXISPEN]);
                    break;
            }
            glPushMatrix();
            switch(i) {
                case 0: glTranslatef( 1.125f,0,0); break;
                case 1: glTranslatef(0, 1.125f,0); break;
                case 2: glTranslatef(0,0, 1.125f); break;
                case 3: glTranslatef(-1.125f,0,0); break;
                case 4: glTranslatef(0,-1.125f,0); break;
                case 5: glTranslatef(0,0,-1.125f); break;
            }
            
            glMultMatrixf(m);

            glBegin(GL_LINES);
            if(i<3) {
                glVertex2f(-0.075f,-0.025f);
                glVertex2f(-0.075f, 0.025f);
            }
            switch(i) {
                case 0:
                case 3:
                    glCallList(AppData::GLObjBase+AD_OBJ_CHARX);
                    break;
                case 1:
                case 4:
                    glCallList(AppData::GLObjBase+AD_OBJ_CHARY);
                    break;
                case 2:
                case 5:
                    glCallList(AppData::GLObjBase+AD_OBJ_CHARZ);
                    break;
            }
            glEnd();

            glPopMatrix();
        }
    }

    glLineWidth(1.0f);

    for(ii=1;ii<27 && sC>ScaleTable[ii];ii++);

    switch(ii) {
        case  2: case  5: case  8:
        case 11: case 14: case 17:
        case 20: case 23: case 26:
        case 29: case 32: case 35:
            fine=ScaleTable[ii-2];
            break;
        default:
            fine=ScaleTable[ii-1];
            break;
    }
    coarse=ScaleTable[ii];
    ratio=(int)floorf(coarse/fine+.5);
    
    glLoadIdentity();

    glScalef(fine,fine,fine);    

    if(sX>sY)
        sW=(Scale*sX)/160.0f;
    else
        sW=(Scale*sY)/160.0f;

    mX=floorf((pX-sW*1.4f)/fine+.5f);
    mY=floorf((pY-sW*1.4f)/fine+.5f);
    mZ=floorf((pZ-sW*1.4f)/fine+.5f);

    nX=floorf((pX+sW*1.4f)/fine+.5f);
    nY=floorf((pY+sW*1.4f)/fine+.5f);
    nZ=floorf((pZ+sW*1.4f)/fine+.5f);

    glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
//        glEnable(GL_BLEND);

    aX=sqrt(fabs(Orient.z.x))*.75f;
    aY=sqrt(fabs(Orient.z.y))*.75f;
    aZ=sqrt(fabs(Orient.z.z))*.75f;

    glBegin(GL_LINES);
    for(fV=j=mX;fV<nX+.01;fV++,j++) {
        if((j%ratio)==0) {
            SelectGLColorAlpha(AppData::PenColor[GRIDPEN],aY);
        } else {
            SelectGLColorAlpha(AppData::PenColor[GRIDPEN],aY*.5f);
        }
        if(Orient.z.y<0) {
            glVertex3f(fV,mY,mZ);
            glVertex3f(fV,mY,nZ);
        } else {
            glVertex3f(fV,nY,mZ);
            glVertex3f(fV,nY,nZ);
        }
        if((j%ratio)==0) {
            SelectGLColorAlpha(AppData::PenColor[GRIDPEN],aZ);
        } else {
            SelectGLColorAlpha(AppData::PenColor[GRIDPEN],aZ*.5f);
        }
        if(Orient.z.z<0) {
            glVertex3f(fV,mY,mZ);
            glVertex3f(fV,nY,mZ);
        } else {
            glVertex3f(fV,mY,nZ);
            glVertex3f(fV,nY,nZ);
        }
    }
    for(fV=j=mY;fV<nY+.01;fV++,j++) {
        if((j%ratio)==0) {
            SelectGLColorAlpha(AppData::PenColor[GRIDPEN],aZ);
        } else {
            SelectGLColorAlpha(AppData::PenColor[GRIDPEN],aZ*.5f);
        }
        if(Orient.z.z<0) {
            glVertex3f(mX,fV,mZ);
            glVertex3f(nX,fV,mZ);
        } else {
            glVertex3f(mX,fV,nZ);
            glVertex3f(nX,fV,nZ);
        }
        if((j%ratio)==0) {
            SelectGLColorAlpha(AppData::PenColor[GRIDPEN],aX);
        } else {
            SelectGLColorAlpha(AppData::PenColor[GRIDPEN],aX*.5f);
        }
        if(Orient.z.x<0) {
            glVertex3f(mX,fV,mZ);
            glVertex3f(mX,fV,nZ);
        } else {
            glVertex3f(nX,fV,mZ);
            glVertex3f(nX,fV,nZ);
        }
    }
    for(fV=j=mZ;fV<nZ+.01;fV++,j++) {
        if((j%ratio)==0) {
            SelectGLColorAlpha(AppData::PenColor[GRIDPEN],aY);
        } else {
            SelectGLColorAlpha(AppData::PenColor[GRIDPEN],aY*.5f);
        }
        if(Orient.z.y<0) {
            glVertex3f(mX,mY,fV);
            glVertex3f(nX,mY,fV);
        } else {
            glVertex3f(mX,nY,fV);
            glVertex3f(nX,nY,fV);
        }
        if((j%ratio)==0) {
            SelectGLColorAlpha(AppData::PenColor[GRIDPEN],aX);
        } else {
            SelectGLColorAlpha(AppData::PenColor[GRIDPEN],aX*.5f);
        }
        if(Orient.z.x<0) {
            glVertex3f(mX,mY,fV);
            glVertex3f(mX,nY,fV);
        } else {
            glVertex3f(nX,mY,fV);
            glVertex3f(nX,nY,fV);
        }
    }
    glEnd();

    if((Flags&VF_NONPARALLEL)==0) {
        char numBuff[20];
        float Vec3::*cr;
        float Vec3::*cu;
        float Vec3::*cf;
                
        Vec3 ul;
        Vec3 ur;
        Vec3 ll;
        Vec3 lr;
        float sgnR,sgnU,fR,fU;
        float bitsize;

        glDisable(GL_DEPTH_TEST);

        if(Orient.x.x!=0.0) { cr=&Vec3::x; sgnR=Orient.x.x; }
        if(Orient.x.y!=0.0) { cr=&Vec3::y; sgnR=Orient.x.y; }
        if(Orient.x.z!=0.0) { cr=&Vec3::z; sgnR=Orient.x.z; }

        if(Orient.y.x!=0.0) { cu=&Vec3::x; sgnU=Orient.y.x; }
        if(Orient.y.y!=0.0) { cu=&Vec3::y; sgnU=Orient.y.y; }
        if(Orient.y.z!=0.0) { cu=&Vec3::z; sgnU=Orient.y.z; }

        if(Orient.z.x!=0.0) cf=&Vec3::x;
        if(Orient.z.y!=0.0) cf=&Vec3::y;
        if(Orient.z.z!=0.0) cf=&Vec3::z;

        fR=sX/128.0f*Scale;
        fU=sY/128.0f*Scale;

        bitsize=Scale/128.0f;

        mX=floor((Orient.l.*cr-fR)/coarse+.5f);
        nX=floor((Orient.l.*cr+fR)/coarse+.5f);
        mY=floor((Orient.l.*cu-fU)/coarse+.5f);
        nY=floor((Orient.l.*cu+fU)/coarse+.5f);

        ul.*cf=
        ur.*cf=
        ll.*cf=
        lr.*cf=Orient.l.*cf;

        glLoadIdentity();
        // label axes for vertical axes

        for(fV=mX*coarse;fV<(nX+.5f)*coarse;fV+=coarse) {
            sprintf(numBuff,"%g",fV);
            ul.*cr=ll.*cr=fV-sgnR*bitsize*8.0f*strlen(numBuff);
            ur.*cr=lr.*cr=fV+sgnR*bitsize*8.0f*strlen(numBuff);
            ul.*cu=ur.*cu=Orient.l.*cu-sgnU*(fU-26.0f*bitsize);
            ll.*cu=lr.*cu=Orient.l.*cu-sgnU*(fU- 2.0f*bitsize);

            SelectGLColorAlpha((COLORREF)0x0,.5f);
            glBegin(GL_QUADS);
            ul.glPlot();
            ur.glPlot();
            lr.glPlot();
            ll.glPlot();
            glEnd();

            ll.*cu+=(bitsize*(4.0f))*sgnU;
            SelectGLColor((COLORREF)0xffffff);
            ll.glRasterPos();
            glCallLists(strlen(numBuff),GL_UNSIGNED_BYTE,numBuff);
        }

        // and for the horizontal axes
        for(fV=mY*coarse;fV<(nY+.1f)*coarse;fV+=coarse) {
            sprintf(numBuff,"%g",fV);
            ul.*cr=ll.*cr=Orient.l.*cr+sgnR*(fR-bitsize*16.0f*strlen(numBuff));
            ur.*cr=lr.*cr=Orient.l.*cr+sgnR*fR;
            ul.*cu=ur.*cu=fV+sgnU*12.0f*bitsize;
            ll.*cu=lr.*cu=fV-sgnU*12.0f*bitsize;

            SelectGLColorAlpha((COLORREF)0x0,.5f);
            glBegin(GL_QUADS);
            ul.glPlot();
            ur.glPlot();
            lr.glPlot();
            ll.glPlot();
            glEnd();

            ll.*cu+=(bitsize*(4.0f))*sgnU;
            SelectGLColor((COLORREF)0x0ffffff);
            ll.glRasterPos();
            glCallLists(strlen(numBuff),GL_UNSIGNED_BYTE,numBuff);
        }
        glEnable(GL_DEPTH_TEST);
    }
}

void lsmView::jumpToOrigin() {
        Orient.setLVec(0.0,0.0,0.0);
        InvalidateView();
}

void lsmView::ViewFromFront() {
        if((Flags&VF_HANDMASK)==VF_LEFTHANDED) {
// right is -x, up is y, forward is -z
                Orient.setYVec(0.0,1.0, 0.0);
                Orient.setZVec(0.0,0.0,-1.0);
        } else {
// right is -x, up is z, forward is -y
                Orient.setYVec(0.0, 0.0,1.0);
                Orient.setZVec(0.0,-1.0,0.0);
        }
        Orient.setXVec(-1.0,0.0,0.0);
        Flags&=~VF_NONPARALLEL;
        InvalidateView();
}

void lsmView::ViewFromBack() {
        if((Flags&VF_HANDMASK)==VF_LEFTHANDED) {
// right is +x, up is +y, forward is +z
                Orient.setYVec(0.0,1.0,0.0);
                Orient.setZVec(0.0,0.0,1.0);
        } else {
// right is +x, up is +z, forward is +y
                Orient.setYVec(0.0,0.0,1.0);
                Orient.setZVec(0.0,1.0,0.0);
        }
        Orient.setXVec(1.0,0.0,0.0);
        Flags&=~VF_NONPARALLEL;
        InvalidateView();
}

void lsmView::ViewFromUp() {
        if((Flags&VF_HANDMASK)==VF_LEFTHANDED) {
// right is +x, up is +z, forward is -y
                Orient.setYVec(0.0, 0.0,1.0);
                Orient.setZVec(0.0,-1.0,0.0);
        } else {
// right is +x, up is +y, foward is -z
                Orient.setYVec(0.0,1.0, 0.0);
                Orient.setZVec(0.0,0.0,-1.0);
        }
        Orient.setXVec(1.0,0.0,0.0);
        Flags&=~VF_NONPARALLEL;
        InvalidateView();
}

void lsmView::ViewFromDown() {
        if((Flags&VF_HANDMASK)==VF_LEFTHANDED) {
// right is -x, up is +z, forward is y
                Orient.setYVec(0.0,0.0,1.0);
                Orient.setZVec(0.0,1.0,0.0);
        } else {
// right is -x, up is +y, forward is +z
                Orient.setYVec(0.0,1.0,0.0);
                Orient.setZVec(0.0,0.0,1.0);
        }
        Orient.setXVec(-1.0,0.0,0.0);

        Flags&=~VF_NONPARALLEL;
        InvalidateView();
}

void lsmView::ViewFromRight() {
        if((Flags&VF_HANDMASK)==VF_LEFTHANDED) {
// right is +z, up is +y, forward is -x
                Orient.setXVec(0.0,0.0,1.0);
                Orient.setYVec(0.0,1.0,0.0);
        } else {
// right is +y, up is +z, forward is -x
                Orient.setXVec(0.0,1.0,0.0);
                Orient.setYVec(0.0,0.0,1.0);
        }
        Orient.setZVec(-1.0,0.0,0.0);

        Flags&=~VF_NONPARALLEL;
        InvalidateView();
}

void lsmView::ViewFromLeft() {
        if((Flags&VF_HANDMASK)==VF_LEFTHANDED) {
// right is -z, up is +y, forward is +x
                Orient.setXVec(0.0,0.0,-1.0);
                Orient.setYVec(0.0,1.0, 0.0);
        } else {
// right is -y, up is +z, forward is +x
                Orient.setXVec(0.0,-1.0,0.0);
                Orient.setYVec(0.0, 0.0,1.0);
        }
        Orient.setZVec(1.0,0.0,0.0);

        Flags&=~VF_NONPARALLEL;
        InvalidateView();
}

void lsmView::ViewUV(void) {
//      Orient.identity();

        Flags&=~VF_NONPARALLEL;
        InvalidateView();
}

void lsmView::zoomInTenPercent() {
        Scale*=.909091f;
        if(Scale<.00001f) Scale=.00001f;
        InvalidateView();
}

void lsmView::zoomOutTenPercent() {
    Scale*=1.1f;
    if(Scale>100000.0f) Scale=100000.0f;
    InvalidateView();
}

void lsmView::zoomToBound(class Bound& b) {
    Vec3 redux,center;

    if(b.range_x()<=0.0f) return;
    if(b.range_y()<=0.0f) return;

    if(b.range_x()<.01 && b.range_y()<.01) {
        redux.x=(b.get_minx()+b.get_maxx())*.5f-.005f;
        center.x=(b.get_minx()+b.get_maxx())*.5f+.005f;
        redux.y=(b.get_miny()+b.get_maxy())*.5f-.005f;
        center.y=(b.get_miny()+b.get_maxy())*.5f+.005f;
        redux.z=b.get_minz();
        center.z=b.get_maxz();
        b.empty();
        b.include(redux);
        b.include(center);
    }

    if(sX*b.range_y() > sY*b.range_x()) {
        Scale=b.range_y()*67.2f/(float)sY; // =.525 * 128.0
    } else {
        Scale=b.range_x()*67.2f/(float)sX; // =.525 * 128.0
    }

    center.x=(b.get_maxx()+b.get_minx())*.5f;
    center.y=(b.get_maxy()+b.get_miny())*.5f;
    center.z=(b.get_maxz()+b.get_minz())*.5f;
    (redux=center)*=Orient;
    Orient.l=redux;
}

